; This file is part of Intellidiscs.

; Intellidiscs is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; Intellidiscs is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with Intellidiscs.  If not, see <http://www.gnu.org/licenses/>.

; Copyright 2021 Phaser Cat Games LLC

; Main game logic

; Constants.
subpixel_limit          equ 64
left_boundary           equ 49
right_boundary          equ 199
top_boundary            equ $3a
bottom_boundary         equ $cc
disc_left_boundary      equ left_boundary
disc_right_boundary     equ right_boundary + 6
disc_top_boundary       equ top_boundary
disc_bottom_boundary    equ bottom_boundary + 14
disc_return_speed       equ #32
disc_height             equ 3
held_disc_y_delta       equ 5
crouch_disc_y_delta     equ 7
crouch_left_offset      equ 0
crouch_right_offset     equ 6
player_disc_speed       equ 8
player_sprite_height    equ 17
wave_spawn_time         equ 2
follow_on_spawn_time    equ 10
health_recovery_time    equ 240
cpu_reaction_time       equ 120
player_frame_time       equ 5
shatter_frame_time      equ 10
disc_respawn_time       equ 240
recognizer_height       equ 76
recognizer_top_bound    equ $30
recognizer_center_x     equ 124
recognizer_center_y     equ $78
door_repair_lifetime    equ 120

; Color constants.
player_1_color          equ $8e
drone_color             equ $3c
armored_color           equ $36
leader_color            equ $2a
double_color            equ $1c
seeker_color            equ $0e
guard_color             equ $c8
recognizer_body_color   equ $32

; Game mode constants.
mode_single_player  equ 0
mode_game_over      equ 1

; Door state constants.
door_closed         equ 0
door_open           equ 1
door_jammed         equ 2

; Disc state constants.
disc_in_hand        equ 0
disc_in_flight      equ 1
disc_returning      equ 2
disc_blocking       equ 3
disc_shatter        equ 4
disc_floppy         equ 5
disc_destroyed      equ 6
disc_crouch         equ 7
disc_dead           equ 8
disc_stick          equ 9

; Disc action request constants.
action_none         equ -1
action_block        equ 16
action_return       equ 17

; Single player phase constants.
wave_full           equ 0
wave_reinforce      equ 1
wave_clear          equ 2

; Facing constants.
player_left         equ 0
player_right        equ 1

; CPU control flags.
cpu_player_flag     equ 128
cpu_double_flag     equ 64
cpu_seeker_flag     equ 32

; Enemy type flags.
enemy_drone         equ 0
enemy_armored       equ 64
enemy_leader        equ 128
enemy_guard         equ 192

; Recognizer battle phases.
recognizer_inactive equ 0
recognizer_init     equ 1
recognizer_enter    equ 2
recognizer_repair   equ 3
recognizer_leave    equ 4

; Recognizer eye states.
recognizer_eye_closed   equ 0
recognizer_eye_open     equ 30
recognizer_eye_sparks   equ -1

; Recognizer constants.
recognizer_eye_move_time    equ 30
recognizer_flap_move_time   equ 20
recognizer_vulnerable_time  equ 60
recognizer_spark_time       equ 10
recognizer_spark_sound_time equ 60

; Player data structure.
player_xpos             equ 0
player_ypos             equ 1
player_sub_x            equ 2
player_sub_y            equ 3
player_x_speed          equ 4
player_y_speed          equ 5
player_disc_state       equ 6
player_disc_xpos        equ 7
player_disc_ypos        equ 8
player_disc_sub_x       equ 9
player_disc_sub_y       equ 10
player_disc_x_speed     equ 11
player_disc_y_speed     equ 12
player_cpu_flags        equ 13
player_cpu_angle        equ 14
player_frame_table      equ 15
player_frame_index      equ 17
player_table_end        equ 18
player_frame_tick       equ 19
player_speed_index      equ 20
player_facing           equ 21
player_health           equ 22
player_max_health       equ 23
player_disc_offset      equ 24
player_disc_timer       equ 25
player_disc_frame       equ 26
player_last_move        equ 27
player_hit_tag          equ 28
player_recover_timer    equ 29
player_disc_action      equ 30
player_last_angle       equ 31
player_size             equ 32

; Game data locations.
player_data_1       equ $20
player_data_2       equ player_data_1 + player_size
player_data_3       equ player_data_2 + player_size
player_data_4       equ player_data_3 + player_size
player_throw_key    equ player_data_4 + player_size
spawn_timer_low     equ player_throw_key + 4
spawn_timer_high    equ spawn_timer_low + 1
door_states         equ spawn_timer_high + 1
cpu_react_timer     equ door_states + 12
score_level         equ cpu_react_timer + 1
warrior_level       equ score_level + 1
reinforced          equ warrior_level + 1
game_mode           equ reinforced + 1
enemies_killed      equ game_mode + 1
recognizer_state    equ enemies_killed + 1
wave_phase          equ recognizer_state + 1
game_over_delay     equ wave_phase + 1

; Battle mode constants.
round_start_delay   equ #120
door_interval       equ #180

; Battle mode data.  Re-uses some of the above
; locations.
round_timer equ spawn_timer_low
door_timer  equ spawn_timer_high
round_over  equ cpu_react_timer

; Recognizer data re-uses space from enemy
; data blocks.
recognizer_xpos_1           equ player_data_2
recognizer_xpos_2           equ recognizer_xpos_1 + 1
recognizer_ypos             equ recognizer_xpos_2 + 1
recognizer_side             equ recognizer_ypos + 1
recognizer_sub_x            equ recognizer_side + 1
recognizer_sub_y            equ recognizer_sub_x + 1
recognizer_hspeed           equ recognizer_sub_y + 1
recognizer_vspeed           equ recognizer_hspeed + 1
recognizer_bottom           equ recognizer_vspeed + 1
recognizer_from_top         equ recognizer_bottom + 1
recognizer_active_door      equ recognizer_from_top + 1
recognizer_repair_x         equ recognizer_active_door + 1
recognizer_repair_y         equ recognizer_repair_x + 1
recognizer_repair_sub_x     equ recognizer_repair_y + 1
recognizer_repair_sub_y     equ recognizer_repair_sub_x + 1
recognizer_repair_life      equ recognizer_repair_sub_y + 1
recognizer_probe_x          equ recognizer_repair_life + 1
recognizer_probe_y          equ recognizer_probe_x + 1
recognizer_probe_sub_x      equ recognizer_probe_y + 1
recognizer_probe_sub_y      equ recognizer_probe_sub_x + 1
recognizer_probe_hspeed     equ recognizer_probe_sub_y + 1
recognizer_probe_vspeed     equ recognizer_probe_hspeed + 1
recognizer_p3_state         equ recognizer_probe_vspeed + 1
recognizer_probe_state      equ recognizer_p3_state + 1
recognizer_eye_state        equ recognizer_probe_state + 1
recognizer_eye_xpos         equ recognizer_eye_state + 1
recognizer_eye_move         equ recognizer_eye_xpos + 1
recognizer_eye_direction    equ recognizer_eye_move + 1
recognizer_flap_timer       equ recognizer_eye_direction + 1
recognizer_spark_offset     equ recognizer_flap_timer + 1
recognizer_spark_timer      equ recognizer_spark_offset + 1
recognizer_data_size        equ recognizer_spark_timer + 1

; Button data.
trigger_states          equ $300        ; 4 values here.
shift_states            equ $304        ; 4 values here.
pausing_player          equ $308

; Configurable data and routines.
recognizer_image    equ $3000
p0_disc_input       equ $3100
p1_disc_input       equ $3200
p2_disc_input       equ $3300
p3_disc_input       equ $3400

start_game subroutine
    jsr     start_transition
    mwa     game_display_list,sDLISTL
    mwa     game_interrupt,VDSLST
    mwa     game_vblank,VVBLKD
    jsr     safe_clear_page_0
    jsr     clear_sprites
    jsr     keypad_init
    lda     recognizer_body_style
    beq     .new_recognizer
    mwa     recognizer_sprites_classic,$fe
    jmp     .copy_recognizer_sprites
.new_recognizer
    mwa     recognizer_sprites_new,$fe
.copy_recognizer_sprites
    ldy     #0
.recognizer_copy_loop
    lda     ($fe),y
    sta     $3000,y
    beq     .recognizer_loop_done
    iny
    lda     ($fe),y
    sta     $3000,y
    iny
    lda     ($fe),y
    sta     $3000,y
    iny
    jmp     .recognizer_copy_loop
.recognizer_loop_done
    mwa     p3_disc_input,$fe
    ldx     #3
.input_subroutine_loop
    lda     control_modes,x
    beq     .control_mode_keypad
    cmp     #1
    beq     .control_mode_trigger
    mwa     twin_stick_disc_handler,$fc
    ldy     #twin_stick_disc_handler_end - #twin_stick_disc_handler
    jmp     .copy_disc_handler
.control_mode_keypad
    mwa     keypad_disc_handler,$fc
    ldy     #keypad_disc_handler_end - #keypad_disc_handler
    jmp     .copy_disc_handler
.control_mode_trigger
    mwa     trigger_disc_handler,$fc
    ldy     #trigger_disc_handler_end - #trigger_disc_handler
.copy_disc_handler
    jsr     copy_memory
    jsr     adjust_input_for_player
    dec     $ff
    dex
    bpl     .input_subroutine_loop
    ldy     #0
    tya
.playfield_clear_loop
    sta     $300,y                  ; Not playfield, but clear anyway.
    sta     $2000,y
    sta     $2100,y
    sta     $2200,y
    sta     $2300,y
    dey
    bne     .playfield_clear_loop
    sta     SIZEP1
    sta     SIZEP2
    ldx     player_count
    cpx     #2
    bpl     .battle_mode_score_line
    lda     #$0e
    sta     sCOLOR0
    lda     #$86
    sta     sCOLOR1
    lda     #$46
    sta     sCOLOR2
    lda     #$1e
    sta     sCOLOR3
    lda     #$10
    sta     $2013
    jmp     .score_line_done
.battle_mode_score_line
    lda     #player_1_color
    sta     sCOLOR0
    lda     #armored_color
    sta     sCOLOR1
    sta     sCOLPM1
    lda     #guard_color
    sta     sCOLOR2
    sta     sCOLPM2
    lda     #leader_color
    sta     sCOLOR3
    sta     sCOLPM3
    lda     #$10
    sta     $2002
    lda     #$10 | $40
    sta     $2007
    cpx     #3
    bmi     .score_line_done
    lda     #$10 | $80
    sta     $200c
    cpx     #4
    bne     .score_line_done
    lda     #$10 | $c0
    sta     $2011
.score_line_done
    mwa     back_wall_pattern,$fc
    mwa     $2014,$fe
    ldy     #120
    jsr     copy_memory
    ldy     #39
.horizontal_loop
    lda     #4
    sta     $235c,y
    lda     #7
    sta     $22e4,y
    sta     $2244,y
    sta     $21a4,y
    sta     $2104,y
    dey
    bpl     .horizontal_loop
    ldx     #5
    stx     $235c
    inx
    stx     $235c + 39
    mwa     $2334,$fe
    ldx     #18
.side_wall_loop
    lda     #2
    ldy     #0
    sta     ($fe),y
    lda     #3
    ldy     #39
    sta     ($fe),y
    lda     #10
    tay
    sta     ($fe),y
    ldy     #20
    sta     ($fe),y
    ldy     #30
    sta     ($fe),y
    dex
    beq     .side_wall_done
    lda     $fe
    sec
    sbc     #40
    sta     $fe
    bcs     .side_wall_loop
    dec     $ff
    bcc     .side_wall_loop         ; Always true.
.side_wall_done
    ldx     #8                      ; Cap the grid lines.
    stx     $22e4
    stx     $2244
    stx     $21a4
    stx     $2104
    inx
    stx     $22e4 + 39
    stx     $2244 + 39
    stx     $21a4 + 39
    stx     $2104 + 39
    ldx     #11                     ; Grid intersections.
    stx     $22e4 + 10
    stx     $2244 + 10
    stx     $21a4 + 10
    stx     $2104 + 10
    stx     $22e4 + 20
    stx     $2244 + 20
    stx     $21a4 + 20
    stx     $2104 + 20
    stx     $22e4 + 30
    stx     $2244 + 30
    stx     $21a4 + 30
    stx     $2104 + 30
    inx                             ; Bottom row grid points.
    stx     $235c + 10
    stx     $235c + 20
    stx     $235c + 30
    mva     #1,PRIOR
    ldy     #3
.trigger_loop
    lda     #1
    sta     trigger_states,y
    lda     #8
    sta     shift_states,y
    dey
    bpl     .trigger_loop
    ldy     #player_data_1
    jsr     init_player_data
    lda     #action_none
    sta     player_data_1 + player_disc_action
    sta     player_data_2 + player_disc_action
    sta     player_data_3 + player_disc_action
    sta     player_data_4 + player_disc_action
    lda     #wave_clear
    sta     wave_phase
    mva     #player_1_color,sCOLPM0
    mva     #wave_spawn_time,spawn_timer_high
    mva     #cpu_reaction_time,cpu_react_timer
    jsr     complete_transition
    mva     #$c8,AUDCTL
    mva     #$8a,SKCTL
    lda     player_count
    cmp     #1
    beq     start_arcade_game
    lda     #-1                     ; Terminate the tie-list.
    sta     $f0
    jmp     start_battle_game

; Main game loop.
start_arcade_game subroutine
    mva     #124,HPOSP0
    sta     player_data_1 + player_xpos
    mva     #126,HPOSM0
    sta     player_data_1 + player_disc_xpos
    lda     #2
    sta     player_data_1 + player_disc_offset
    lda     #$86
    sta     player_data_1 + player_ypos
    lda     #$86 + held_disc_y_delta
    sta     player_data_1 + player_disc_ypos
    lda     #disc_dead
    sta     player_data_2 + player_disc_state
    sta     player_data_3 + player_disc_state
    sta     player_data_4 + player_disc_state
game_loop
    sta     HITCLR
    jsr     wait_for_vb
    jsr     keypad_process
    lda     game_mode
    cmp     #mode_game_over
    bne     .game_active
    jmp     start_game_over
.game_active
    lda     keypad_last_key
    cmp     #KEYPAUSE
    bne     .not_paused
    jsr     pause_loop
    jmp     game_loop
.not_paused
    jsr     process_player_1_input
    jsr     process_all_players
    jsr     process_player_1_disc
    ldy     #player_data_2
    jsr     enemy_is_attacking
    bne     .no_player_hit_from_2
    lda     M1PL
    and     #1
    beq     .no_player_hit_from_2
    jsr     enemy_is_guard
    beq     .guard_hit
    ldy     #player_data_1
    ldx     player_data_2 + player_cpu_flags
    lda     #$fc
    jsr     hit_player
    beq     .no_player_hit_from_2
    ldy     #player_data_2
    ldx     #$f3
    jsr     handle_disc_after_player_hit
.no_player_hit_from_2
    ldy     #player_data_3
    jsr     enemy_is_attacking
    bne     .no_player_hit_from_3
    lda     M2PL
    and     #1
    beq     .no_player_hit_from_3
    jsr     enemy_is_guard
    beq     .guard_hit
    ldy     #player_data_1
    ldx     player_data_3 + player_cpu_flags
    lda     #$fc
    jsr     hit_player
    beq     .no_player_hit_from_3
    ldy     #player_data_3
    ldx     #$cf
    jsr     handle_disc_after_player_hit
.no_player_hit_from_3
    ldy     #player_data_4
    jsr     enemy_is_attacking
    bne     .no_player_hit_from_4
    lda     M3PL
    and     #1
    beq     .no_player_hit_from_4
    jsr     enemy_is_guard
    beq     .guard_hit
    ldy     #player_data_1
    ldx     player_data_4 + player_cpu_flags
    lda     #$fc
    jsr     hit_player
    beq     .no_player_hit_from_4
    ldy     #player_data_4
    ldx     #$3f
    jsr     handle_disc_after_player_hit
    jmp     .no_player_hit_from_4
.guard_hit
    lda     #4
    jsr     DOTUNE
    lda     #5
    jsr     DOTUNE
    lda     #mode_game_over
    sta     game_mode
.no_player_hit_from_4
    lda     player_data_1 + player_disc_state
    cmp     #disc_blocking
    bne     .no_block_check
    ldx     player_data_1 + player_disc_xpos
    ldy     player_data_1 + player_disc_ypos
    jsr     check_block
.no_block_check
    lda     wave_phase              ; Process the wave phase.
    beq     .process_full_wave
    cmp     #wave_reinforce
    beq     .process_reinforce_wave
.process_spawn_timer
    dec     spawn_timer_low
    bpl     .wave_phase_done
    lda     #59
    sta     spawn_timer_low
    dec     spawn_timer_high
    bpl     .wave_phase_done
    jsr     spawn_new_wave
    jmp     .wave_phase_done
.process_full_wave
    jsr     count_living_discs      ; If a disc has been eliminated,
    cpx     #3                      ; start the reinforcement timer.
    beq     .wave_phase_done
    inc     wave_phase              ; Go to reinforcements.
    jmp     .wave_phase_done
.process_reinforce_wave
    jsr     count_living_discs      ; If 3 discs, go back to full.
    cpx     #3
    bne     .not_full
    dec     wave_phase              ; Full, go back a state.
    jmp     .wave_phase_done
.not_full
    stx     $ff                     ; Save the disc count.
    jsr     count_living_enemies    ; If 0 enemies and 0 discs, wave clear.
    bne     .process_spawn_timer    ; Still reinforcing.
    ldx     $ff
    bne     .wave_phase_done
    lda     #wave_spawn_time
    sta     spawn_timer_high
    lda     #0
    sta     spawn_timer_low
    inc     wave_phase
.wave_phase_done
    dec     cpu_react_timer         ; Check the CPU timer.
    bne     .no_cpu_react
    lda     #cpu_reaction_time      ; Reset the timer.
    sta     cpu_react_timer
    ldy     #player_data_2          ; Evaluate all three enemies.
    lda     #$f3                    ; Disc erasure mask.
    jsr     cpu_adjust
    ldy     #player_data_3
    lda     #$cf
    jsr     cpu_adjust
    ldy     #player_data_4
    lda     #$3f
    jsr     cpu_adjust
.no_cpu_react
    jmp     game_loop

; Logic loop for the recognizer battles,
recognizer_loop subroutine
    sta     HITCLR
    jsr     wait_for_vb
    jsr     keypad_process
    lda     keypad_last_key
    cmp     #KEYPAUSE
    bne     .not_paused
    jsr     pause_loop
    lda     recognizer_state
    cmp     #recognizer_enter
    beq     .unpause_enter_sound
    cmp     #recognizer_leave
    bne     recognizer_loop
    lda     recognizer_eye_state
    cmp     #recognizer_eye_sparks
    bne     .unpause_enter_sound
    lda     #$e
    jsr     DOTUNE
    lda     #$f
    jsr     DOTUNE
    jmp     recognizer_loop
.unpause_enter_sound
    lda     #$16
    jsr     DOTUNE
    lda     #$17
    jsr     DOTUNE
    jmp     recognizer_loop
.not_paused
    lda     recognizer_state
    cmp     #recognizer_init
    beq     .init_recognizer
    cmp     #recognizer_enter
    bne     .not_entering
    jmp     .handle_entry
.not_entering
    cmp     #recognizer_repair
    bne     .leaving
    jmp     .handle_repair
.leaving
    jmp     .handle_leaving
.init_recognizer
    lda     #$16
    jsr     DOTUNE
    lda     #$17
    jsr     DOTUNE
    lda     #1                      ; Double size players 1 and 2.
    sta     SIZEP1
    sta     SIZEP2
    sta     cpu_react_timer         ; Probe direction on first frame.
    lda     #recognizer_body_color
    sta     sCOLPM1
    sta     sCOLPM2
    lda     #recognizer_spark_time
    sta     recognizer_spark_timer
.pick_entry_side
    lda     RANDOM                  ; Choose an entry side.
    and     #$03                    ; Mask to 0-3.
    beq     .pick_entry_side        ; Choose until not 0.
    sta     recognizer_side         ; Store the side.
    cmp     #1
    beq     .enter_left
    cmp     #2
    beq     .enter_top
.enter_right
    lda     #223
    sta     recognizer_xpos_1
    lda     #239
    sta     recognizer_xpos_2
    lda     #-12
    sta     recognizer_hspeed
    ldy     #-1
    sty     recognizer_vspeed
    iny
    sty     recognizer_from_top
    ldx     #$6b
    jmp     .init_common
.enter_left
    lda     #0
    sta     recognizer_xpos_1
    sta     recognizer_from_top
    lda     #16
    sta     recognizer_xpos_2
    lda     #12
    sta     recognizer_hspeed
    lda     #1
    sta     recognizer_vspeed
    ldx     #$60
    jmp     .init_common
.enter_top
    lda     #80
    sta     recognizer_xpos_1
    lda     #96
    sta     recognizer_xpos_2
    lda     #4
    sta     recognizer_hspeed
    lda     #16
    sta     recognizer_vspeed
    sta     recognizer_from_top
    ldx     #recognizer_top_bound
.init_common
    stx     recognizer_ypos
    stx     recognizer_bottom       ; Only used on top entry.
    ldx     #1
    stx     recognizer_eye_direction
    dex
    stx     recognizer_sub_x
    stx     recognizer_sub_y
    stx     recognizer_probe_sub_x
    stx     recognizer_probe_sub_y
    stx     recognizer_p3_state
    stx     recognizer_probe_state
    stx     recognizer_eye_state
    dex
    stx     recognizer_active_door  ; -1 to trigger new door.
    lda     #7
    sta     recognizer_spark_offset
    lda     #recognizer_center_x
    sta     recognizer_probe_x
    lda     #recognizer_center_y
    sta     recognizer_probe_y
    lda     #recognizer_eye_move_time
    sta     recognizer_eye_move
    lda     #recognizer_flap_move_time
    sta     recognizer_flap_timer
    lda     #16
    sta     recognizer_eye_xpos
    jsr     draw_recognizer
    inc     recognizer_state
.handle_entry
    jsr     move_recognizer
    lda     recognizer_xpos_1
    cmp     #113
    beq     .end_entry
    jmp     .handle_eye
.end_entry
    lda     #$16
    jsr     KILLTUNE
    lda     #$17
    jsr     KILLTUNE
    inc     recognizer_state
    jmp     .handle_eye
.handle_repair
    ldy     recognizer_active_door  ; Check the door state.
    bpl     .door_processing_end
    ldx     #0                      ; Check for the first jammed door.
.find_door_loop
    ldy     door_repair_sequence,x
    lda     door_states,y
    cmp     #door_jammed            ; If it's jammed, fix it.
    beq     .found_door
    cpx     #11                     ; Check for list end.
    beq     .end_repair
    inx                             ; Next door.
    jmp     .find_door_loop
.found_door
    lda     #recognizer_vulnerable_time
    sta     recognizer_eye_state    ; Expose the eye.
    sty     recognizer_active_door
    lda     #recognizer_center_x
    sta     recognizer_repair_x
    lda     #recognizer_center_y
    sta     recognizer_repair_y
    lda     #0
    sta     recognizer_repair_sub_x
    sta     recognizer_repair_sub_y
    lda     #door_repair_lifetime
    sta     recognizer_repair_life
    jmp     .door_processing_end
.end_repair
    lda     #$16
    jsr     DOTUNE
    lda     #$17
    jsr     DOTUNE
.start_exit
    jsr     erase_recognizer_door
    jsr     erase_recognizer_probe
    lda     #16
    sta     recognizer_vspeed
    lda     #0                      ; Release the player.
    sta     recognizer_probe_state
    sta     recognizer_hspeed
    inc     recognizer_state
    lda     recognizer_eye_state    ; Erase the eye if necessary.
    bpl     .recognizer_intact
    ldy     recognizer_ypos
    lda     #$0f
    and     $1304,y
    sta     $1304,y
    lda     #$0f
    and     $1305,y
    sta     $1305,y
    lda     #$0e
    sta     sCOLPM3
    lda     #125
    sta     HPOSP3
.recognizer_intact
    jmp     recognizer_loop
.door_processing_end
    ldy     recognizer_active_door
    dec     recognizer_repair_life  ; Process the active door panel.
    bne     .door_still_alive
    tya
    pha
    lda     #$c
    jsr     DOTUNE
    lda     #$d
    jsr     DOTUNE
    pla
    tay
    jsr     close_door
    jsr     erase_recognizer_door
    ldy     #-1
    sty     recognizer_active_door
    lda     #recognizer_center_x
    sta     recognizer_repair_x
    lda     #recognizer_center_y
    sta     recognizer_repair_y
    jmp     .flip_p3_sprite
.door_still_alive
    lda     recognizer_repair_sub_x
    clc
    adc     door_repair_h_speed,y
    bmi     .door_move_left
    cmp     #subpixel_limit
    bmi     .store_door_sub_x
    inc     recognizer_repair_x
    sec
    sbc     #subpixel_limit
    jmp     .store_door_sub_x
.door_move_left
    dec     recognizer_repair_x
    clc
    adc     #subpixel_limit
.store_door_sub_x
    sta     recognizer_repair_sub_x
    lda     recognizer_repair_sub_y
    clc
    adc     door_repair_v_speed,y
    bmi     .door_move_up
    cmp     #subpixel_limit
    bmi     .store_door_sub_y
    inc     recognizer_repair_y
    sec
    sbc     #subpixel_limit
    jmp     .store_door_sub_y
.door_move_up
    dec     recognizer_repair_y
    clc
    adc     #subpixel_limit
.store_door_sub_y
    sta     recognizer_repair_sub_y
    lda     recognizer_probe_state
    beq     .no_p3_sprite_flip
    jmp     .flip_p3_sprite
.no_p3_sprite_flip
    lda     player_data_1 + player_health
    beq     .no_new_probe_direction ; Already dead.
    dec     cpu_react_timer
    bne     .no_new_probe_direction
    lda     #cpu_reaction_time
    sta     cpu_react_timer
    clc
    lda     recognizer_probe_x
    adc     #3
    sta     $fe
    lda     recognizer_probe_y
    adc     #7
    sta     $ff
    jsr     angle_to_player
    sta     $ff                     ; Save the angle.
    lda     player_data_1 + player_ypos
    clc
    adc     #player_sprite_height
    cmp     recognizer_ypos
    bcs     .normal_probe_speed     ; Not above the body.
    lda     player_data_1 + player_xpos
    cmp     recognizer_xpos_1       ; Pick a speed based on the player's
    bcc     .player_not_above       ; position around the recognizer.
    sbc     recognizer_xpos_2       ; Carry is set.
    cmp     #8
    bcs     .player_not_above
    lda     #52                     ; Super fast speed.
    jmp     .got_probe_speed
.player_not_above                   ; Above, but not directly above,
    lda     #35                     ; slightly faster.
    jmp     .got_probe_speed
.normal_probe_speed
    lda     #18                     ; Normal speed.
.got_probe_speed
    clc
    adc     $ff                     ; Add the angle.
    tay
    lda     player_hspeed_table,y
    sta     recognizer_probe_hspeed
    lda     player_vspeed_table,y
    sta     recognizer_probe_vspeed
.no_new_probe_direction
    lda     recognizer_probe_hspeed
    clc
    adc     recognizer_probe_sub_x
    bmi     .probe_move_left
    cmp     #subpixel_limit
    bmi     .store_probe_sub_x
    inc     recognizer_probe_x
    sec
    sbc     #subpixel_limit
    jmp     .store_probe_sub_x
.probe_move_left
    dec     recognizer_probe_x
    clc
    adc     #subpixel_limit
.store_probe_sub_x
    sta     recognizer_probe_sub_x
    lda     recognizer_probe_vspeed
    clc
    adc     recognizer_probe_sub_y
    bmi     .probe_move_up
    cmp     #subpixel_limit
    bmi     .store_probe_sub_y
    inc     recognizer_probe_y
    sec
    sbc     #subpixel_limit
    jmp     .store_probe_sub_y
.probe_move_up
    dec     recognizer_probe_y
    clc
    adc     #subpixel_limit
.store_probe_sub_y
    sta     recognizer_probe_sub_y
.flip_p3_sprite
    lda     recognizer_p3_state     ; Set up the p3 sprite for
    eor     #1                      ; this frame, 0 is the probe.
    sta     recognizer_p3_state
    beq     .show_probe
    lda     P3PL                    ; Before leaving the probe state,
    and     #1                      ; check for collision with the player.
    beq     .no_probe_collision
    sta     recognizer_probe_state  ; Player captured.
    ldx     #0
    stx     player_data_1 + player_x_speed
    stx     player_data_1 + player_y_speed
    ldy     #player_data_1
    jsr     update_player_animation
    jsr     erase_recognizer_probe  ; Lock the probe to the player.
    lda     player_data_1 + player_xpos
    sta     recognizer_probe_x
    lda     player_data_1 + player_ypos
    sta     recognizer_probe_y
.no_probe_collision
    jsr     erase_recognizer_probe
    ldx     #16                     ; 16 bytes.
    ldy     recognizer_repair_y
    lda     #$ff
.repair_draw_loop
    sta     $1700,y
    dex
    beq     .repair_draw_done
    iny
    jmp     .repair_draw_loop
.repair_draw_done
    lda     recognizer_repair_x
    sta     HPOSP3
    lda     #$7c
    sta     COLPM3
    jmp     .p3_sprite_done
.show_probe
    jsr     erase_recognizer_door
    lda     #13                     ; 13 bytes.
    sta     $ff
    ldy     recognizer_probe_y      ; Write index.
    ldx     #0                      ; Read index.
.probe_draw_loop
    lda     probe_sprite,x
    sta     $1700,y
    dec     $ff
    beq     .probe_draw_done
    iny
    inx
    jmp     .probe_draw_loop
.probe_draw_done
    lda     recognizer_probe_x
    sta     HPOSP3
    lda     #$0f
    sta     COLPM3
.p3_sprite_done
    jmp     .handle_eye
.handle_leaving
    jsr     move_recognizer
    lda     recognizer_ypos
    cmp     #$ad + #recognizer_height
    bne     .handle_eye
    lda     #0
    sta     SIZEP1
    sta     SIZEP2
    ldy     #recognizer_data_size - 1
.recognizer_data_clear_loop         ; Blank the recognizer data,
    sta     recognizer_xpos_1,y     ; it belongs to the enemies again.
    dey
    bpl     .recognizer_data_clear_loop
    lda     #disc_dead              ; Set disc states to dead.
    sta     player_data_2 + player_disc_state
    sta     player_data_3 + player_disc_state
    sta     player_data_4 + player_disc_state
    ldy     #0                      ; Clear any lingering sprite data.
    tya
.recognizer_cleanup_loop
    sta     $1500,y
    sta     $1600,y
    sta     $1700,y
    dey
    bne     .recognizer_cleanup_loop
    lda     #$16
    jsr     KILLTUNE
    lda     #$17
    jsr     KILLTUNE
    ldy     player_data_1 + player_health
    bne     .resume_game_loop
    jmp     start_game_over
.resume_game_loop
    jmp     game_loop
.handle_eye
    lda     recognizer_eye_state
    beq     .no_eye_timer
    bpl     .handle_eye_timer
    jmp     .handle_sparks
.handle_eye_timer
    dec     recognizer_eye_state
.no_eye_timer
    dec     recognizer_eye_move
    bpl     .no_eye_move
    lda     #recognizer_eye_move_time
    sta     recognizer_eye_move
    lda     recognizer_eye_direction
    beq     .eye_move_left
    lda     recognizer_eye_xpos
    cmp     #16
    bne     .no_reverse_to_left
    dec     recognizer_eye_direction
.no_reverse_to_left
    inc     recognizer_eye_xpos
    jmp     .no_eye_move
.eye_move_left
    lda     recognizer_eye_xpos
    cmp     #14
    bne     .no_reverse_to_right
    inc     recognizer_eye_direction
.no_reverse_to_right
    dec     recognizer_eye_xpos
.no_eye_move
    lda     recognizer_xpos_1       ; Position the eye.
    clc
    adc     recognizer_eye_xpos
    sta     HPOSM3
    ldx     recognizer_eye_state    ; Eye state 0 is invulnerable.
    beq     .invulnerable_eye
    lda     #0
.invulnerable_eye
    sta     HPOSM2                  ; Place the red eye on top.
.eye_move_done
    dec     recognizer_flap_timer   ; Move the guard flaps.
    beq     .handle_flaps
    jmp     .handle_player
.handle_flaps
    lda     #recognizer_flap_move_time
    sta     recognizer_flap_timer
    lda     recognizer_from_top     ; Check for top-entry hackery.
    beq     .normal_flap_draw
    lda     recognizer_bottom
    sec
    sbc     #recognizer_top_bound + 70
    bcs     .flaps_on_screen
    jmp     .handle_player          ; No flap draw yet.
.flaps_on_screen
    tay
    lda     #$04
    eor     $1500,y
    sta     $1500,y
    lda     #$04
    eor     $14ff,y
    sta     $14ff,y
    lda     #$20
    eor     $1600,y
    sta     $1600,y
    lda     #$20
    eor     $15ff,y
    sta     $15ff,y
    jmp     .handle_player
.normal_flap_draw
    lda     #$04
    ldy     recognizer_ypos
    eor     $1504,y
    sta     $1504,y
    lda     #$04
    eor     $1505,y
    sta     $1505,y
    lda     #$20
    eor     $1604,y
    sta     $1604,y
    lda     #$20
    eor     $1605,y
    sta     $1605,y
    jmp     .handle_player
.handle_sparks
    dec     recognizer_repair_life
    bne     .no_sparks_sound
    lda     #$10
    jsr     KILLTUNE
    lda     #$11
    jsr     KILLTUNE
    lda     #$e
    jsr     DOTUNE
    lda     #$f
    jsr     DOTUNE
    lda     #recognizer_spark_sound_time
    sta     recognizer_repair_life
.no_sparks_sound
    lda     recognizer_ypos
    clc
    adc     #5
    tay
    ldx     #7
    stx     $ff                     ; Row count.
    ldx     recognizer_spark_offset
.sparks_loop
    lda     sparker_sprite,x
    sta     $1700,y
    dec     $ff
    bmi     .erase_top_sparks_row
    dex
    dey
    jmp     .sparks_loop
.erase_top_sparks_row
    dey
    lda     #0
    sta     $1700,y
    dec     recognizer_spark_timer
    bne     .spark_done
    lda     #recognizer_spark_time
    sta     recognizer_spark_timer
    lda     recognizer_spark_offset ; Update the offset.
    cmp     #23
    beq     .spark_animation_loop
    clc
    adc     #8
    jmp     .store_spark_offset
.spark_animation_loop
    lda     #7
.store_spark_offset
    sta     recognizer_spark_offset
.spark_done
    jmp     .eye_move_done
.handle_player
    lda     recognizer_probe_state  ; Do nothing on capture.
    bne     .ignore_player_input
    lda     player_data_1 + player_health
    beq     .ignore_player_input    ; Already dead.
    jsr     process_player_1_input
.ignore_player_input
    jsr     process_player_1
    lda     player_data_1 + player_health
    beq     .no_recognizer_contact  ; Already in death animation.
    lda     #$ff
    sta     keypad_last_key
    lda     P1PL                    ; Contact with the recognizer
    ora     P2PL                    ; is instantly fatal.
    and     #1
    beq     .no_recognizer_contact
    ldy     #player_data_1
    jsr     start_death_animation
    lda     #0
    sta     recognizer_probe_hspeed
    sta     recognizer_probe_vspeed
.no_recognizer_contact
    lda     player_data_1 + player_disc_state
    cmp     #disc_in_flight         ; Check disc interactions.
    beq     .disc_check
    jmp     .no_disc_check
.disc_check
    lda     recognizer_state        ; Allow top disc entry during
    cmp     #recognizer_repair      ; repair mode.
    bne     .disable_top_shot
    lda     player_data_1 + player_disc_x_speed
    bpl     .no_absolute_conversion
    jsr     invert
.no_absolute_conversion
    cmp     #25                     ; Only allow disc angles that
    bcs     .disable_top_shot       ; are mostly down.
    lda     player_data_1 + player_disc_y_speed
    cmp     #48
    bcc     .disable_top_shot
    lda     recognizer_ypos
    clc
    adc     #4
    cmp     player_data_1 + player_disc_ypos
    bcs     .no_floppy
.disable_top_shot
    lda     M0PL                    ; Contact with the recognizer
    and     #$06                    ; causes floppy disc.
    beq     .no_floppy
    jsr     make_floppy_disc
.no_floppy                          ; Check for eye damage on the
    lda     recognizer_eye_state    ; floppy frame.
    cmp     #recognizer_eye_sparks
    beq     .no_disc_check          ; Ignore if eye destroyed.
    lda     recognizer_ypos
    clc                             ; An in-flight disc is at ypos + 1.
    adc     #3
    cmp     player_data_1 + player_disc_ypos
    beq     .disc_y_check_passed
    adc     #1
    cmp     player_data_1 + player_disc_ypos
    bne     .no_disc_check          ; Didn't hit the eye, y is off.
.disc_y_check_passed
    lda     recognizer_eye_xpos     ; Check the horizontal.
    clc
    adc     recognizer_xpos_1
    sec
    sbc     player_data_1 + player_disc_xpos
    beq     .handle_eye_hit
    cmp     #-1
    beq     .handle_eye_hit
    cmp     #1
    bne     .no_disc_check
.handle_eye_hit
    lda     recognizer_eye_state    ; Check the vulnerability.
    bne     .destroy_recognizer     ; If not 0, eye is open.
    lda     player_data_1 + player_disc_state
    cmp     #disc_floppy            ; Make the disc floppy if it
    beq     .no_disc_check          ; isn't already.
    jsr     make_floppy_disc
    jmp     recognizer_loop
.destroy_recognizer
    lda     #$10
    jsr     DOTUNE
    lda     #$11
    jsr     DOTUNE
    lda     #recognizer_spark_sound_time
    sta     recognizer_repair_life  ; Reuse this space for sound timing.
    lda     #recognizer_eye_sparks
    sta     recognizer_eye_state
    ldx     #$11
    jsr     add_to_score
    jmp     .start_exit
.no_disc_check
    jmp     recognizer_loop

; Battle mode game loop.
start_battle_game subroutine
    ldx     player_count
    dex
    dex
    ldy     #player_data_1
    jsr     init_player_data
    lda     battle_p1_x,x
    sta     HPOSP0
    sta     player_data_1 + player_xpos
    lda     #2
    sta     player_data_1 + player_disc_offset
    clc
    adc     player_data_1 + player_xpos
    sta     HPOSM0
    sta     player_data_1 + player_disc_xpos
    lda     battle_p1_y,x
    sta     player_data_1 + player_ypos
    clc
    adc     #held_disc_y_delta
    sta     player_data_1 + player_disc_ypos
    ldy     #player_data_2
    jsr     init_player_data
    lda     battle_p2_x,x
    sta     HPOSP1
    sta     player_data_2 + player_xpos
    lda     #2
    sta     player_data_2 + player_disc_offset
    clc
    adc     player_data_2 + player_xpos
    sta     HPOSM1
    sta     player_data_2 + player_disc_xpos
    lda     battle_p2_y,x
    sta     player_data_2 + player_ypos
    clc
    adc     #held_disc_y_delta
    sta     player_data_2 + player_disc_ypos
    cpx     #1                      ; Count has been decremented twice.
    bmi     .players_ready
    dex
    ldy     #player_data_3
    jsr     init_player_data
    lda     battle_p3_x,x
    sta     HPOSP2
    sta     player_data_3 + player_xpos
    lda     #2
    sta     player_data_3 + player_disc_offset
    clc
    adc     player_data_3 + player_xpos
    sta     HPOSM2
    sta     player_data_3 + player_disc_xpos
    lda     battle_p3_y,x
    sta     player_data_3 + player_ypos
    clc
    adc     #held_disc_y_delta
    sta     player_data_3 + player_disc_ypos
    cpx     #1                      ; Count has been decremented thrice.
    bne     .players_ready
    ldy     #player_data_4
    jsr     init_player_data
    mva     #battle_p4_x,HPOSP3
    sta     player_data_4 + player_xpos
    mva     #battle_p4_x + 2,HPOSM3
    sta     player_data_4 + player_disc_xpos
    lda     #2
    sta     player_data_4 + player_disc_offset
    lda     #battle_p4_y
    sta     player_data_4 + player_ypos
    lda     #battle_p4_y + held_disc_y_delta
    sta     player_data_4 + player_disc_ypos
.players_ready
    bit     $f0                     ; Check the tie-list.
    bmi     .no_tie_breaker
    ldx     #0                      ; Tie index.
    stx     $f5                     ; Expected player.
    ldy     #player_data_1          ; Expected player data.
.tie_breaker_loop
    lda     $f0,x
    bmi     .eliminate_player
    cmp     $f5
    beq     .found_player
.eliminate_player
    lda     $f5
    cmp     #4                      ; All player's done?
    beq     .tie_break_ready
    lda     #0                      ; Eliminate the player.
    sta     player_health,y
    sta     player_xpos,y
    sty     $ff
    ldy     $f5
    sta     HPOSP0,y
    sta     HPOSM0,y
    inc     $f5
    lda     $ff
    clc
    adc     #player_size
    tay
    jmp     .tie_breaker_loop
.found_player
    lda     #1                      ; One hit in the tie-breaker.
    sta     player_health,y
    inc     $f5                     ; Next player.
    tya
    clc
    adc     #player_size
    tay
    inx
    jmp     .tie_breaker_loop
.no_tie_breaker
    lda     #5                      ; Set health to 5..
    sta     player_data_1 + player_health
    sta     player_data_1 + player_max_health
    sta     player_data_2 + player_health
    sta     player_data_2 + player_max_health
    sta     player_data_3 + player_health
    sta     player_data_3 + player_max_health
    sta     player_data_4 + player_health
    sta     player_data_4 + player_max_health
.tie_break_ready
    lda     #68                     ; Maximum speed table.
    sta     player_data_1 + player_speed_index
    sta     player_data_2 + player_speed_index
    sta     player_data_3 + player_speed_index
    sta     player_data_4 + player_speed_index
    ldy     #11
.door_close_loop
    jsr     close_door
    dey
    bpl     .door_close_loop
    iny
    tya
.sprite_clear_loop
    sta     $1300,y
    sta     $1400,y
    sta     $1500,y
    sta     $1600,y
    sta     $1700,y
    dey
    bne     .sprite_clear_loop
    ldx     #1
    stx     door_timer
    dex
    stx     round_over
    lda     #round_start_delay
    sta     round_timer
    ; Fall through into the main loop.

; Main logic loop for battle mode.
battle_loop subroutine
    sta     HITCLR
    jsr     wait_for_vb
    jsr     keypad_process
    lda     round_timer
    beq     .round_in_progress
    jsr     keypad_reset
    dec     round_timer
    bne     .player_4_dead
    lda     round_over
    beq     .player_4_dead
    jmp     .end_round
.round_in_progress
    ldy     player_count
.pause_loop
    dey
    bmi     .pause_loop_done
    lda     keypad_last_key,y
    cmp     #KEYPAUSE
    bne     .pause_loop
.pause_game
    sty     pausing_player
    jsr     pause_loop
    jmp     battle_loop
.pause_loop_done
    dec     door_timer
    bne     .no_doors
    jsr     close_all_open_doors
    ldx     #0                      ; Build a list of open doors.
    ldy     #0
.find_closed_doors_loop
    lda     door_states,y
    bne     .next_door
    sty     $f0,x                   ; Save the number.
    inx
.next_door
    iny
    cpy     #12
    bne     .find_closed_doors_loop
.doors_done
    cpx     #0                      ; Any doors available?
    beq     .no_doors
    stx     $ff                     ; Save the door count.
.random_door_loop
    lda     RANDOM
    and     #$0f                    ; Mask to 16.
    cmp     $ff                     ; Need a value smaller than
    bpl     .random_door_loop       ; the door count.
    tay
    lda     $f0,y
    tay
    jsr     open_door
    lda     #door_interval
    sta     door_timer
.no_doors
    lda     player_data_1 + player_health
    beq     .player_1_dead
    jsr     process_player_1_input
.player_1_dead
    lda     player_data_2 + player_health
    beq     .player_2_dead
    jsr     process_player_2_input
.player_2_dead
    lda     player_data_3 + player_health
    beq     .player_3_dead
    jsr     process_player_3_input
.player_3_dead
    lda     player_data_4 + player_health
    beq     .player_4_dead
    jsr     process_player_4_input
.player_4_dead
    jsr     process_all_players
    jsr     process_all_player_discs
    lda     player_data_1 + player_disc_state
    cmp     #disc_blocking
    bne     .no_block_p1
    ldx     player_data_1 + player_disc_xpos
    ldy     player_data_1 + player_disc_ypos
    jsr     check_block
.no_block_p1
    lda     player_data_2 + player_disc_state
    cmp     #disc_blocking
    bne     .no_block_p2
    ldx     player_data_2 + player_disc_xpos
    ldy     player_data_2 + player_disc_ypos
    jsr     check_block
.no_block_p2
    lda     player_data_3 + player_disc_state
    cmp     #disc_blocking
    bne     .no_block_p3
    ldx     player_data_3 + player_disc_xpos
    ldy     player_data_3 + player_disc_ypos
    jsr     check_block
.no_block_p3
    lda     player_data_4 + player_disc_state
    cmp     #disc_blocking
    bne     .no_block_p4
    ldx     player_data_4 + player_disc_xpos
    ldy     player_data_4 + player_disc_ypos
    jsr     check_block
.no_block_p4
    lda     round_over
    bne     .end_round_check
    ldx     #0                      ; Count the number of living players.
    ldy     player_count
    lda     player_data_1 + player_health
    beq     .p1_dead
    inx
.p1_dead
    dey
    lda     player_data_2 + player_health
    beq     .p2_dead
    inx
.p2_dead
    dey
    beq     .dead_check_done
    lda     player_data_3 + player_health
    beq     .p3_dead
    inx
.p3_dead
    dey
    beq     .dead_check_done
    lda     player_data_4 + player_health
    beq     .dead_check_done
    inx
.dead_check_done
    cpx     #2                      ; End round on less than 2 players
    bpl     .end_round_check        ; alive.  It's possible that everyone
    inc     round_over              ; is dead.
    lda     #round_start_delay
    sta     round_timer
    jsr     stop_cpu_players
    ldy     #player_data_1
    jsr     stop_player
.end_round_check
    jmp     battle_loop
.end_round
    ldx     #0                      ; List all players with 10+ points.
    ldy     #0
    stx     $ff                     ; Current winning score.
    lda     $2001
    beq     .p1_loss
    lda     $2002                   ; Score to beat.
    sta     $ff
    sty     $f0,x
    inx
.p1_loss
    iny
    lda     $2006
    beq     .p2_loss
    lda     $2007
    and     #$3f                    ; Mask off the color bits.
    cmp     $ff                     ; Check against the current high.
    bmi     .p2_loss
    beq     .p2_tie
    ldx     #0                      ; New winner, reset x.
    sta     $ff                     ; Update the winning score.
.p2_tie
    sty     $f0,x
    inx
.p2_loss
    iny
    lda     $200b
    beq     .p3_loss
    lda     $200c
    and     #$3f
    cmp     $ff
    bmi     .p3_loss
    beq     .p3_tie
    ldx     #0
    sta     $ff
.p3_tie
    sty     $f0,x
    inx
.p3_loss
    iny
    lda     $2010
    beq     .p4_loss
    lda     $2011
    and     #$3f
    cmp     $ff
    bmi     .p4_loss
    beq     .p4_tie
    ldx     #0
    sta     $ff
.p4_tie
    sty     $f0,x
    inx
.p4_loss
    cpx     #1                      ; How many players won?
    bmi     .next_round             ; 0, not over yet.
    beq     .end_battle             ; 1 winner, so no tie breaker.
.next_round
    lda     #-1                     ; Terminate the tie-list.
    sta     $f0,x
    dec     round_over
    jmp     start_battle_game
.end_battle
    jmp     start_game_over

; Pause loop!
pause_loop subroutine
    jsr     CLEARTUN
    ldy     pausing_player
    lda     #$ff
    sta     keypad_last_key,y
    jsr     wait_for_vb
    jsr     keypad_process
    ldy     pausing_player
    lda     keypad_last_key,y
    cmp     #KEYPAUSE
    beq     .unpause
    cmp     #KEYSTART
    beq     .restart
    cmp     #KEYRESET
    bne     pause_loop
    jmp     end_game
.unpause
    ldy     #3
    lda     #$ff
.key_clear_loop
    sta     keypad_last_key,y
    dey
    bpl     .key_clear_loop
    rts
.restart
    ldx     #$ff                    ; Reset the stack.
    txs
    jmp     start_game

; Stop all CPU players at game over.
stop_cpu_players subroutine
    ldy     #player_data_2
    jsr     stop_player
    ldy     #player_data_3
    jsr     stop_player
    ldy     #player_data_4
    ; Fall through!

; Stop a player because the game has ended.
; Expects the player's data pointer in y.
stop_player subroutine
    lda     player_health,y         ; Do nothing on dead enemies.
    beq     .done
    lda     #0
    sta     player_x_speed,y
    sta     player_y_speed,y
    sta     player_frame_index,y
    sta     player_table_end,y
.done
    rts

; Start the game over process.
start_game_over subroutine
    lda     #60                 ; Wait a second before starting music
    sta     game_over_delay
    lda     player_count
    cmp     #1
    bne     .game_over_common       ; Battle mode.
    lda     recognizer_state
    bne     .game_over_common
    jsr     stop_cpu_players
    lda     player_data_1 + player_health
    beq     .disc_death
    lda     #0
    sta     player_data_1 + player_x_speed
    sta     player_data_1 + player_y_speed
    lda     #2                      ; Suppress animation.
    sta     player_data_1 + player_frame_tick
    jmp     .game_over_common
.disc_death
    ldy     #player_data_1
    jsr     start_death_animation
.game_over_common
    lda     #4                      ; Fours rows of "game over" tiles.
    sta     $ff
    mwa     $2159,$fc               ; Target address.
    ldx     #0                      ; Read index.
.game_over_text_outer_loop
    ldy     #0                      ; Write index.
.game_over_text_inner_loop
    lda     game_over_layout,x
    ora     #$80                    ; Activate alternate color.
    sta     ($fc),y
    inx
    iny
    cpy     #30                     ; 30 tiles per row.
    bmi     .game_over_text_inner_loop
    cpx     #120                    ; 120 total tiles.
    beq     .game_over_loop
    lda     $fc                     ; Next row.
    clc
    adc     #40
    sta     $fc
    bcc     .game_over_text_outer_loop
    inc     $fd
    jmp     .game_over_text_outer_loop
.game_over_loop
    jsr     wait_for_vb
    jsr     keypad_process
    lda     game_over_delay
    bmi     .music_done             ; Music already started.
    dec     game_over_delay
    bpl     .music_done             ; Not time to start.
    mva     #$65,AUDCTL
    mva     #2,SKCTL
    lda     #$1a
    jsr     DOTUNE
    lda     #$1b
    jsr     DOTUNE
    lda     #$1c
    jsr     DOTUNE
    lda     #$1d
    jsr     DOTUNE
    jsr     STARTTUN
.music_done
    ldy     player_count
.reset_check_loop
    dey
    bmi     .reset_loop_done
    lda     keypad_last_key,y
    cmp     #KEYRESET
    beq     end_game
    jmp     .reset_check_loop
.reset_loop_done
    jsr     process_all_players
    lda     player_count
    cmp     #1
    bne     .battle_mode_game_over
    jsr     process_player_1_disc
    lda     player_data_1 + player_health
    beq     .game_over_loop         ; Dead player, ignore.
    inc     sCOLPM0                 ; Frozen by guard, cycle colors.
    lda     #2                      ; Suppress animation.
    sta     player_data_1 + player_frame_tick
    jmp     .game_over_loop
.battle_mode_game_over
    jsr     process_all_player_discs
    jmp     .game_over_loop

; End the game, updating the high/low/last scores if necessary.
end_game subroutine
    lda     player_count            ; No score update on battle mode.
    cmp     #1
    bne     .return_to_title
    ldy     #0                      ; Check for a new high score.
    ldx     #0                      ; Assume no high score.
    stx     $ff
    stx     $fe                     ; Assume no low score.
    lda     high_score_zone + 19    ; If this isn't 0, no score yet.
    cmp     #$10
    bne     .new_high_score
.high_score_check_loop
    lda     $2000,y
    cmp     high_score_zone,y
    beq     .next_high_digit
    bmi     .high_score_check_done
    jmp     .new_high_score
.next_high_digit
    iny
    cpy     #20
    bne     .high_score_check_loop
    jmp     .high_score_check_done
.new_high_score
    dec     $ff
.high_score_check_done
    lda     low_score_zone + 19     ; Again, look for 0.
    cmp     #$10
    bne     .new_low_score
    ldy     #0                      ; Check for a new low score.
.low_score_check_loop
    lda     $2000,y
    cmp     low_score_zone,y
    beq     .next_low_digit
    bpl     .low_score_check_done
    jmp     .new_low_score
.next_low_digit
    iny
    cpy     #20
    bne     .low_score_check_loop
    jmp     .low_score_check_done
.new_low_score
    dec     $fe
.low_score_check_done
    ldy     #19                     ; Update the last, high, and low scores.
.score_update_loop
    lda     $2000,y
    sta     last_score_zone,y
    bit     $ff
    bpl     .no_high_score
    sta     high_score_zone,y
.no_high_score
    bit     $fe
    bpl     .no_low_score
    sta     low_score_zone,y
.no_low_score
    dey
    bpl     .score_update_loop
.return_to_title
    ldx     #$ff                    ; Reset the stack, might be in a jsr.
    txs
    jmp     start_title_screen

; Initialize a player data block with
; reasonable values.  Pass the block
; in y.
init_player_data subroutine
    lda     #subpixel_limit / 2
    sta     player_sub_x,y
    sta     player_sub_y,y
    sta     player_disc_sub_x,y
    sta     player_disc_sub_y,y
    lda     #disc_in_hand
    sta     player_disc_state,y
    lda     #player_right
    sta     player_facing,y
    lda     #3
    sta     player_health,y
    sta     player_max_health,y
    lda     #player_start_speed_index
    sta     player_speed_index,y
    lda     #5
    sta     player_last_angle,y
    lda     #<player_vertical_cycle
    sta     player_frame_table,y
    lda     #>player_vertical_cycle
    sta     player_frame_table + 1,y
    lda     #0
    sta     player_frame_index,y
    lda     #player_idle_last_frame
    sta     player_table_end,y
    lda     #player_frame_time
    sta     player_frame_tick,y
    rts

; Get the number of living enemies.
; Returns the count in the x register.
; 0 bit will be set on return.
count_living_enemies subroutine
    ldx     #3                      ; Assume 3.
    lda     player_data_2 + player_health
    bne     .player_2_alive
    dex
.player_2_alive
    lda     player_data_3 + player_health
    bne     .player_3_alive
    dex
.player_3_alive
    lda     player_data_4 + player_health
    bne     .done
    dex
.done
    rts

; Get the number of living discs.
; Returns the result in the x register.
; 0 bit will be set on return.
count_living_discs subroutine
    ldx     #3                      ; Assume 3.
    lda     #disc_dead
    cmp     player_data_2 + player_disc_state
    bne     .disc_2_alive
    dex
.disc_2_alive
    cmp     player_data_3 + player_disc_state
    bne     .disc_3_alive
    dex
.disc_3_alive
    cmp     player_data_4 + player_disc_state
    bne     .done
    dex
.done
    rts

; Macro for player movement processing.
; Arguments:
;   {1} The player's data block.
;   {2} The player's HPOS register.
;   {3} The player's sprite data page.
;   {4} The player's disc erasure pattern.
;   {5} The player's missile HPOS register.
;   {6} The player's disc write pattern.
;   {7} Offset to the player's color registers.
    mac process_player_movement
    lda     {1} + player_xpos       ; See if the player is alive.
    bne     .player_alive
    jmp     .done_player_move
.player_alive
    lda     {1} + player_health     ; Skip recovery on 0 health.
    beq     .health_done
    dec     {1} + player_recover_timer
    bne     .health_done            ; Timer still running.
    ldy     #{1}
    jsr     recover_one_health
.health_done
    lda     {1} + player_disc_state ; See if the player is blocking.
    cmp     #disc_blocking
    bne     .player_not_blocking
    ldx     {1} + player_ypos
    jmp     .refresh_player_frame
.player_not_blocking
    lda     {1} + player_x_speed        
    bmi     .move_left              ; Check the horizontal movement.
    bne     .move_right
    lda     {1} + player_y_speed    ; Check for no vertical movement
    bne     .not_idle
    ldx     {1} + player_ypos
    jmp     .refresh_player_frame
.not_idle
    jmp     .check_vertical
.move_right
    clc
    adc     {1} + player_sub_x
    cmp     #subpixel_limit + 1
    bpl     .move_coarse_right
    jmp     .update_x_subpixel
.move_coarse_right
    sec
    sbc     #subpixel_limit
    ldx     {1} + player_xpos
    inx
    cpx     #right_boundary
    bne     .right_boundary_ok
    pha                             ; Save the subpixel.
    bit     {1} + player_cpu_flags  ; Is this a CPU player?
    bpl     .human_right_boundary
    ldy     #{1}
    jsr     check_right_wall_bounce
    pla                             ; Retrieve the subpixel.
    jmp     .update_x_subpixel
.human_right_boundary
    ldx     #1
    ldy     {1} + player_ypos
    jsr     check_left_right_door_wrap
    beq     .do_right_wrap
    pla                             ; Retreive the subpixel.
    jmp     .update_x_subpixel
.do_right_wrap
    jsr     summon_recognizer
    ldy     #{1}
    jsr     recover_one_health
    lda     {1} + player_disc_state
    bne     .no_right_disc_wrap
    ldx     #left_boundary - 1
    stx     {1} + player_disc_xpos
.no_right_disc_wrap
    ldx     #left_boundary
    pla                             ; Retrieve the subpixel.
.right_boundary_ok
    stx     {1} + player_xpos
    stx     {2}
    ldx     {1} + player_disc_state
    if {1} != #player_data_1
    beq     .move_held_disc_right
    cpx     #disc_stick
    endif
    bne     .update_x_subpixel
.move_held_disc_right
    inc     {1} + player_disc_xpos
    jmp     .update_x_subpixel
.move_left
    clc
    adc     {1} + player_sub_x
    bmi     .move_coarse_left
    jmp     .update_x_subpixel
.move_coarse_left
    clc
    adc     #subpixel_limit
    ldx     {1} + player_xpos
    dex
    cpx     #left_boundary
    bne     .left_boundary_ok
    pha                             ; Save the subpixel.
    bit     {1} + player_cpu_flags  ; Is this a CPU player?
    bpl     .human_left_boundary
    ldy     #{1}
    jsr     check_left_wall_bounce
    pla                             ; Retrieve the subpixel.
    jmp     .update_x_subpixel
.human_left_boundary
    ldx     #0
    ldy     {1} + player_ypos
    jsr     check_left_right_door_wrap
    beq     .do_left_wrap
    pla                             ; Retrieve the subpixel.
    jmp     .update_x_subpixel
.do_left_wrap
    jsr     summon_recognizer
    ldy     #{1}
    jsr     recover_one_health
    lda     {1} + player_disc_state
    bne     .no_left_disc_wrap
    ldx     #right_boundary + 1
    stx     {1} + player_disc_xpos
.no_left_disc_wrap
    ldx     #right_boundary
    pla                             ; Retrieve the subpixel.
.left_boundary_ok
    stx     {1} + player_xpos
    stx     {2}
    ldx     {1} + player_disc_state
    if {1} != #player_data_1
    beq     .move_held_disc_left
    cpx     #disc_stick
    endif
    bne     .update_x_subpixel
.move_held_disc_left
    dec     {1} + player_disc_xpos
.update_x_subpixel
    sta     {1} + player_sub_x
.check_vertical
    ldx     {1} + player_ypos
    lda     {1} + player_y_speed
    bmi     .move_up                ; Check the vertical movement.
    bne     .move_down
    jmp     .refresh_player_frame
.move_down
    clc
    adc     {1} + player_sub_y
    cmp     #subpixel_limit + 1
    bpl     .y_subpixel_positive
    jmp     .update_y_subpixel
.y_subpixel_positive
    sec
    sbc     #subpixel_limit
    cpx     #bottom_boundary
    bne     .bottom_boundary_ok
    pha                             ; Save the subpixel.
    bit     {1} + player_cpu_flags  ; Is this a CPU player?
    bpl     .human_bottom_boundary
    ldy     #{1}
    jsr     check_bottom_wall_bounce
    pla                             ; Retrieve the subpixel.
    jmp     .update_y_subpixel
.human_bottom_boundary
    ldy     #1
    ldx     {1} + player_xpos
    jsr     check_top_bottom_door_wrap
    beq     .wrap_to_top
    ldx     {1} + player_ypos
    pla                             ; Retrieve the subpixel.
    jmp     .update_y_subpixel
.wrap_to_top
    jsr     .erase_player
    jsr     summon_recognizer
    ldy     #{1}
    jsr     recover_one_health
    lda     {1} + player_disc_state
    bne     .no_bottom_disc_wrap
    ldy     {1} + player_disc_ypos
    lda     #{4}
    jsr     erase_disc
    lda     #top_boundary + #held_disc_y_delta
    sta     {1} + player_disc_ypos
.no_bottom_disc_wrap
    pla                             ; Retrieve the subpixel.
    ldx     #top_boundary
    jmp     .update_y_subpixel
.bottom_boundary_ok
    inx
    ldy     {1} + player_disc_state
    beq     .move_disc_down
    cpy     #disc_stick
    bne     .update_y_subpixel
.move_disc_down
    inc     {1} + player_disc_ypos
    jmp     .update_y_subpixel
.move_up
    clc
    adc     {1} + player_sub_y
    bpl     .update_y_subpixel
    clc
    adc     #subpixel_limit
    cpx     #top_boundary
    bne     .top_boundary_ok
    pha                             ; Save the subpixel.
    bit     {1} + player_cpu_flags  ; Is this a CPU player?
    bpl     .human_top_boundary
    ldy     #{1}
    jsr     check_top_wall_bounce
    pla                             ; Retrieve the subpixel.
    jmp     .update_y_subpixel
.human_top_boundary
    ldy     #0
    ldx     {1} + player_xpos
    jsr     check_top_bottom_door_wrap
    beq     .wrap_to_bottom
    ldx     {1} + player_ypos
    pla                             ; Retrieve the subpixel.
    jmp     .update_y_subpixel
.wrap_to_bottom
    jsr     .erase_player
    jsr     summon_recognizer
    ldy     #{1}
    jsr     recover_one_health
    lda     {1} + player_disc_state
    bne     .no_top_disc_wrap
    ldy     {1} + player_disc_ypos
    lda     #{4}
    jsr     erase_disc
    lda     #bottom_boundary + #held_disc_y_delta
    sta     {1} + player_disc_ypos
.no_top_disc_wrap
    pla                             ; Retreive the subpixel.
    ldx     #bottom_boundary
    jmp     .update_y_subpixel
.top_boundary_ok
    dex
    ldy     {1} + player_disc_state
    beq     .move_disc_up
    cpy     #disc_stick
    bne     .update_y_subpixel
.move_disc_up
    dec     {1} + player_disc_ypos
.update_y_subpixel
    sta     {1} + player_sub_y
    stx     {1} + player_ypos
.refresh_player_frame
    stx     $fe
    lda     #>{3}
    sta     $ff
    dec     {1} + player_frame_tick ; Check for frame change.
    beq     .new_frame
    ldy     {1} + player_frame_index
    jmp     .draw_frame
.new_frame
    lda     #player_frame_time
    sta     {1} + player_frame_tick
    ldy     {1} + player_frame_index
    cpy     {1} + player_table_end  ; Check for animation loop.
    bne     .no_animation_loop
    lda     {1} + player_health     ; Is this the death animation?
    bne     .not_dead_yet
    sta     {1} + player_xpos       ; Move off screen.
    sta     {2}
    lda     {1} + player_disc_state ; If the disc is in flight, leave it.
    cmp     #disc_in_flight
    beq     .in_flight_dont_erase
    lda     #0
    sta     {1} + player_disc_xpos
    sta     {5}
    lda     #disc_dead
    sta     {1} + player_disc_state
.in_flight_dont_erase
    jsr     .erase_player
    jmp     .done_player_move
.not_dead_yet
    ldy     #0                      ; Go back to start.
    .byte   $2c
.no_animation_loop
    iny                             ; Go to next frame.
    iny
.draw_frame
    sty     {1} + player_frame_index
    lda     ({1} + player_frame_table),y
    sta     $fc
    iny
    lda     ({1} + player_frame_table),y
    sta     $fd
    ldy     #player_sprite_height
    jsr     copy_memory
    jmp     .done_player_move
.erase_player
    ldy     {1} + player_ypos
    sta     $fe
    ldx     #player_sprite_height
    lda     #0
.erase_player_loop
    sta     {3},y
    dex
    beq     .erase_player_return
    iny
    jmp     .erase_player_loop
.erase_player_return
    rts
.done_player_move
.handle_disc_movement subroutine
    lda     {1} + player_health     ; Only handle this if the player is
    bne     .player_alive_disc_ok   ; alive, or the disc is in flight.
    lda     {1} + player_disc_state
    cmp     #disc_in_flight
    beq     .dead_and_inflight
    cmp     #disc_shatter
    beq     .dead_and_shatter
    jmp     .end_handle_disc
.dead_and_inflight
    jmp     .handle_in_flight
.dead_and_shatter
    jmp     .handle_shatter
.player_alive_disc_ok
    lda     {1} + player_disc_state
    bne     .not_held
    jmp     .handle_held
.not_held
    cmp     #disc_crouch
    bne     .not_crouch
    jmp     .handle_held
.not_crouch
    cmp     #disc_in_flight
    bne     .not_in_flight
    jmp     .handle_in_flight
.not_in_flight
    cmp     #disc_returning
    bne     .check_blocking
    jmp     .handle_returning
.check_blocking
    cmp     #disc_blocking
    bne     .not_blocking
    jmp     .handle_blocking
.not_blocking
    cmp     #disc_shatter
    bne     .not_shatter
    jmp     .handle_shatter
.not_shatter
    cmp     #disc_destroyed
    beq     .handle_destroyed
    cmp     #disc_floppy
    beq     .handle_floppy
    if {1} != #player_data_1        ; Player doesn't get a stick.
    cmp     #disc_stick
    beq     .handle_stick
    endif
    jmp     .end_handle_disc
    if {1} != #player_data_1        ; Player doesn't get a stick.
.handle_stick
    lda     {1} + player_disc_xpos
    sta     {5}
    lda     #$13
    sta     $ff
    ldy     {1} + player_disc_ypos
    dey
    sty     $fe
    lda     #{4}                    ; Erase one pixel above,
    ldy     #0                      ; and one below.
    and     ($fe),y
    sta     ($fe),y
    ldy     #8
    lda     #{4}
    and     ($fe),y
    sta     ($fe),y
    ldy     #7                      ; Fill in the middle.
.draw_stick_loop
    lda     [#{6} << 1] & #{6}      ; Shift to one pixel.
    ora     ($fe),y
    sta     ($fe),y
    dey
    bne     .draw_stick_loop
    jmp     .end_handle_disc
    endif
.handle_destroyed
    dec     {1} + player_disc_timer
    bne     .not_respawn
    jmp     .respawn_disc
.not_respawn
    jmp     .end_handle_disc
.handle_floppy
    if {1} == #player_data_1        ; Only the human can reach this.
    dec     {1} + player_disc_timer
    bne     .continue_floppy
    jmp     .start_return
.continue_floppy
    lda     player_data_1 + player_disc_sub_x
    clc
    adc     player_data_1 + player_disc_x_speed
    bmi     .floppy_move_left
    cmp     #subpixel_limit
    bmi     .floppy_update_x_subpixel
    sec
    sbc     #subpixel_limit
    ldy     player_data_1 + player_disc_xpos
    iny
    jmp     .floppy_update_mpos
.floppy_move_left
    clc
    adc     #subpixel_limit
    ldy     player_data_1 + player_disc_xpos
    dey
.floppy_update_mpos
    sty     player_data_1 + player_disc_xpos
    sty     HPOSM0
.floppy_update_x_subpixel
    sta     player_data_1 + player_disc_sub_x
    lda     player_data_1 + player_disc_sub_y
    clc
    adc     player_data_1 + player_disc_y_speed
    bmi     .floppy_move_up
    cmp     #subpixel_limit
    bmi     .floppy_update_y_subpixel
    inc     player_data_1 + player_disc_ypos
    sec
    sbc     #subpixel_limit
    jmp     .floppy_update_y_subpixel
.floppy_move_up
    dec     player_data_1 + player_disc_ypos
    clc
    adc     #subpixel_limit
.floppy_update_y_subpixel
    sta     player_data_1 + player_disc_sub_y
    ldy     player_data_1 + player_disc_ypos
    lda     $12ff,y                 ; Blank one row high and low to
    and     #$fc                    ; cover the animations.
    sta     $12ff,y
    lda     $1302,y
    and     #$fc
    sta     $1302,y
    lda     {1} + player_disc_timer
    and     #$08                    ; Animate based on the 8 bit.
    beq     .floppy_frame_0         ; Only player 1 can hit this state.
    lda     $1300,y
    and     #$fc
    ora     #$01
    sta     $1300,y
    lda     $1301,y
    and     #$fc
    ora     #$02
    sta     $1301,y
    jmp     .end_handle_disc
.floppy_frame_0
    lda     $1300,y
    and     #$fc
    ora     #$02
    sta     $1300,y
    lda     $1301,y
    and     #$fc
    ora     #$01
    sta     $1301,y
    jmp     .end_handle_disc
    endif
.respawn_disc
    lda     #disc_returning
    sta     {1} + player_disc_state
    lda     #50
    sta     {1} + player_disc_xpos
    sta     {5}
    lda     #$10
    sta     {1} + player_disc_ypos
    lda     {1} + player_cpu_flags  ; Clear the special disc flags.
    and     #cpu_double_flag | #cpu_seeker_flag
    beq     .disc_reset_done        ; Not a leader, ignore.
    lda     {1} + player_cpu_flags
    and     #~[#cpu_double_flag | #cpu_seeker_flag]
    sta     {1} + player_cpu_flags
    lda     #leader_color
    sta     {7} + COLPM0
    sta     {7} + sCOLPM0
.disc_reset_done
    jmp     .end_handle_disc
.handle_shatter
    dec     {1} + player_disc_timer
    beq     .next_shatter_frame
    jmp     .end_handle_disc
.next_shatter_frame
    lda     #shatter_frame_time
    sta     {1} + player_disc_timer
    ldy     {1} + player_disc_ypos
    lda     {1} + player_disc_frame
    beq     .shatter_frame_1
    cmp     #1
    beq     .shatter_frame_2
    lda     #{4}
    jsr     erase_disc
    lda     {1} + player_health     ; If player is dead, kill disc.
    bne     .shatter_not_dead
    lda     #disc_dead
    sta     {1} + player_disc_state
    jmp     .end_handle_disc
.shatter_not_dead
    lda     #disc_destroyed
    sta     {1} + player_disc_state
    lda     #disc_respawn_time
    sta     {1} + player_disc_timer
    jmp     .end_handle_disc
.shatter_frame_2
    inc     {1} + player_disc_frame
    lda     $1300,y
    and     #{4}
    ora     #$aa & #{6}
    sta     $1300,y
    lda     $1301,y
    and     #{4}
    sta     $1301,y
    lda     $1302,y
    and     #{4}
    sta     $1302,y
    lda     #$55 & #{6}
    ora     $1303,y
    sta     $1303,y
    ldx     {1} + player_disc_xpos
    lda     {1} + player_disc_x_speed
    bmi     .shatter_to_right
    dex
    dex
    .byte   $2c
.shatter_to_right
    inx
    inx
    stx     {1} + player_disc_xpos
    stx     {5}
    jmp     .end_handle_disc
.shatter_frame_1
    inc     {1} + player_disc_frame
    lda     $1300,y
    ora     #$aa & #{6}
    sta     $1300,y
    lda     $1301,y
    and     #{4}
    ora     #$55 & #{6}
    sta     $1301,y
    jmp     .end_handle_disc
.handle_blocking
    lda     #{6}
    ldy     #{1}
    jsr     position_block_disc
    lda     {1} + player_disc_xpos
    sta     {5}
    jmp     .end_handle_disc
.handle_held
    lda     {1} + player_xpos       ; See if the player is alive.
    bne     .player_alive_held
    jmp     .end_handle_disc
.player_alive_held
    lda     {1} + player_disc_xpos
    sta     {5}
    ldx     {1} + player_disc_ypos
    dex
    dex
    stx     $fe
.draw_square_disc
    lda     #$13
    sta     $ff
    ldy     #0
    lda     #{4}
    and     ($fe),y
    sta     ($fe),y
    iny
    lda     #{4}
    and     ($fe),y
    sta     ($fe),y
    ldy     #6
    lda     #{4}
    and     ($fe),y
    sta     ($fe),y
    ldy     #5
    lda     #{4}
    and     ($fe),y
    sta     ($fe),y
.held_vertical_loop
    dey
    cpy     #1
    beq     .held_vertical_loop_done
    lda     #{6}
    ora     ($fe),y
    sta     ($fe),y
    jmp     .held_vertical_loop
.held_vertical_loop_done
    jmp     .end_handle_disc
.handle_in_flight
    lda     {1} + player_cpu_flags  ; Is this a CPU player?
    bmi     .check_seeker_redirect
    lda     {1} + player_disc_action
    cmp     #action_return
    bne     .no_flight_cancel
    jmp     .start_return
.check_seeker_redirect
    and     #cpu_seeker_flag
    beq     .no_flight_cancel
    lda     cpu_react_timer         ; Re-aim every second.
    cmp     #120
    beq     .reaim_disc
    cmp     #60
    bne     .no_flight_cancel
.reaim_disc
    lda     {1} + player_disc_xpos
    sta     $fe
    ldy     {1} + player_disc_ypos
    iny
    iny
    sty     $ff
    lda     #{4}
    sta     $f0
    ldy     #{1}
    jsr     cpu_aim_disc
.no_flight_cancel
    jsr     .horizontal_move_common
    cpx     #disc_right_boundary
    bcc     .check_left_boundary
    bit     {1} + player_cpu_flags  ; Is this a CPU player?
    bmi     .right_wall_cpu
    ldy     {1} + player_disc_ypos
    jsr     check_disc_right_wall
.right_wall_cpu
    jmp     .start_return
.check_left_boundary
    cpx     #disc_left_boundary + 1
    bcs     .update_x_position
    bit     {1} + player_cpu_flags  ; Is this a CPU player?
    bmi     .left_wall_cpu
    ldy     {1} + player_disc_ypos
    jsr     check_disc_left_wall
.left_wall_cpu
    jmp     .start_return
.update_x_position
    stx     {1} + player_disc_xpos
    stx     {5}
    ldy     {1} + player_disc_ypos
    sty     $fe
    ldy     #$13
    sty     $ff
    lda     {1} + player_disc_sub_y
    clc
    adc     {1} + player_disc_y_speed
    bmi     .in_flight_up
    cmp     #subpixel_limit
    bmi     .update_y_subpixel
    sec
    sbc     #subpixel_limit
    tax
    ldy     {1} + player_disc_ypos
    iny
    iny
    cpy     #disc_bottom_boundary
    bcc     .in_flight_move_down
    bit     {1} + player_cpu_flags  ; Is this a CPU player?
    bmi     .bottom_wall_cpu
    ldx     {1} + player_disc_xpos
    jsr     check_disc_bottom_wall
.bottom_wall_cpu
    jmp     .start_return
.in_flight_move_down
    sty     {1} + player_disc_ypos
    lda     #{6}
    ldy     #3
    ora     ($fe),y
    sta     ($fe),y
    jmp     .finish_vmove_draw
.in_flight_up
    clc
    adc     #subpixel_limit
    tax
    ldy     {1} + player_disc_ypos
    dey
    dey
    cpy     #disc_top_boundary + 1
    bcs     .in_flight_move_up
    bit     {1} + player_cpu_flags  ; Is this a CPU player?
    bmi     .top_wall_cpu
    ldx     {1} + player_disc_xpos
    jsr     check_disc_top_wall
.top_wall_cpu
    jmp     .start_return
.in_flight_move_up
    sty     {1} + player_disc_ypos
    lda     #{6}
    ldy     #0
    dec     $fe
    ora     ($fe),y
    sta     ($fe),y
    inc     $fe
.finish_vmove_draw
    ldy     #1
    lda     #{4}
    and     ($fe),y
    sta     ($fe),y
    txa
.update_y_subpixel
    sta     {1} + player_disc_sub_y
    jmp     .end_handle_disc
.start_return
    lda     {1} + player_health     ; See if the player is removed.
    bne     .start_return_proper
    lda     #disc_dead              ; Erase the disc.
    sta     {1} + player_disc_state
    lda     #{4}
    ldy     {1} + player_disc_ypos
    jsr     erase_disc
    jmp     .end_handle_disc
.start_return_proper
    lda     #disc_returning
    sta     {1} + player_disc_state
    lda     #$13
    sta     $ff
    lda     {1} + player_disc_ypos
    sta     $fe
    ldy     #0
    lda     #{6}
    ora     ($fe),y
    sta     ($fe),y
    ldy     #2
    lda     #{6}
    ora     ($fe),y
    sta     ($fe),y
    jmp     .end_handle_disc
.handle_returning
    lda     {1} + player_xpos
    bne     .player_alive_returning
    jmp     .end_handle_disc
.player_alive_returning
    lda     {1} + player_disc_x_speed
    bne     .not_caught             ; When both speeds are 0, catch.
    lda     {1} + player_disc_y_speed
    bne     .not_caught
    lda     {1} + player_table_end
    bne     .no_stand_up
    lda     {1} + player_frame_table
    cmp     #<player_vertical_cycle
    bne     .stand_up_on_catch
    lda     {1} + player_frame_table + 1
    cmp     #>player_vertical_cycle
    beq     .no_stand_up
.stand_up_on_catch
    lda     {1} + player_facing
    beq     .stand_up_left
    lda     player_disc_offsets + 3
    sta     {1} + player_disc_offset
    lda     #<player_right_cycle
    sta     {1} + player_frame_table
    lda     #>player_right_cycle
    jmp     .stand_up_common
.stand_up_left
    lda     player_disc_offsets + 13
    sta     {1} + player_disc_offset
    lda     #<player_left_cycle
    sta     {1} + player_frame_table
    lda     #>player_left_cycle
.stand_up_common
    sta     {1} + player_frame_table + 1
.no_stand_up
    lda     #disc_in_hand
    sta     {1} + player_disc_state
    lda     {1} + player_disc_ypos
    sta     $fe
    lda     {1} + player_ypos
    clc
    adc     #held_disc_y_delta
    sta     {1} + player_disc_ypos
    sta     $fc
    lda     {1} + player_xpos
    clc
    adc     {1} + player_disc_offset
    sta     {1} + player_disc_xpos
    sta     {5}
    lda     #$13
    sta     $ff
    sta     $fd
    ldy     #2
.caught_draw_erase_loop
    lda     ($fe),y
    and     #{4}
    sta     ($fe),y
    lda     ($fc),y
    ora     #{6}
    sta     ($fc),y
    dey
    bpl     .caught_draw_erase_loop
    jmp     .end_handle_disc
.not_caught
    lda     {1} + player_disc_xpos
    sec
    sbc     #3
    cmp     {1} + player_xpos
    bcc     .return_to_right
    sbc     #2
    cmp     {1} + player_xpos
    bcs     .return_to_left
    lda     #0
    .byte   $2c
.return_to_right
    lda     #disc_return_speed
    .byte   $2c
.return_to_left
    lda     -#disc_return_speed
    sta     {1} + player_disc_x_speed
    jsr     .horizontal_move_common
    lda     {1} +  player_disc_ypos
    sec
    sbc     #player_sprite_height / 2 - 4
    cmp     {1} + player_ypos
    bcc     .return_to_bottom
    sbc     #3
    cmp     {1} + player_ypos
    bcs     .return_to_top
    lda     #0
    .byte   $2c
.return_to_bottom
    lda     #disc_return_speed
    .byte   $2c
.return_to_top
    lda     -#disc_return_speed
    stx     {1} + player_disc_xpos
    stx     {5}
    sta     {1} + player_disc_y_speed
    ldy     {1} + player_disc_ypos
    clc
    adc     {1} + player_disc_sub_y
    bmi     .return_move_up
    cmp     #subpixel_limit
    bmi     .return_update_y_subpixel
    sec
    sbc     #subpixel_limit
    iny
    iny
    jmp     .return_update_y_subpixel
.return_move_up
    clc
    adc     #subpixel_limit
    dey
    dey
.return_update_y_subpixel
    sta     {1} + player_disc_sub_y
    sty     {1} + player_disc_ypos
    dey
    dey
    sty     $fe
    jmp     .draw_square_disc
.horizontal_move_common
    ldx     {1} + player_disc_xpos
    lda     {1} + player_disc_sub_x
    clc
    adc     {1} + player_disc_x_speed
    bmi     .common_move_left
    cmp     #subpixel_limit
    bmi     .return_update_x_subpixel
    sec
    sbc     #subpixel_limit
    inx
    inx
    jmp     .return_update_x_subpixel
.common_move_left
    clc
    adc     #subpixel_limit
    dex
    dex
.return_update_x_subpixel
    sta     {1} + player_disc_sub_x
    rts
.end_handle_disc
    endm

; Process movement for all four players.
process_all_players subroutine
    process_player_movement player_data_2,HPOSP1,$1500,$f3,HPOSM1,$0c,1
    process_player_movement player_data_3,HPOSP2,$1600,$cf,HPOSM2,$30,2
    process_player_movement player_data_4,HPOSP3,$1700,$3f,HPOSM3,$c0,3
    ; Fall through into player 1.

; Process player 1 alone.  This is used for recognizer
; battles where the other player sprites are being used
; for recognizer type stuff.
process_player_1 subroutine
    process_player_movement player_data_1,HPOSP0,$1400,$fc,HPOSM0,$03,0
    rts

; This macro handles human player input.
; Arguments:
;   {1} The player's data block.
;   {2} The player's disc input location.
;   {3} The player's disc erasure mask.
;   {4} The player's horizontal POT shadow.
;   {5} The player's vertical POT shadow.
;   {6} The player's missile position register.
;   {7} The player's block draw mask.
    mac process_player_input
    jsr     {2}
    ldx     {1} + player_disc_action
    bmi     .no_throw
    cpx     #action_block
    bpl     .no_throw
    lda     {1} + player_disc_state
    beq     .throw_ok
    cmp     #disc_blocking
    beq     .throw_ok
    cmp     #disc_crouch
    bne     .no_throw
.throw_ok
    ldy     #{1}
    lda     #{3}
    jsr     throw_disc
.no_throw
    ldx     {4}                     ; Check player movement.
    ldy     {5}
    jsr     check_joystick_pots
    beq     .no_last_angle_update
    sta     {1} + player_last_angle
.no_last_angle_update
    ldx     {1} + player_hit_tag
    beq     .no_hit_movement_check
    cmp     {1} + player_last_move
    bne     .no_hit_movement_check  ; Direction changed, OK to move.
    lda     #0                      ; Convert to a null move
    jmp     .skip_last_move_update
.no_hit_movement_check
    sta     {1} + player_last_move
    ldx     #0
    stx     {1} + player_hit_tag
.skip_last_move_update
    pha                             ; Save the direction code.
    clc                             ; Convert to a lookup index.
    adc     {1} + player_speed_index
    tax
    lda     player_hspeed_table,x
    sta     {1} + player_x_speed
    lda     player_vspeed_table,x
    sta     {1} + player_y_speed
    pla                             ; Retrieve the direction code.
    tax
.block_crouch
    lda     {1} + player_disc_action
    cmp     #action_block
    beq     .start_block_crouch
    jmp     .no_block_crouch
.start_block_crouch
    lda     {1} + player_disc_state
    beq     .start_block
    cmp     #disc_crouch
    beq     .start_block
    cmp     #disc_blocking
    bne     .start_crouch
    lda     #disc_crouch
    sta     {1} + player_disc_state
    ldy     {1} + player_disc_ypos
    lda     #{3}
    jsr     erase_disc
    lda     #crouch_disc_y_delta
    clc
    adc     {1} + player_ypos
    sta     {1} + player_disc_ypos
.start_crouch
    lda     {1} + player_facing
    beq     .crouch_left
    lda     {1} + player_disc_state
    cmp     #disc_crouch
    bne     .no_crouch_right_disc_adjust
    lda     #crouch_right_offset
    clc
    adc     {1} + player_xpos
    sta     {1} + player_disc_xpos
.no_crouch_right_disc_adjust
    lda     #<player_right_crouch_cycle
    sta     {1} + player_frame_table
    lda     #>player_right_crouch_cycle
    jmp     .store_block_crouch
.crouch_left
    lda     {1} + player_disc_state
    cmp     #disc_crouch
    bne     .no_crouch_left_disc_adjust
    lda     #crouch_left_offset
    clc
    adc     {1} + player_xpos
    sta     {1} + player_disc_xpos
.no_crouch_left_disc_adjust
    lda     #<player_left_crouch_cycle
    sta     {1} + player_frame_table
    lda     #>player_left_crouch_cycle
    jmp     .store_block_crouch
.start_block
    ldy     {1} + player_disc_ypos
    lda     #{3}
    jsr     erase_disc
    lda     {1} + player_ypos
    clc
    adc     #4
    sta     {1} + player_disc_ypos
    tay
    ldx     #6
.draw_block_disc_loop
    lda     #{7}
    ora     $1300,y
    sta     $1300,y
    iny
    dex
    bne     .draw_block_disc_loop
    lda     #disc_blocking
    sta     {1} + player_disc_state
    lda     {1} + player_facing
    beq     .block_left
    lda     {1} + player_xpos
    clc
    adc     #7
    sta     {1} + player_disc_xpos
    sta     {6}
    lda     #<player_right_block_cycle
    sta     {1} + player_frame_table
    lda     #>player_right_block_cycle
    jmp     .store_block_crouch
.block_left
    lda     {1} + player_xpos
    sta     {1} + player_disc_xpos
    sta     {6}
    lda     #<player_left_block_cycle
    sta     {1} + player_frame_table
    lda     #>player_left_block_cycle
.store_block_crouch
    sta     {1} + player_frame_table + 1
    lda     #0
    sta     {1} + player_frame_index
    sta     {1} + player_table_end
.no_block_crouch
    lda     {1} + player_x_speed
    ora     {1} + player_y_speed
    beq     .no_crouch_undo
    lda     {1} + player_disc_state
    cmp     #disc_crouch
    bne     .no_crouch_undo
    ldy     {1} + player_disc_ypos
    lda     #{3}
    jsr     erase_disc
    lda     {1} + player_ypos
    clc
    adc     #held_disc_y_delta
    sta     {1} + player_disc_ypos
    lda     #disc_in_hand
    sta     {1} + player_disc_state
.no_crouch_undo
    ldy     #{1}
    jmp     update_player_animation
    endm

; Input processing for human players.
process_player_1_input subroutine
    process_player_input player_data_1,p0_disc_input,$fc,sPOT0,sPOT1,HPOSM0,$02

process_player_2_input subroutine
    process_player_input player_data_2,p1_disc_input,$f3,sPOT2,sPOT3,HPOSM1,$08

process_player_3_input subroutine
    process_player_input player_data_3,p2_disc_input,$cf,sPOT4,sPOT5,HPOSM2,$20

process_player_4_input subroutine
    process_player_input player_data_4,p3_disc_input,$3f,sPOT6,sPOT7,HPOSM3,$80

; Macro for disc-to-player collisions.
; Arguments:
;   {1} The player's data block.
;   {2} The player's missile collision register.
;   {3} The player's disc erasure mask.
;   {4} The first bit to check for hits.
;   {5} The player data block associated with {3}.
;   {6} The disc erasure mask assocaited with {3}.
;   {7 - 12} Repeat for the other players.
    mac process_player_disc
    lda     {1} + player_disc_state
    cmp     #disc_in_flight
    bne     .no_hit_check
    lda     {2}
    and     #$0f                    ; Mask out the high bits so the
    tax                             ; damage subroutine doesn't apply
    and     #{4}                    ; double damage.
    beq     .no_hit_p2
    ldy     #{5}
    lda     #{6}
    jsr     hit_player
    bne     .player_disc_return
.no_hit_p2
    txa
    and     #{7}
    beq     .no_hit_p3
    ldy     #{8}
    lda     #{9}
    jsr     hit_player
    bne     .player_disc_return
.no_hit_p3
    txa
    and     #{10}
    beq     .no_hit_check
    ldy     #{11}
    lda     #{12}
    jsr     hit_player
    beq     .no_hit_check
.player_disc_return
    ldy     #{1}
    ldx     #{3}
    jsr     handle_disc_after_player_hit
.no_hit_check
    endm

; Do disc processing for all players in battle mode.
process_all_player_discs
    process_player_disc player_data_2,M1PL,$f3,$01,player_data_1,$fc,$04,player_data_3,$cf,$08,player_data_4,$3f
    process_player_disc player_data_3,M2PL,$cf,$01,player_data_1,$fc,$02,player_data_2,$f3,$08,player_data_4,$3f
    process_player_disc player_data_4,M3PL,$3f,$01,player_data_1,$fc,$02,player_data_2,$f3,$04,player_data_3,$cf
    ; Fall through into player 1.

; Handle player 1's disc flight.  Factored out
; so the game over code can use it.
process_player_1_disc subroutine
    process_player_disc player_data_1,M0PL,$fc,$02,player_data_2,$f3,$04,player_data_3,$cf,$08,player_data_4,$3f
    rts

; See if a enemy should be processed for damage to
; the player.  Returns 'eq' if the enemy's disc
; is in flight, or if the enemy is a guard.  The
; enemy's data pointer should be passed in y.
enemy_is_attacking subroutine
    lda     player_disc_state,y     ; Check for disc in flight.
    cmp     #disc_in_flight
    beq     .return
    ; Fallthrough into the other test.

; Quick test for the guard status of an enemy.
; Pass the enemy's data pointer in y.  Result
; is in the 0 bit.
enemy_is_guard
    lda     player_max_health,y     ; Guards have 4 health.
    cmp     #4
.return
    rts

; Deal with fallout of a disc going to return
; mode after hitting a player.  Expects the disc
; owner's data block in y and the disc erasure
; pattern in x.
handle_disc_after_player_hit subroutine
    lda     player_health,y         ; Check for a dead thrower.
    bne     .normal_return
    lda     #disc_dead              ; Kill the disc.
    sta     player_disc_state,y
    lda     player_disc_ypos,y
    tay
    txa
    jmp     erase_disc
.normal_return
    lda     #disc_returning         ; Player is alive, just go
    sta     player_disc_state,y     ; to return mode.
    rts

; Start a disc throw.  Pass the disc throwing angle
; in x, the player's data block in y, and the player's
; disc erasure mask in the accumulator.
throw_disc subroutine
    sta     $f0                     ; Save the mask.
    sty     $f1                     ; Save the pointer.
    stx     $f2                     ; Save the angle.
    lda     #$18
    jsr     DOTUNE
    lda     #$19
    jsr     DOTUNE
    lda     $f0
    ldy     $f1
    ldx     $f2
    lda     player_disc_x_speeds,x
    sta     player_disc_x_speed,y
    lda     player_disc_y_speeds,x
    sta     player_disc_y_speed,y
    lda     player_disc_state,y     ; If already in flight, bail out.
    cmp     #disc_in_flight         ; The seeker disc re-uses this code.
    beq     .done
    lda     #disc_in_flight
    sta     player_disc_state,y
    lda     player_xpos,y           ; Center the hpos.
    clc
    adc     #3
    sta     player_disc_xpos,y
    lda     player_disc_ypos,y
    tay
    lda     $f0
    jsr     erase_disc
    ldy     $f1                     ; Retrieve the pointer.
    lda     player_ypos,y           ; Set the new ypos.
    clc
    adc     #4
    sta     player_disc_ypos,y
    tay
    lda     $f0                     ; Flip the mask.
    eor     #$ff
    ora     $1301,y
    sta     $1301,y
.done
    rts

; Erase the disc at offset y using the mask in
; the accumulator.
erase_disc subroutine
    ldx     #8                      ; Erase 8 rows.
    sta     $ff                     ; Save the mask.
.loop
    lda     $1300,y                 ; Get the byte.
    and     $ff                     ; Mask it out.
    sta     $1300,y                 ; Write it back.
    iny
    dex
    bne     .loop
    rts

; Check for discs intercepted by a blocking player.
; Expects the blocking player's disc xpos in x,
; ypos in y.
check_block subroutine
    stx     $fc                     ; Save the position.
    dey                             ; Adjust for easier comparison.
    sty     $fd                     ; add_to_score clobbers $fe-$ff.
    ldy     #player_data_1          ; Look for in-flight discs.
.player_loop
    lda     player_disc_state,y
    cmp     #disc_in_flight         ; Disregard other states.
    bne     .next_player
    lda     player_disc_xpos,y      ; Check the xpos.
    sec
    sbc     $fc
    beq     .xhit
    cmp     #1
    beq     .xhit
    cmp     #-1
    bne     .next_player
.xhit
    lda     player_disc_ypos,y      ; Check the ypos.
    cmp     $fd
    bcc     .next_player
    sbc     $fd
    cmp     #6
    bcs     .next_player
.yhit
    tya
    pha
    lda     #8
    jsr     DOTUNE
    lda     #9
    jsr     DOTUNE
    pla
    tay
    lda     #disc_shatter
    sta     player_disc_state,y
    lda     #1
    sta     player_disc_timer,y
    lda     #0
    sta     player_disc_frame,y
    ldx     #$12
    lda     player_count            ; No score in battle mode.
    cmp     #1
    bne     .next_player
    sty     $fb                     ; Save y.
    jsr     add_to_score
    ldy     $fb                     ; Recover y.
.next_player
    cpy     #player_data_4
    beq     .done
    tya
    clc
    adc     #player_size
    tay
    jmp     .player_loop
.done
    rts

; Position the disc while in blocking mode.
; Expects the player's data block in y and
; the player's missile mask in the accumulator.
position_block_disc subroutine
    sta     $fe                     ; Save the mask.
    lda     player_x_speed,y        ; Figure out the direction.
    ora     player_y_speed,y
    bne     .directions_ok
    rts                             ; Do nothing on neutral.
.directions_ok
    tya                             ; Save the block.
    pha
    lda     player_disc_ypos,y
    tay
    lda     $fe
    eor     #$ff
    jsr     erase_disc
    pla
    tay
    lda     player_x_speed,y
    bmi     .speed_left
    beq     .hspeed_neutral
    lda     #1                      ; Right flag.
    .byte   $2c
.speed_left
    lda     #-1                     ; Left flag.
    .byte   $2c
.hspeed_neutral
    lda     #0                      ; Neutral flag.
    sta     $fd                     ; Save the hspeed.
    lda     player_y_speed,y
    bmi     .speed_up
    beq     .vspeed_neutral
    lda     $fd
    bmi     .disc_down_left
    lda     $fe                     ; Mask to the left bit.
    asl
    and     $fe
    sta     $fe
    lda     player_xpos,y
    clc
    adc     #7
    sta     player_disc_xpos,y
    jsr     .set_right_block_frame
    lda     #10
    jmp     .draw_flat_disc
.speed_up
    lda     $fd
    bmi     .disc_up_left
    lda     $fe                     ; Mask to the left bit.
    asl
    and     $fe
    sta     $fd                     ; Retain the original mask.
    lda     player_xpos,y
    clc
    adc     #6
    sta     player_disc_xpos,y
    jsr     .set_right_block_frame
    jmp     .disc_up_common
.vspeed_neutral
    lda     $fd
    bmi     .disc_left
    lda     $fe                     ; Mask to the left bit.
    asl
    and     $fe
    sta     $fe
    lda     player_xpos,y
    clc
    adc     #7
    sta     player_disc_xpos,y
    jsr     .set_right_block_frame
    lda     #4
    jmp     .draw_flat_disc
.disc_down_left
    lda     $fe                     ; Mask to the left bit.
    asl
    and     $fe
    sta     $fe
    lda     player_xpos,y
    sta     player_disc_xpos,y
    jsr     .set_left_block_frame
    lda     #10
    jmp     .draw_flat_disc
.disc_up_left
    lda     $fe                     ; Mask to the right bit,
    lsr
    and     $fe
    sta     $fd                     ; Retain the original mask.
    lda     player_xpos,y
    sta     player_disc_xpos,y
    jsr     .set_left_block_frame
.disc_up_common
    lda     player_ypos,y
    sta     player_disc_ypos,y
    tay
    ldx     #6
.disc_up_draw_loop
    lda     $fd                     ; Get the one-bit mask.
    ora     $1300,y
    sta     $1300,y
    iny
    dex
    beq     .draw_done
    cpx     #4
    beq     .flip_up_draw_bits
    jmp     .disc_up_draw_loop
.flip_up_draw_bits
    lda     $fd
    eor     $fe                     ; Flip the bits.
    sta     $fd
    jmp     .disc_up_draw_loop
.disc_left
    lda     $fe                     ; Mask to the left bit.
    asl
    and     $fe
    sta     $fe
    lda     player_xpos,y
    sta     player_disc_xpos,y
    jsr     .set_left_block_frame
    lda     #4
.draw_flat_disc
    clc
    adc     player_ypos,y
    sta     player_disc_ypos,y
    tay
    ldx     #6
.draw_flat_disc_loop
    lda     $fe
    ora     $1300,y
    sta     $1300,y
    iny
    dex
    bne     .draw_flat_disc_loop
.draw_done
    rts
.set_left_block_frame
    lda     #player_left
    sta     player_facing,y
    lda     #<player_left_block_cycle
    sta     player_frame_table,y
    lda     #>player_left_block_cycle
    jmp     .set_frame_common
.set_right_block_frame
    lda     #player_right
    sta     player_facing,y
    lda     #<player_right_block_cycle
    sta     player_frame_table,y
    lda     #>player_right_block_cycle
.set_frame_common
    sta     player_frame_table + 1,y
    rts

; Check disc impact against the doors on the right wall.
; Expects the disc y position in y.
check_disc_right_wall subroutine
    jsr     check_disc_side_wall_impact
    bmi     .no_contact
    clc                             ; Convert to a door index.
    adc     #3
    tay
    lda     door_states,y
    cmp     #door_open
    bne     .no_contact
    jmp     open_door
.no_contact
    rts

; Check disc impact against the doors on the left wall.
; Expects the disc y position in y.
check_disc_left_wall subroutine
    jsr     check_disc_side_wall_impact
    bmi     .no_contact
    beq     .top_door               ; Convert to a door index.
    cmp     #1
    beq     .middle_door
    ldy     #9
    .byte   $2c
.top_door
    ldy     #11
    .byte   $2c
.middle_door
    ldy     #10
    lda     door_states,y
    cmp     #door_open
    bne     .no_contact
    jmp     open_door
.no_contact
    rts

; Check disc impact against the doors on the bottom wall.
; Expects the disc x position in x.
check_disc_bottom_wall subroutine
    jsr     check_disc_upper_lower_wall_impact
    bmi     .no_contact
    beq     .left_door              ; Convert to a door index.
    cmp     #1
    beq     .middle_door
    ldy     #6
    .byte   $2c
.left_door
    ldy     #8
    .byte   $2c
.middle_door
    ldy     #7
    lda     door_states,y
    cmp     #door_open
    bne     .no_contact
    jmp     open_door
.no_contact
    rts

; Check disc impact against the doors on the top wall.
; Expects the disc x position in x.
check_disc_top_wall subroutine
    jsr     check_disc_upper_lower_wall_impact
    bmi     .no_contact
    tay
    lda     door_states,y
    cmp     #door_open
    bne     .no_contact
    jmp     open_door
.no_contact
    rts

; Checks a disc against side-wall impact.
; Result is placed in the accumulator.
; 0  - upper door
; 1  - middle door
; 2  - bottom door
; -1 - no impact
; Expects the disc y position in y.
check_disc_side_wall_impact subroutine
    cpy     #$5f
    bcc     .no_contact
    cpy     #$6f
    bcc     .top_door
    cpy     #$87
    bcc     .no_contact
    cpy     #$97
    bcc     .middle_door
    cpy     #$af
    bcc     .no_contact
    cpy     #$bf
    bcs     .no_contact
    lda     #2
    rts
.no_contact
    lda     #-1
    rts
.top_door
    lda     #0
    rts
.middle_door
    lda     #1
    rts

; Checks a disc against upper and lower wall impact.
; Result is placed in the accumulator.
; 0  - left door
; 1  - middle door
; 2  - right door
; -1 - no impact
; Expects the disc x position in x.
check_disc_upper_lower_wall_impact subroutine
    cpx     #$54
    bcc     .no_contact
    cpx     #$5b
    bcc     .left_door
    cpx     #$7c
    bcc     .no_contact
    cpx     #$83
    bcc     .middle_door
    cpx     #$a4
    bcc     .no_contact
    cpx     #$ab
    bcs     .no_contact
    lda     #2
    rts
.no_contact
    lda     #-1
    rts
.left_door
    lda     #0
    rts
.middle_door
    lda     #1
    rts

; Spawns a door at the top of the screen.
; Expects the door number in y, should be 0-2.
spawn_top_door subroutine
    lda     door_states,y           ; Check the door state.
    cmp     #door_jammed
    beq     .done                   ; Don't need to do anything.
    ldx     door_offsets,y          ; Get the byte offset.
    cmp     #door_open
    beq     .jam_door               ; Draw a jammed door.
    lda     #31                     ; Open door index.
    sta     $2000,x
    lda     #32
    sta     $2001,x
    lda     #33
    sta     $2028,x
    lda     #34
    sta     $2029,x
    lda     #door_open
    sta     door_states,y
.done
    rts
.jam_door
    lda     #27                     ; Jammed door tile.
    sta     $2000,x
    lda     #28
    sta     $2001,x
    lda     #29
    sta     $2028,x
    lda     #30
    sta     $2029,x
    lda     #door_jammed
    sta     door_states,y
    rts

; Spawns a door on the right side of the screen.
; Expects the door number in y, should be 3-5.
spawn_right_door subroutine
    cpy     #3                      ; Check the door ID.
    beq     .top_door
    cpy     #4
    beq     .middle_door
    lda     door_states,y
    cmp     #door_jammed
    beq     .done                   ; Nothing to do.
    cmp     #door_open
    beq     .jam_bottom_door        ; Draw the jammed door.
    lda     #35                     ; Door open tile.
    sta     $22bb
    sta     $22bb + 40
    lda     #door_open
    sta     door_states,y
    rts
.jam_bottom_door
    lda     #37                     ; Door jammed tile.
    sta     $22bb
    lda     #39
    sta     $22bb + 40
    lda     #door_jammed
    sta     door_states,y
    rts
.top_door
    lda     door_states,y
    cmp     #door_jammed
    beq     .done                   ; Nothing to do.
    cmp     #door_open
    beq     .jam_top_door
    lda     #36                     ; Door open tile with grid line.
    sta     $212b
    lda     #35                     ; Door open tile.
    sta     $212b + 40
    lda     #door_open
    sta     door_states,y
    rts
.jam_top_door
    lda     #38                     ; Door jammed tile with grid line.
    sta     $212b
    lda     #39
    sta     $212b + 40
    lda     #door_jammed
    sta     door_states,y
    rts
.middle_door
    lda     door_states,y
    cmp     #door_jammed
    beq     .done                   ; Nothing to do.
    cmp     #door_open
    beq     .jam_middle_door;       ; Draw the jammed door.
    lda     #35                     ; Door open tile.
    sta     $21f3
    sta     $21f3 + 40
    lda     #door_open
    sta     door_states,y
.done
    rts
.jam_middle_door
    lda     #37                     ; Door jammed tile.
    sta     $21f3
    lda     #39
    sta     $21f3 + 40
    lda     #door_jammed
    sta     door_states,y
    rts

; Spawns a door on the left side of the screen.
; Expects the door number in y, should be 9-11.
spawn_left_door subroutine
    cpy     #11                     ; Check the door ID.
    beq     .top_door
    cpy     #10
    beq     .middle_door
    lda     door_states,y
    cmp     #door_jammed
    beq     .done                   ; Nothing to do.
    cmp     #door_open
    beq     .jam_bottom_door        ; Draw a jammed door.
    lda     #44                     ; Door opened tile.
    sta     $2294
    sta     $2294 + 40
    lda     #door_open
    sta     door_states,y
    rts
.jam_bottom_door
    lda     #46                     ; Door jammed tile.
    sta     $2294
    lda     #48
    sta     $2294 + 40
    lda     #door_jammed
    sta     door_states,y
    rts
.top_door
    lda     door_states,y
    cmp     #door_jammed
    beq     .done                   ; Nothing to do.
    cmp     #door_open
    beq     .jam_top_door           ; Draw the jammed door.
    lda     #45                     ; Door open tile with grid line.
    sta     $2104
    lda     #44                     ; Door open tile.
    sta     $2104 + 40
    lda     #door_open
    sta     door_states,y
    rts
.jam_top_door
    lda     #47                     ; Door jammed tile with grid line.
    sta     $2104
    lda     #48                     ; Door jammed tile.
    sta     $2104 + 40
    lda     #door_jammed
    sta     door_states,y
    rts
.middle_door
    lda     door_states,y
    cmp     #door_jammed
    beq     .done                   ; Nothing to do.
    cmp     #door_open
    beq     .jam_middle_door        ; Draw the jammed door.
    lda     #44                     ; Door open tile.
    sta     $21cc
    sta     $21cc + 40
    lda     #door_open
    sta     door_states,y
.done
    rts
.jam_middle_door
    lda     #46                     ; Door jammed tile.
    sta     $21cc
    lda     #48
    sta     $21cc + 40
    lda     #door_jammed
    sta     door_states,y
    rts

; Open a given door.  Places a closed door if
; the door doesn't yet exist, opens it otherwise.
; Expects the door number in y.
; The doors are numbered clockwise from the top left.
; Preserves y.
open_door subroutine
    lda     door_states,y
    cmp     #door_open
    bne     .no_jam_sound
    tya
    pha
    lda     #6
    jsr     DOTUNE
    lda     #7
    jsr     DOTUNE
    pla
    tay
.no_jam_sound
    cpy     #3                      ; Figure out the door side.
    bmi     .top_door
    cpy     #6
    bmi     .right_door
    cpy     #9
    bmi     .bottom_door
    jmp     spawn_left_door
.top_door
    jmp     spawn_top_door
.right_door
    jmp     spawn_right_door
.bottom_door
    ; Fall through.

; Spawns a door on the bottom side of the screen.
; Expects the door number in y, should be 6-8.
spawn_bottom_door subroutine
    jsr     get_bottom_door_offset
    lda     door_states,y           ; Check the door state.
    cmp     #door_jammed
    beq     .done                   ; Nothing to do.
    cmp     #door_open
    beq     .jam_door               ; Jam the door.
    lda     #40                     ; Door open tile left.
    sta     $2320,x
    lda     #41                     ; Door open tile right.
    sta     $2321,x
    lda     #door_open
    sta     door_states,y
.done
    rts
.jam_door
    lda     #42                     ; Door jammed tile left.
    sta     $2320,x
    lda     #43
    sta     $2321,x
    lda     #door_jammed
    sta     door_states,y
    rts

; Get the byte offset for a door index on the bottom.
; Expects the index in y, must be 6-8.  Returns
; the offset in x.
get_bottom_door_offset subroutine
    cpy     #6                      ; Convert to a lookup index.
    beq     .right_door
    cpy     #7
    beq     .middle_door
    ldx     #0
    .byte   $2c
.right_door
    ldx     #2
    .byte   $2c
.middle_door
    ldx     #1
.draw_door
    lda     door_offsets,x          ; Get the byte offset.
    tax
    rts

; Open three doors starting with the one in y.
; This is to spawn new enemy waves.  Only
; operates on closed doors.
open_three_doors subroutine
    lda     door_states,y
    cmp     #door_jammed
    beq     .skip_first
    jsr     open_door
.skip_first
    iny
    lda     door_states,y
    cmp     #door_jammed
    beq     .skip_second
    jsr     open_door
.skip_second
    iny
    lda     door_states,y
    cmp     #door_jammed
    beq     .skip_third
    jmp     open_door
.skip_third
    rts

; Checks to see if a left-right side door
; wrap is possible.  Expects the player's
; ypos in the y register.  Pass 0 in the x
; register if the player is on the left side,
; 1 if the player is on the right.  Sets the 0 bit
; if the player should wrap, clears it
; otherwise.
check_left_right_door_wrap subroutine
    cpy     #$5d
    bcc     .no_wrap
    cpy     #$63
    bcc     .top_wrap
    cpy     #$85
    bcc     .no_wrap
    cpy     #$8b
    bcc     .middle_wrap
    cpy     #$ad
    bcc     .no_wrap
    cpy     #$b3
    bcs     .no_wrap
    cpx     #0                      ; Check for contact jam.
    bne     .check_bottom_right_jam
    lda     door_states + 9
    cmp     #door_open
    bne     .no_bottom_jam
    ldy     #9
    jmp     .jam_bottom_door
.check_bottom_right_jam
    lda     door_states + 5
    cmp     #door_open
    bne     .no_bottom_jam
    ldy     #5
.jam_bottom_door
    jsr     open_door
.no_bottom_jam
    lda     door_states + 5         ; Check the door states.
    beq     .no_wrap
    lda     door_states + 9
    beq     .no_wrap
    ldy     #5
    jsr     open_door
    ldy     #9
    jsr     open_door
    lda     #0                      ; Return value.
    rts
.no_wrap
    lda     #1                      ; Return value.
    rts
.top_wrap
    cpx     #0                      ; Check for contact jam.
    bne     .check_top_right_jam
    lda     door_states + 11
    cmp     #door_open
    bne     .no_top_jam
    ldy     #11
    jmp     .jam_top_door
.check_top_right_jam
    lda     door_states + 3
    cmp     #door_open
    bne     .no_top_jam
    ldy     #3
.jam_top_door
    jsr     open_door
.no_top_jam
    lda     door_states + 3
    beq     .no_wrap
    lda     door_states + 11
    beq     .no_wrap
    ldy     #3                      ; Open both doors.
    jsr     open_door
    ldy     #11
    jsr     open_door
    lda     #0                      ; Return value.
    rts
.middle_wrap
    cpx     #0                      ; Check for contact jam.
    bne     .check_middle_right_jam
    lda     door_states + 10
    cmp     #door_open
    bne     .no_middle_jam
    ldy     #10
    jmp     .jam_middle_door
.check_middle_right_jam
    lda     door_states + 4
    cmp     #door_open
    bne     .no_middle_jam
    ldy     #4
.jam_middle_door
    jsr     open_door
.no_middle_jam
    lda     door_states + 4
    beq     .no_wrap
    lda     door_states + 10
    beq     .no_wrap
    ldy     #4
    jsr     open_door
    ldy     #10
    jsr     open_door
    lda     #0
    rts

; Checks to see if a top-bottom side door
; wrap is possible.  Expects the player's
; xpos in the x register.  Pass 0 in the y
; register if the player in on the top, 1
; if the player is on the bottom.  Sets the 0 bit
; if the player should wrap, clears it
; otherwise.
check_top_bottom_door_wrap subroutine
    cpx     #$52
    bcc     .no_wrap
    cpx     #$56
    bcc     .left_wrap
    cpx     #$7a
    bcc     .no_wrap
    cpx     #$7e
    bcc     .middle_wrap
    cpx     #$a2
    bcc     .no_wrap
    cpx     #$a6
    bcs     .no_wrap
    cpy     #0                      ; Check for contact jam.
    bne     .check_low_right_jam
    lda     door_states + 2
    cmp     #door_open
    bne     .no_right_jam
    ldy     #2
    jmp     .jam_right_door
.check_low_right_jam
    lda     door_states + 6
    cmp     #door_open
    bne     .no_right_jam
    ldy     #6
.jam_right_door
    jsr     open_door
.no_right_jam
    lda     door_states + 2
    beq     .no_wrap
    lda     door_states + 6
    beq     .no_wrap
    ldy     #2
    jsr     open_door
    ldy     #6
    jsr     open_door
    lda     #0                      ; Return value.
    rts
.no_wrap
    lda     #1                      ; Return value.
    rts
.left_wrap
    cpy     #0                      ; Check for contact jam.
    bne     .check_low_left_jam
    lda     door_states
    cmp     #door_open
    bne     .no_left_jam
    jmp     .jam_left_door
.check_low_left_jam
    lda     door_states + 8
    cmp     #door_open
    bne     .no_left_jam
    ldy     #8
.jam_left_door
    jsr     open_door
.no_left_jam
    lda     door_states
    beq     .no_wrap
    lda     door_states + 8
    beq     .no_wrap
    ldy     #0
    jsr     open_door
    ldy     #8
    jsr     open_door
    lda     #0                      ; Return value.
    rts
.middle_wrap
    cpy     #0                      ; Check for contact jam.
    bne     .check_low_middle_jam
    lda     door_states + 1
    cmp     #door_open
    bne     .no_middle_jam
    iny
    jmp     .jam_middle_door
.check_low_middle_jam
    lda     door_states + 7
    cmp     #door_open
    bne     .no_middle_jam
    ldy     #7
.jam_middle_door
    jsr     open_door
.no_middle_jam
    lda     door_states + 1
    beq     .no_wrap
    lda     door_states + 7
    beq     .no_wrap
    ldy     #1
    jsr     open_door
    ldy     #7
    jsr     open_door
    lda     #0                      ; Return value.
    rts

; Applies a hit to the player whose
; data block is in y.  If the player
; is dead, will adjust scores accordingly.
; Requires the disc erasure mask in the accumulator.
; Returns 1 in the accumulator if the
; striking disc should return, 0 otherwise.
; Pass the CPU flags in x to test for a
; double damage disc.  The player->enemy
; hit checking code keeps the player's
; missile hit register in x with the upper
; four bits masked out.  This is OK, since it
; will cause the double-damage check to always fail.
hit_player subroutine
    sta     $f1                     ; Save the mask.
    sty     $f2                     ; Save the pointer.
    lda     player_health,y         ; Is the player alive?
    bne     .alive                  ; 0 bit set, no disc return.
    rts
.alive
    txa                             ; Check the double damage flag.
    and     #cpu_double_flag
    beq     .single_damage
    lda     #2                      ; 2 damage.
    jmp     .check_for_death
.single_damage
    lda     #1                      ; 1 damage.
.check_for_death
    cmp     player_health,y
    bmi     .player_alive
    jmp     .kill_player
.player_alive
    jsr     invert                  ; Invert the damage total.
    adc     player_health,y         ; Apply the damage.
    sta     player_health,y
    lda     player_count            ; Guards need an adjustment.
    cmp     #1
    bne     .not_guard
    jsr     enemy_is_guard
    bne     .not_guard
    lda     player_health,y
    tay
    iny
    jmp     .update_speed
.not_guard
    lda     player_health,y
    tay
.update_speed
    lda     health_speed_table,y
    ldy     $f2
    sta     player_speed_index,y
    lda     #health_recovery_time   ; Reset the recovery timer.
    sta     player_recover_timer,y
    lda     player_facing,y         ; Switch to the hit animation.
    beq     .hit_facing_left
    lda     player_disc_state,y
    beq     .right_disc_position
    cmp     #disc_stick
    bne     .no_right_disc_position
.right_disc_position
    lda     player_xpos,y
    clc
    adc     #5
    sta     player_disc_xpos,y
    jsr     apply_hit_missile
.no_right_disc_position
    lda     #<player_right_hit_cycle
    sta     player_frame_table,y
    lda     #>player_right_hit_cycle
.hit_common
    sta     player_frame_table + 1,y
    lda     #$12
    jsr     DOTUNE
    lda     #$13
    jsr     DOTUNE
    ldy     $f2                     ; Recover the pointer.
    lda     #0
    sta     player_frame_index,y
    sta     player_table_end,y
    lda     player_cpu_flags,y      ; Is this a CPU player?
    bmi     .handle_cpu_hit
    lda     player_last_move,y      ; If the player was moving,
    beq     .no_player_hit_tag      ; tag a hit to "stick" the player.
    lda     #1                      ; Flag disc return.
    sta     player_hit_tag,y        ; Indicate a hit.
    rts
.handle_cpu_hit
    lda     #0                      ; Stop all CPU movement.
    sta     player_cpu_angle,y
    sta     player_x_speed,y
    sta     player_y_speed,y
.no_player_hit_tag
    lda     #1                      ; Flag disc return.
    rts
.hit_facing_left
    lda     player_disc_state,y
    beq     .left_disc_position
    cmp     #disc_stick
    bne     .no_left_disc_position
.left_disc_position
    lda     player_xpos,y
    clc
    adc     #2
    sta     player_disc_xpos,y
    jsr     apply_hit_missile
.no_left_disc_position
    lda     #<player_left_hit_cycle
    sta     player_frame_table,y
    lda     #>player_left_hit_cycle
    jmp     .hit_common
.kill_player
    lda     player_count            ; Skip during battle mode.
    cmp     #1
    bne     .continue_kill
    cpy     #player_data_1          ; Don't adjust enemy count if
    beq     .player_dead            ; the player is killed.
    lda     enemies_killed          ; Increase the kill count, but
    cmp     #4                      ; not beyond 4.  If 4 is reached,
    beq     .continue_kill          ; there will be no wave bonus.
    inc     enemies_killed
    jmp     .continue_kill
.player_dead
    lda     #mode_game_over
    sta     game_mode
.continue_kill
    jsr     start_death_animation
    lda     player_count
    cmp     #1
    bne     .battle_scoring
    cpy     #player_data_1
    bne     .arcade_scoring
    lda     #0                      ; No disc return.
    rts
.apply_score
    lda     player_count
    cmp     #1
    beq     .arcade_scoring
.battle_scoring
    lda     player_data_1 + player_health
    beq     .no_p1_score
    inc     $2002
    lda     $2002
    cmp     #$1a
    bne     .no_p1_score
    ldy     #$10
    sty     $2002
    iny
    sty     $2001     
.no_p1_score
    lda     player_data_2 + player_health
    beq     .no_p2_score
    inc     $2007
    lda     $2007
    cmp     #$1a | $40
    bne     .no_p2_score
    ldy     #$10 | $40
    sty     $2007
    iny
    sty     $2006
.no_p2_score
    lda     player_count            ; Bail on 2 player battles.
    cmp     #2
    beq     .no_p4_score
    lda     player_data_3 + player_health
    beq     .no_p3_score
    inc     $200c
    lda     $200c
    cmp     #$1a | $80
    bne     .no_p3_score
    ldy     #$10 | $80
    sty     $200c
    iny
    sty     $200b
.no_p3_score
    lda     player_count            ; Bail on 3 player battles.
    cmp     #3
    beq     .no_p4_score
    lda     player_data_4 + player_health
    beq     .no_p4_score
    inc     $2011
    lda     $2011
    cmp     #$1a | $c0
    bne     .no_p4_score
    ldy     #$10 | $c0
    sty     $2011
    iny
    sty     $2010
.no_p4_score
    lda     #0                      ; Disc pass-through.
    rts
.arcade_scoring
    txa                             ; Save x.
    pha
    lda     enemies_killed          ; Check for the wave bonus.
    cmp     #3
    bne     .no_wave_bonus
    lda     reinforced
    bne     .no_wave_bonus
    lda     #6                      ; Bonus is kill value * 5, so
    .byte   $2c                     ; score 6 times (bonus + kill).
.no_wave_bonus
    lda     #1                      ; Loop counter.
    sta     $f0
.score_loop
    ldx     #$12
    jsr     add_to_score
    dec     $f0
    bne     .score_loop
    ; Fall through!

; Check to see if the player has reached
; the next tier on the scoring table.
check_score_threshold subroutine
    ldy     score_level             ; See if the table is already maxed.
    cpy     #max_score_level
    beq     .done
    ldx     score_thresholds,y      ; Get the start digit.
    iny
.loop
    lda     score_thresholds,y
    bmi     .next_level             ; On $ff, the check has passed.
    cmp     $2000,x                 ; Check the digit.
    beq     .next
    bpl     .done                   ; If greater, not there yet.
.next
    iny                             ; Next digit.
    inx                             ; Next character.
    jmp     .loop
.next_level
    iny                             ; Update the stored level.
    sty     score_level
    ldy     warrior_level           ; Go to the next warrior level.
.warrior_loop
    iny
    lda     warrior_scores,y        ; Look for $ff.
    bpl     .warrior_loop
    iny                             ; One more to reach the next level.
    sty     warrior_level
.done
    pla
    tax                             ; Retrieve x.
    lda     #0                      ; Clear the 0 bit.  No disc return.
    rts

; Quick and hacky routine for applying missile
; pos after a player has been hit.  Expects the
; players data block in y and the new missile
; position in the accumulator.
apply_hit_missile subroutine
    cpy     #player_data_1
    bne     .not_1
    sta     HPOSM0
    rts
.not_1
    cpy     #player_data_2
    bne     .not_2
    sta     HPOSM1
    rts
.not_2
    cpy     #player_data_3
    bne     .not_3
    sta     HPOSM2
    rts
.not_3
    sta     HPOSM3
    rts

; Start the death animation for the player
; whose data block is in y.
start_death_animation subroutine
    tya
    pha                             ; Save y.
    lda     #$a
    jsr     DOTUNE
    lda     #$b
    jsr     DOTUNE
    pla
    tay
    lda     player_disc_state,y     ; Let the player's disc keep going
    cmp     #disc_in_flight         ; if it's currently in flight.
    beq     .no_disc_erase          ; Otherwise, erase it.
    lda     #disc_dead
    sta     player_disc_state,y
    sty     $f2                     ; Save the pointer.
    lda     player_disc_ypos,y
    tay
    lda     $f1                     ; Get the mask.
    stx     $f3                     ; Save the hitmask.
    jsr     erase_disc
    ldx     $f3                     ; Retrieve the hitmask.
    ldy     $f2                     ; Retrieve the pointer.
.no_disc_erase
    lda     #0
    sta     player_frame_index,y
    sta     player_health,y
    sta     player_x_speed,y
    sta     player_y_speed,y
    lda     #player_frame_time
    sta     player_frame_tick,y
    lda     #player_death_last_frame
    sta     player_table_end,y
    lda     player_facing,y         ; Switch to the death animation.
    beq     .death_facing_left
    lda     #<player_right_death_cycle
    sta     player_frame_table,y
    lda     #>player_right_death_cycle
    jmp     .death_common
.death_facing_left
    lda     #<player_left_death_cycle
    sta     player_frame_table,y
    lda     #>player_left_death_cycle
.death_common
    sta     player_frame_table + 1,y
    rts

; Add to the player's score.  Expects the digit
; to start adding from in x.  This allows the
; recognizer's x10 score to be easily applied.
add_to_score subroutine
    ldy     warrior_level           ; Offset into the score table.
.loop
    lda     warrior_scores,y        ; Get the value to add.
    beq     .next                   ; If 0, do nothing.
    bmi     .done                   ; If $ff, all done.
.normal_add
    clc
    adc     $2000,x                 ; Add the digit.
    cmp     #10                     ; Check for new digit.
    bpl     .existing_digit
    adc     #$10                    ; Add $10 to reach the character code.
.existing_digit
    cmp     #$1a                    ; Check for carry.
    bmi     .no_carry
    sta     $ff                     ; Save the value.
    stx     $fe                     ; Save the digit.
.carry_loop
    dex                             ; Back up a digit.
    lda     $2000,x                 ; Check it.
    bne     .digit_exists           ; If 0, add a new digit.
    lda     #$11
    sta     $2000,x
    jmp     .done_carry
.digit_exists
    cmp     #$19                    ; If 9, set it to 0 and continue.
    bne     .simple_carry
    lda     #$10
    sta     $2000,x
    jmp     .carry_loop
.simple_carry
    inc     $2000,x
.done_carry
    lda     $ff                     ; Retrieve the original_value.
    ldx     $fe                     ; Retrieve the original character.
    sec                             ; Reduce the value.
    sbc     #$a
.no_carry
    sta     $2000,x                 ; Write the new digit.
.next
    iny                             ; Next source digit.
    dex                             ; Next screen digit.
    jmp     .loop
.done
    rts

; Spawn a new wave of enemies.
spawn_new_wave subroutine
    jsr     count_living_enemies    ; If all enemies are dead, wait
    bne     .reinforce_ok           ; for discs to clear before
    jsr     count_living_discs      ; spawning the new wave.
    beq     .reinforce_ok
    lda     #0                      ; Reset the timer.
    sta     spawn_timer_low
    lda     #follow_on_spawn_time
    sta     spawn_timer_high
    rts
.reinforce_ok
    jsr     close_all_open_doors
    lda     wave_phase
    cmp     #wave_reinforce
    beq     .spawn_reinforcement
    lda     recognizer_state        ; Check for recognizer entry.
    cmp     #recognizer_leave       ; No back-to-back recognizers
    bne     .no_prior_recognizer
    lda     #recognizer_inactive    ; Reset the state.
    sta     recognizer_state
    jmp     .spawn_full_wave
.no_prior_recognizer
    cmp     #recognizer_init
    beq     .spawn_recognizer
    cpx     #4                      ; At least four jammed doors?
    bpl     .random_door_loop
    jmp     .spawn_full_wave
.random_door_loop
    lda     RANDOM
    and     #$0f                    ; Mask to 16.
    cmp     #13                     ; Want 12 or lower.
    bpl     .random_door_loop
    cmp     $fe                     ; <= jammed doors, spawn recognizer.
    beq     .spawn_recognizer_from_doors
    bmi     .spawn_recognizer_from_doors
    jmp     .spawn_full_wave        ; No enemies?  Spawn a new wave.
.spawn_recognizer_from_doors
    inc     recognizer_state
.spawn_recognizer
    ldx     #$ff                    ; Clear the stack, cancel the return.
    txs
    jmp     recognizer_loop
.spawn_reinforcement
    lda     #1                      ; Set the reinforcements flag.
    sta     reinforced
    lda     #0                      ; Reset the timer.
    sta     spawn_timer_low
    lda     #follow_on_spawn_time
    sta     spawn_timer_high
.find_spawn_door
    lda     RANDOM                  ; Find a door for the follow-on.
    and     #$03                    ; Mask to 0-3.
    tay
    lda     reinforcement_doors,y   ; Get the door index.
    tay                             ; Open it if necessary.
    lda     door_states,y
    cmp     #door_jammed
    beq     .no_door_open
    jsr     open_door
.no_door_open
    lda     spawn_xpos,y            ; Get the spawn positions and save them.
    tax
    lda     spawn_ypos,y
    sta     $f1
    lda     player_data_2 + player_xpos
    bne     .player_1_active        ; Find an empty player slot.
    lda     player_data_2 + player_disc_state
    cmp     #disc_dead
    bne     .player_1_active
    lda     #0                      ; Enemy table adjustment.
    sta     $f0
    lda     #$0c                    ; Disc write mask.
    sta     $fb
    lda     #$15                    ; Graphics page.
    stx     HPOSP1
    ldy     #player_data_2
    jmp     .found_player
.player_1_active
    lda     player_data_3 + player_xpos
    bne     .player_2_active
    lda     player_data_3 + player_disc_state
    cmp     #disc_dead
    bne     .player_2_active
    lda     #1                      ; Enemy table adjustment.
    sta     $f0
    lda     #$30                    ; Disc write mask.
    sta     $fb
    lda     #$16                    ; Graphics page.
    stx     HPOSP2
    ldy     #player_data_3
    jmp     .found_player
.player_2_active
    lda     #2                      ; Enemy table adjustment.
    sta     $f0
    lda     #$17                    ; Must be player 3.
    stx     HPOSP3
    ldy     #player_data_4
.found_player
    sta     $ff                     ; Graphics page.
    lda     #cpu_player_flag        ; Identify the CPU player.
    sta     player_cpu_flags,y
    lda     $f1
    jsr     create_enemy
    jsr     get_difficulty          ; Check the difficulty ramp.
    sta     $ff                     ; Multiply by 3 to index the
    asl                             ; enemy table.
    adc     $ff                     ; Carry is always clear.
    adc     $f0                     ; Add the enemy offset.
    sta     $fe                     ; Save the index.
    sty     $ff                     ; Save the enemy pointer.
    tay
    ldx     enemy_spawn_table,y
    ldy     $ff                     ; Retrieve the enemy pointer.
    jsr     apply_enemy_speedup
    jmp     apply_enemy_definition
.spawn_full_wave
    ldx     #3                      ; Abuse the speed adjust function
    jsr     apply_enemy_speedup     ; to get the player's max health.
    stx     player_data_1 + player_max_health
    stx     player_data_1 + player_health
    lda     health_speed_table,x
    sta     player_data_1 + player_speed_index
    lda     #0
    sta     enemies_killed
    sta     spawn_timer_low         ; Set the timer for follow-ons.
    sta     reinforced              ; Clear the reinforcements flag.
    lda     #follow_on_spawn_time
    sta     spawn_timer_high
    lda     #wave_full
    sta     wave_phase
    lda     #cpu_player_flag        ; Identify the CPU players.
    sta     player_data_2 + player_cpu_flags
    sta     player_data_3 + player_cpu_flags
    sta     player_data_4 + player_cpu_flags
    lda     RANDOM                  ; Pick a side.
    and     #$03                    ; Mask to 0-3.
    beq     .spawn_top
    cmp     #1
    beq     .spawn_right
    cmp     #2
    bne     .spawn_left
    jmp     .spawn_bottom
.spawn_left
    ldy     #9                      ; Open the doors.
    jsr     open_three_doors
    ldx     #left_boundary + 1
    jmp     .side_spawn_common
.spawn_top
    ldy     #0                      ; Open the doors.
    jsr     open_three_doors
    ldy     #player_data_2
    ldx     #84
    stx     HPOSP1
    lda     #$15
    sta     $ff
    lda     #top_boundary + 1
    jsr     create_enemy
    ldy     #player_data_3
    ldx     #124
    stx     HPOSP2
    inc     $ff
    lda     #top_boundary + 1
    jsr     create_enemy
    ldy     #player_data_4
    ldx     #164
    stx     HPOSP3
    inc     $ff
    lda     #top_boundary + 1
    jsr     create_enemy
    jmp     .apply_difficulty
.spawn_right
    ldy     #3                      ; Open the doors.
    jsr     open_three_doors
    ldx     #right_boundary - 1
.side_spawn_common
    stx     HPOSP1
    stx     HPOSP2
    stx     HPOSP3
    lda     #$15
    sta     $ff
    ldy     #player_data_2
    lda     #$5f
    jsr     create_enemy
    ldy     #player_data_3
    inc     $ff
    lda     #$87
    jsr     create_enemy
    ldy     #player_data_4
    inc     $ff
    lda     #$af
    jsr     create_enemy
    jmp     .apply_difficulty
.spawn_bottom
    ldy     #6
    jsr     open_three_doors
    ldy     #player_data_2
    ldx     #84
    stx     HPOSP1
    lda     #$15
    sta     $ff
    lda     #bottom_boundary - 1
    jsr     create_enemy
    ldy     #player_data_3
    ldx     #124
    stx     HPOSP2
    inc     $ff
    lda     #bottom_boundary - 1
    jsr     create_enemy
    ldy     #player_data_4
    ldx     #164
    stx     HPOSP3
    inc     $ff
    lda     #bottom_boundary - 1
    jsr     create_enemy
.apply_difficulty
    jsr     get_difficulty          ; Check the difficulty ramp.
    sta     $ff                     ; Multiply by 3 to index the
    asl                             ; enemy table.
    adc     $ff                     ; Carry is always clear.
    sta     $fe                     ; Save the index.
    tay
    ldx     enemy_spawn_table,y
    jsr     apply_enemy_speedup
    ldy     #player_data_2
    jsr     apply_enemy_definition
    ldy     $fe
    ldx     enemy_spawn_table + 1,y
    jsr     apply_enemy_speedup
    ldy     #player_data_3
    jsr     apply_enemy_definition
    ldy     $fe
    ldx     enemy_spawn_table + 2,y
    jsr     apply_enemy_speedup
    ldy     #player_data_4
    ; Fall through!

; Apply an enemy definition byte to the enemy
; whose data block is in y.  Expects the byte
; to apply in x.
apply_enemy_definition subroutine
    txa
    and     #$c0                    ; Mask the type bits.
    beq     .make_drone             ; Drone, no special action needed.
    cmp     #enemy_armored
    beq     .make_armored
    cmp     #enemy_leader
    beq     .make_leader
    lda     #4                      ; Make a guard.
    sta     player_health,y
    sta     player_max_health,y
    lda     #disc_stick
    sta     player_disc_state,y
    lda     player_disc_ypos,y      ; Move the "disc" up a little.
    sec
    sbc     #2
    sta     player_disc_ypos,y
    sty     $ff
    lda     player_cpu_angle,y      ; Change the xpos of the stick.
    tay
    lda     player_stick_offsets,y
    ldy     $ff
    clc
    adc     player_xpos,y
    sta     player_disc_xpos,y
    lda     #guard_color
    jmp     .apply_speed_and_color
.make_drone
    lda     #1
    sta     player_health,y
    sta     player_max_health,y
    lda     #drone_color
    jmp     .apply_speed_and_color
.make_armored
    lda     #2
    sta     player_health,y
    sta     player_max_health,y
    lda     health_speed_table + 2
    sta     player_speed_index,y
    lda     #armored_color
    pha                             ; Armored enemies don't get speedup.
    jmp     .apply_color
.make_leader
    lda     #1
    sta     player_health,y
    sta     player_max_health,y
    lda     $200e                   ; Check the 100,000s digit.
    bne     .allow_seeker           ; Seekers in play.
    lda     $200f                   ; Check the 10,000s digit.
    cmp     #$12
    bcs     .allow_seeker           ; 20,000+, Seekers in play.
    bit     RANDOM                  ; 50/50 split on normal/double.
    bpl     .normal_leader
.double_leader
    lda     #cpu_player_flag | #cpu_double_flag
    sta     player_cpu_flags,y
    lda     #double_color
    jmp     .apply_speed_and_color
.allow_seeker
    lda     RANDOM                  ; 1/3 chance of each disc type.
    and     #$03                    ; Mask down to two bits.
    beq     .allow_seeker           ; Draw until 1-3.
    cmp     #1
    beq     .double_leader
    cmp     #2
    beq     .normal_leader
    lda     #cpu_player_flag | #cpu_seeker_flag
    sta     player_cpu_flags,y
    lda     #seeker_color
    jmp     .apply_speed_and_color
.normal_leader
    lda     #leader_color
.apply_speed_and_color
    pha                             ; Save the color.
    txa                             ; Recover the definition.
    and     #$3f                    ; Mask off the type bits.
    sta     $ff                     ; Store a copy for adding.
    asl                             ; Multiply by 17, first shift
    asl                             ; left to multiply by 16.
    asl                             ; No need to clear carry, the
    asl                             ; shifts will do it.
    adc     $ff                     ; Add the original value to reach 17.
    sta     player_speed_index,y    ; This is the enemy speed index.
.apply_color
    jsr     get_color_offset
    pla                             ; Get the color back.
    sta     COLPM0,x
    sta     sCOLPM0,x
    jmp     apply_starting_speed

; Find the color register offset for the data
; block in y.  Places the result in x.
get_color_offset subroutine
    ldx     #0                      ; Assume 0.
    cpy     #player_data_1
    beq     .done
    inx
    cpy     #player_data_2
    beq     .done
    inx
    cpy     #player_data_3
    beq     .done
    inx
.done
    rts

; Adjust an enemy definition byte to reflect the
; speed boosts that come at 20,000 and 100,000 points.
; Expects the byte in question in x.
apply_enemy_speedup subroutine
    lda     $200e                   ; Check the 100,000s digit.
    bne     .double_speedup         ; At least 100,000 points.
    lda     $200f                   ; Check the 10,000s digit.
    cmp     #$12
    bcs     .single_speedup         ; At least 20,000 points.
    rts                             ; No speedup.
.double_speedup
    inx
.single_speedup
    inx
    rts

; Find position on the difficulty ramp.
; Returns a value in the accumulator based
; on the following table:
; 0:  < 2,000 points
; 1:  < 3,000 points
; 2:  < 7,000 points
; 3:  < 10,000 points
; 4:  < 15,000 points
; 5:  < 20,000 points
; 6:  < 25,000 points
; 7:  < 30,000 points
; 8:  < 40,000 points
; 9:  < 50,000 points
; 10: < 75,000 points
; 11: < 100,000 points
; 12; < 1,000,000 points
; 12 - 18: 1,000,000+ points
get_difficulty subroutine
    lda     $200d                   ; Check the 1,000,000s digit.
    beq     .under_1m
.get_high_table_index
    lda     RANDOM                  ; Get a value from 1 - 7.
    and     #$07
    beq     .get_high_table_index   ; Loop until 1 - 7.
    clc
    adc     #11                     ; Offset into the high table.
    rts
.under_1m
    lda     $200e                   ; Check the 100,000s digit.
    beq     .under_100k
    lda     #12                     ; 100,000 - 1,000,000.
    rts
.under_100k
    lda     $200f                   ; Check the 10,000s digit.
    beq     .under_10k
    cmp     #$11
    beq     .under_20k
    cmp     #$12
    beq     .under_30k
    cmp     #$13
    beq     .under_40k
    cmp     #$14
    beq     .under_50k
    cmp     #$17
    bmi     .under_75k
    lda     $2010                   ; Check the 1,000s digit.
    cmp     #$15
    bmi     .under_75k
    lda     #11                     ; < 100,000 points.
    rts
.under_75k
    lda     #10
    rts
.under_50k
    lda     #9
    rts
.under_40k
    lda     #8
    rts
.under_30k
    lda     $2010                   ; Check the 1,000s digit.
    cmp     #$15
    bcc     .under_25k
    lda     #7
    rts
.under_25k
    lda     #6
    rts
.under_20k
    lda     $2010                   ; Check the 1,000s digit.
    cmp     #$15
    bcc     .under_15k
    lda     #5
    rts
.under_15k
    lda     #4
    rts
.under_10k
    lda     $2010                   ; Check the 1,000s digit.
    cmp     #$12
    bcc     .under_2k
    cmp     #$13
    bcc     .under_3k
    cmp     #$17
    bcc     .under_7k
    lda     #3                      ; < 10,000 points.
    rts
.under_2k
    lda     #0
    rts
.under_3k
    lda     #1
    rts
.under_7k
    lda     #2
    rts

; Set the initial speeds for a new CPU enemy.
; Expects the enemy's data block in y.
apply_starting_speed subroutine
    lda     player_speed_index,y
    clc
    adc     player_cpu_angle,y
    tax
    lda     player_hspeed_table,x
    sta     player_x_speed,y
    lda     player_vspeed_table,x
    sta     player_y_speed,y
    rts

; Initialize a player data block and
; draw the initial sprite.
; Expects the base address in y,
; the starting xpos in x, and the
; the starting ypos in the accumulator.
; The high-byte of the player's graphics
; page should be in $ff.
create_enemy subroutine
    stx     player_xpos,y
    sta     player_ypos,y
    sta     $fe
    clc
    adc     #held_disc_y_delta
    sta     player_disc_ypos,y
    lda     #subpixel_limit / 2
    sta     player_sub_x,y
    sta     player_sub_y,y
    lda     #0
    sta     player_x_speed,y
    sta     player_y_speed,y
    sta     player_disc_state,y
    sta     player_hit_tag,y
    cpx     #left_boundary + 1
    beq     .enter_from_left
    cpx     #right_boundary - 1
    beq     .enter_from_right
    lda     $fe
    cmp     #top_boundary + 1
    bne     .enter_from_bottom
    jmp     .enter_from_top
.enter_from_bottom
    lda     #player_left
    sta     player_facing,y
    lda     #<player_vertical_cycle
    sta     player_frame_table,y
    lda     #>player_vertical_cycle
    sta     player_frame_table + 1,y
    lda     #player_vertical_last_frame
    sta     player_table_end,y
    lda     #player_frame_time
    sta     player_frame_tick,y
    lda     #1
    sta     player_cpu_angle,y
    lda     player_disc_offsets + 1
    sta     player_disc_offset,y
    txa
    clc
    adc     player_disc_offsets + 1
    sta     player_disc_xpos,y
    jmp     .speed_set
.enter_from_left
    lda     #player_right
    sta     player_facing,y
    lda     #<player_right_cycle
    sta     player_frame_table,y
    lda     #>player_right_cycle
    sta     player_frame_table + 1,y
    lda     #player_right_last_frame
    sta     player_table_end,y
    lda     #player_frame_time
    sta     player_frame_tick,y
    lda     #5
    sta     player_cpu_angle,y
    lda     player_disc_offsets + 5
    sta     player_disc_offset,y
    txa
    clc
    adc     player_disc_offsets + 5
    sta     player_disc_xpos,y
    jmp     .speed_set
.enter_from_right
    lda     #player_left
    sta     player_facing,y
    lda     #<player_left_cycle
    sta     player_frame_table,y
    lda     #>player_left_cycle
    sta     player_frame_table + 1,y
    lda     #player_left_last_frame
    sta     player_table_end,y
    lda     #player_frame_time
    sta     player_frame_tick,y
    lda     #13
    sta     player_cpu_angle,y
    lda     player_disc_offsets + 13
    sta     player_disc_offset,y
    txa
    clc
    adc     player_disc_offsets + 13
    sta     player_disc_xpos,y
    jmp     .speed_set
.enter_from_top
    lda     #player_right
    sta     player_facing,y
    lda     #<player_vertical_cycle
    sta     player_frame_table,y
    lda     #>player_vertical_cycle
    sta     player_frame_table + 1,y
    lda     #player_vertical_last_frame
    sta     player_table_end,y
    lda     #player_frame_time
    sta     player_frame_tick,y
    lda     #9
    sta     player_cpu_angle,y
    lda     player_disc_offsets + 9
    sta     player_disc_offset,y
    txa
    clc
    adc     player_disc_offsets + 9
    sta     player_disc_xpos,y
.speed_set
    lda     #0
    sta     player_frame_index,y
    rts

; Close the door whose index is in y.  Preserves y.
close_door subroutine
    lda     #door_closed
    sta     door_states,y
    cpy     #3
    bmi     .close_top
    cpy     #6
    bmi     .close_right
    cpy     #9
    bmi     .close_bottom
    lda     #8                      ; Door closed tile top.
    ldx     #2                      ; Door closed tile bottom.
    cpy     #11
    beq     .close_left_top
    cpy     #10
    beq     .close_left_middle
    stx     $2294
    stx     $2294 + 40
    rts
.close_left_top
    sta     $2104
    stx     $2104 + 40
    rts
.close_left_middle
    stx     $21cc
    stx     $21cc + 40
    rts
.close_top
    sty     $ff                     ; Save the door index.
    lda     door_offsets,y
    tax
    sec
    sbc     #$14                    ; Convert to pattern offset.
    tay
    lda     back_wall_pattern,y
    sta     $2000,x
    lda     back_wall_pattern + 1,y
    sta     $2001,x
    lda     back_wall_pattern + 40,y
    sta     $2028,x
    lda     back_wall_pattern + 41,y
    sta     $2029,x
    ldy     $ff                     ; Retrieve the door index.
    rts
.close_right
    lda     #9                      ; Door closed tile top.
    ldx     #3                      ; Door closed tile bottom.
    cpy     #3
    beq     .close_right_top
    cpy     #4
    beq     .close_right_middle
    stx     $22bb
    stx     $22bb + 40
    rts
.close_right_top
    sta     $212b
    stx     $212b + 40
    rts
.close_right_middle
    stx     $21f3
    stx     $21f3 + 40
    rts
.close_bottom
    jsr     get_bottom_door_offset
    lda     #4                      ; Door closed tile left.
    sta     $2320,x
    lda     #12                     ; Door closed tile right.
    sta     $2321,x
    rts

; Close all open doors.  Used when spawning a new wave.
; Returns the number of jammed doors in x and $fe.
close_all_open_doors subroutine
    ldx     #0                      ; Jammed door count.
    stx     $fe
    ldy     #11
.loop
    lda     door_states,y
    cmp     #door_jammed
    bne     .not_jammed
    inc     $fe
    jmp     .next
.not_jammed
    cmp     #door_open
    bne     .next
    jsr     close_door
.next
    dey
    bpl     .loop
    ldx     $fe
    rts

; CPU wall bounce checking routines.
; All of them expect the player data
; block address in y.
check_top_wall_bounce subroutine
    lda     game_mode               ; Ignore on game over.
    beq     .game_active
    rts
.game_active
    lda     player_cpu_angle,y
    cmp     #4
    bmi     .bounce_1_3
    cmp     #15
    bmi     no_bounce
    sec                             ; Subtract 12 to get the index.
    sbc     #12
    jmp     .get_new_angle
.bounce_1_3
    sec                             ; Subtract 1 to get the index.
    sbc     #1
.get_new_angle
    sty     $ff                     ; Save the pointer.
    tay
    lda     cpu_direction_inversion_top,y
    ldy     $ff
    sta     player_cpu_angle,y
    jsr     invert_vertical_speed
    jmp     apply_bounce_animation

invert_vertical_speed subroutine
    lda     player_y_speed,y
    jsr     invert
    sta     player_y_speed,y
    rts

invert_horizontal_speed subroutine
    lda     player_x_speed,y
    jsr     invert
    sta     player_x_speed,y
    rts

check_bottom_wall_bounce subroutine
    lda     game_mode               ; Ignore on game over.
    beq     .game_active
    rts
.game_active
    lda     player_cpu_angle,y
    cmp     #7
    bmi     no_bounce
    cmp     #12
    bpl     no_bounce
    sec                             ; Subtract 7 to get the index.
    sbc     #7
    sty     $ff                     ; Save the pointer.
    tay
    lda     cpu_direction_inversion_bottom,y
    ldy     $ff
    sta     player_cpu_angle,y
    jsr     invert_vertical_speed
    jmp     apply_bounce_animation

apply_bounce_animation
    stx     $ff                     ; Save the caller's x.
    ldx     player_cpu_angle,y
    jsr     update_player_animation
    ldx     $ff
no_bounce
    rts

check_right_wall_bounce subroutine
    lda     game_mode               ; Ignore on game over.
    beq     .game_active
    rts
.game_active
    lda     player_cpu_angle,y
    cmp     #3
    bmi     no_bounce
    cmp     #8
    bpl     no_bounce
    sec                             ; Subtract 3 to get the index.
    sbc     #3
    sty     $ff                     ; Save the pointer.
    tay
    lda     cpu_direction_inversion_right,y
    ldy     $ff
    sta     player_cpu_angle,y
    jsr     invert_horizontal_speed
    jmp     apply_bounce_animation

check_left_wall_bounce subroutine
    lda     game_mode               ; Ignore on game over.
    beq     .game_active
    rts
.game_active
    lda     player_cpu_angle,y
    cmp     #11
    bmi     no_bounce
    cmp     #16
    beq     no_bounce
    sec                             ; Subtract 11 to get the index.
    sbc     #11
    sty     $ff                     ; Save the pointer.
    tay
    lda     cpu_direction_inversion_left,y
    ldy     $ff
    sta     player_cpu_angle,y
    jsr     invert_horizontal_speed
    jmp     apply_bounce_animation

; Recover one point of health for a character.
; Expects the character's data block in y.
recover_one_health subroutine
    lda     player_count            ; Ignore in battle mode.
    cmp     #1
    bne     .done
    lda     player_health,y         ; See if health is full.
    cmp     player_max_health,y
    beq     .done
    tax
    lda     #health_recovery_time   ; Reset the health timer.
    sta     player_recover_timer,y
    inx
    stx     player_health,y
    cpy     #player_data_1          ; Guards need an adjustment.
    beq     .not_guard
    jsr     enemy_is_guard
    bne     .not_guard
    inx                             ; Read one value higher.
.not_guard
    lda     health_speed_table,x
    sta     player_speed_index,y
.done
    rts

; CPU processing routine.  Expects the player's
; data block address in y and its disc erasure
; mask in the accumulator.  If this isn't a CPU
; player or the player is dead, will do nothing.
; Otherwise, selects a new movement direction
; and considers throwing a disc at the player.
cpu_adjust subroutine
    sta     $f0                     ; Save the mask.
    lda     player_health,y         ; Is this player dead?
    bne     .player_alive
.dont_process
    rts
.player_alive
    lda     game_mode               ; Do nothing on game over.
    bne     .dont_process
    lda     player_cpu_flags,y      ; Is this a CPU player?
    bpl     .dont_process
    jsr     enemy_is_guard
    bne     .direction_loop
    lda     player_xpos,y
    clc
    adc     #3
    sta     $fe
    lda     player_ypos,y
    adc     #player_sprite_height / 2
    sta     $ff
    jsr     angle_to_player         ; Chase the player.
    clc                             ; Convert to a movement angle.
    adc     #1
    jmp     .no_cpu_crouch
.direction_loop
    lda     RANDOM                  ; Select a random direction. This
    and     #$1f                    ; includes direction 0, which is
    cmp     #17                     ; the crouching state.
    bpl     .direction_loop         ; Draw until valid.
    cmp     #0
    bne     .no_cpu_crouch          ; Switch to the crouch animation on 0.
    lda     player_disc_state,y
    bne     .no_crouch_disc_erase
    lda     $f0                     ; Erase the disc.
    sty     $f1                     ; Save the data pointer.
    jsr     erase_disc
    ldy     $f1
    lda     #crouch_disc_y_delta    ; Position the disc.
    clc
    adc     player_ypos,y
    sta     player_disc_ypos,y
    lda     #disc_crouch
    sta     player_disc_state,y
.no_crouch_disc_erase
    lda     player_facing,y         ; Check the facing.
    beq     .crouch_left
    lda     player_disc_state,y
    cmp     #disc_crouch
    bne     .no_crouch_right_disc_adjust
    lda     #crouch_right_offset
    clc
    adc     player_xpos,y
    sta     player_disc_xpos,y
.no_crouch_right_disc_adjust
    lda     #<player_right_crouch_cycle
    sta     player_frame_table,y
    lda     #>player_right_crouch_cycle
    jmp     .store_crouch
.crouch_left
    lda     player_disc_state,y
    cmp     #disc_crouch
    bne     .no_crouch_left_disc_adjust
    lda     #crouch_left_offset
    clc
    adc     player_xpos,y
    sta     player_disc_xpos,y
.no_crouch_left_disc_adjust
    lda     #<player_left_crouch_cycle
    sta     player_frame_table,y
    lda     #>player_left_crouch_cycle
.store_crouch
    sta     player_frame_table + 1,y
    lda     #0                      ; Restore the direction.
.no_cpu_crouch
    sta     player_cpu_angle,y      ; Store the value.
    beq     .skip_disc_adjust
    lda     player_disc_state,y     ; Move the disc if leaving crouch.
    cmp     #disc_crouch
    bne     .no_disc_adjust         ; Not in hand.
    sty     $f1
    lda     player_disc_ypos,y
    tay
    lda     $f0
    jsr     erase_disc
    ldy     $f1
    lda     player_ypos,y
    clc
    adc     #held_disc_y_delta
    sta     player_disc_ypos,y
    lda     #disc_in_hand
    sta     player_disc_state,y
.no_disc_adjust
    lda     player_cpu_angle,y
.skip_disc_adjust
    pha                             ; Save the direction code.
    clc                             ; Get the speed table index.
    adc     player_speed_index,y
    tax
    lda     player_hspeed_table,x   ; Get the speeds.
    sta     player_x_speed,y
    lda     player_vspeed_table,x
    sta     player_y_speed,y
    pla                             ; Retrieve the direction code.
    tax
    jsr     update_player_animation ; Figure out the animation.
    jsr     enemy_is_guard          ; Is this a guard?
    bne     .not_guard
    rts                             ; Guards have no discs.
.not_guard
    lda     player_disc_state,y     ; Is the disc in hand?
    beq     .process_disc_in_hand
    rts
.process_disc_in_hand
    lda     RANDOM                  ; 50% chance of attack.
    bmi     .cpu_throw_disc
    rts
.cpu_throw_disc
    lda     player_xpos,y
    clc
    adc     #3
    sta     $fe
    lda     player_ypos,y
    adc     #player_sprite_height / #2
    sta     $ff
cpu_aim_disc
    jsr     angle_to_player
    tax
    lda     $f0                     ; Retrieve the mask.
    jmp     throw_disc

; Calculates the angle from the x, y position in
; $fe-$ff to the human player.
; This value is returned in the accumulator.  The
; angle is in disc terms, and needs to be
; incremented if used as a movement direction
; by the guards and paralyzer probe.
angle_to_player subroutine
.source_x equ $fe
.source_y equ $ff       ; MAKE SURE YOU DON'T CLOBBER ANYTHING!
.delta_x equ $fd
.delta_y equ $fc
.player_quad equ $fb
.player_x equ $fa
.player_y equ $f9
.range_45 equ 9
    clc
    lda     player_data_1 + player_xpos
    adc     #3
    sta     .player_x
    lda     player_data_1 + player_ypos
    adc     #player_sprite_height / #2
    sta     .player_y
    lda     .source_x
    cmp     .player_x
    bcs     .player_on_left
    lda     .player_x
    sec
    sbc     .source_x
    sta     .delta_x
    lda     #0
    jmp     .check_vertical_quad
.player_on_left
    lda     .source_x               ; Carry is already set.
    sbc     .player_x
    sta     .delta_x
    lda     #2
.check_vertical_quad
    sta     .player_quad
    lda     .source_y
    cmp     .player_y
    bcs     .player_on_top
    lda     .player_quad
    lsr
    eor     #1
    ora     .player_quad
    sta     .player_quad
    lda     .player_y
    sec
    sbc     .source_y
    sta     .delta_y
    jmp     .quadrant_found
.player_on_top
    lda     .player_quad
    lsr
    ora     .player_quad
    sta     .player_quad
    lda     .source_y               ; Carry is already set.
    sbc     .player_y
    sta     .delta_y
.quadrant_found
    lda     .player_quad
    beq     .top_right
    cmp     #1
    beq     .bottom_right
    cmp     #2
    beq     .bottom_left
    lda     .delta_x
    cmp     .delta_y
    bcs     .octant_6
    cmp     #5
    bcc     .throw_up
    lda     .delta_y
    sbc     .delta_x                ; Carry is set.
    cmp     #.range_45
    bcc     .throw_quad_3_45
    lda     #15
    rts
.octant_6
    lda     .delta_y
    cmp     #player_sprite_height / 2 + 1
    bcc     .throw_left
    lda     .delta_x
    sbc     .delta_y                ; Carry is set.
    cmp     #.range_45
    bcc     .throw_quad_3_45
    lda     #13
    rts
.throw_quad_3_45
    lda     #14
    rts
.top_right
    lda     .delta_y
    cmp     .delta_x
    bcs     .octant_0
    cmp     #player_sprite_height / 2 + 1
    bcc     .throw_right
    lda     .delta_x
    sbc     .delta_y                ; Carry is set.
    cmp     #.range_45
    bcc     .throw_quad_0_45
    lda     #3
    rts
.throw_right
    lda     #4
    rts
.octant_0
    lda     .delta_x
    cmp     #5
    bcc     .throw_up
    lda     .delta_y
    sbc     .delta_x                ; Carry is set.
    cmp     #.range_45
    bcc     .throw_quad_0_45
    lda     #1
    rts
.throw_quad_0_45
    lda     #2
    rts
.throw_up
    lda     #0
    rts
.bottom_right
    lda     .delta_x
    cmp     .delta_y
    bcs     .octant_2
    cmp     #5
    bcc     .throw_down
    sbc     .delta_y                ; Carry is set.
    cmp     #.range_45
    bcc     .throw_quad_1_45
    lda     #7
    rts
.octant_2
    lda     .delta_y
    cmp     #player_sprite_height / 2 + 1
    bcc     .throw_right
    lda     .delta_x
    sbc     .delta_y                ; Carry is set.
    cmp     #.range_45
    bcc     .throw_quad_1_45
    lda     #5
    rts
.throw_quad_1_45
    lda     #6
    rts
.bottom_left
    lda     .delta_y
    cmp     .delta_x
    bcs     .octant_4
    cmp     #player_sprite_height / 2 + 1
    bcc     .throw_left
    lda     .delta_x
    sbc     .delta_y                ; Carry is set.
    cmp     #.range_45
    bcc     .throw_quad_2_45
    lda     #11
    rts
.throw_left
    lda     #12
    rts
.octant_4
    lda     .delta_x
    cmp     #5
    bcc     .throw_down
    lda     .delta_y
    sbc     .delta_x                ; Carry is set.
    cmp     #.range_45
    bcc     .throw_quad_2_45
    lda     #9
    rts
.throw_quad_2_45
    lda     #10
    rts
.throw_down
    lda     #8
    rts

; Update the animation state for a player.
; Expects the new direction of movement in
; x, and the player's data block in y.
update_player_animation subroutine
    lda     player_disc_state,y     ; Do nothing on block.
    cmp     #disc_blocking
    bne     .not_blocking
    rts
.not_blocking
    cpx     #0                      ; Figure out the animation.
    bne     .not_idle
    stx     player_frame_index,y    ; Go idle.
    stx     player_table_end,y
    rts
.not_idle
    lda     player_animation_low_byte,x
    cmp     player_frame_table,y
    bne     .new_animation          ; See if this animation is already going.
    lda     player_animation_high_byte,x
    cmp     player_frame_table + 1,y
    bne     .new_animation
    lda     player_animation_length,x
    sta     player_table_end,y      ; Start the animation if idle.
    jmp     .adjust_held_disc
.new_animation
    lda     player_facing_table,x
    sta     player_facing,y
    lda     player_animation_low_byte,x
    sta     player_frame_table,y
    lda     player_animation_high_byte,x
    sta     player_frame_table + 1,y
    lda     player_animation_length,x
    sta     player_table_end,y
    lda     #0
    sta     player_frame_index,y
.adjust_held_disc
    lda     player_disc_state,y
    cmp     #disc_stick
    bne     .not_stick
    lda     player_stick_offsets,x
    jmp     .update_disc_offset
.not_stick
    lda     player_disc_offsets,x   ; Set the new disc offset.
.update_disc_offset
    sta     player_disc_offset,y
    lda     player_disc_state,y     ; Adjust the disc position if held.
    beq     .disc_adjust
    cmp     #disc_stick
    bne     .no_disc_adjust
.disc_adjust
    lda     player_xpos,y
    clc
    adc     player_disc_offset,y
    sta     player_disc_xpos,y
.no_disc_adjust
    rts

; Draw the recognizer at its current location.
draw_recognizer subroutine
    lda     recognizer_from_top     ; Check for any hackery.
    beq     .normal_draw
    lda     recognizer_bottom       ; Calculate skip bytes.
    sec
    sbc     #recognizer_top_bound
    sta     $fe
    lda     #recognizer_height - 1
    sbc     $fe
    jmp     .draw_ready
.normal_draw
    lda     #0                      ; No skip bytes.
.draw_ready
    ldx     recognizer_ypos
    sta     $fe                     ; Store the skip bytes.
    ldy     #0                      ; Index into image data.
.sprites_loop
    lda     recognizer_image,y      ; Get the repeat count.
    beq     draw_recognizer_eye     ; Terminate on 0.
    sta     $ff                     ; Save the repeat count.
.byte_loop
    lda     $fe                     ; Check for skip bytes.
    beq     .no_skip
    dec     $fe
    jmp     .next_byte
.no_skip
    lda     recognizer_image + 1,y
    sta     $1500,x                 ; Player 1 byte.
    lda     recognizer_image + 2,y
    sta     $1600,x                 ; Player 2 byte.
    inx                             ; Next write_location.
.next_byte
    dec     $ff                     ; Check the loop count.
    bne     .byte_loop
    iny                             ; Go to the next entry.
    iny
    iny
    jmp     .sprites_loop
    ; Fall through!

; Draw the recognizer's eye.
draw_recognizer_eye subroutine
    lda     recognizer_eye_state
    bmi     .no_draw
    lda     recognizer_from_top     ; Check for top-entry hackery.
    beq     .normal_draw
    lda     recognizer_bottom
    sec
    sbc     #recognizer_top_bound + 70
    bcc     .no_draw                ; Don't draw the eye yet.
    beq     .draw_partial_eye       ; Only one row on screen.
    clc
    adc     recognizer_ypos
    tay
    lda     #$f0
    ora     $1300,y
    sta     $1300,y
    lda     #$0f
    and     $12ff,y
    sta     $12ff,y
    rts
.draw_partial_eye
    ldy     recognizer_ypos
    lda     #$f0
    ora     $1300,y
    sta     $1300,y
    rts
.normal_draw
    ldy     recognizer_ypos         ; Eye is a constant offset.
    lda     #$f0                    ; Missiles 2 and 3 are the eye.
    ora     $1304,y
    sta     $1304,y
    lda     #$f0
    ora     $1305,y
    sta     $1305,y
    lda     #$0f                    ; Erase old data,
    and     $1303,y
    sta     $1303,y
    lda     #$0f
    and     $1306,y
    sta     $1306,y
.no_draw
    rts

; Process recognizer movement.
move_recognizer subroutine
    lda     recognizer_sub_x        ; First check horizontal.
    clc
    adc     recognizer_hspeed
    bmi     .move_left
    cmp     #subpixel_limit
    bmi     .update_sub_x
    inc     recognizer_xpos_1
    inc     recognizer_xpos_2
    ldx     recognizer_xpos_1
    stx     HPOSP1           
    ldx     recognizer_xpos_2
    stx     HPOSP2
    sec
    sbc     #subpixel_limit
    jmp     .update_sub_x
.move_left
    dec     recognizer_xpos_1
    dec     recognizer_xpos_2
    ldx     recognizer_xpos_1
    stx     HPOSP1
    ldx     recognizer_xpos_2
    stx     HPOSP2
    clc
    adc     #subpixel_limit
.update_sub_x
    sta     recognizer_sub_x
    lda     recognizer_sub_y
    clc
    adc     recognizer_vspeed
    bpl     .not_move_up
    jmp     .move_up
.not_move_up
    cmp     #subpixel_limit
    bmi     .update_sub_y
    ldy     recognizer_from_top     ; Top entry is special.  Clips
    beq     .no_top_entry_hacks     ; against the score line.
    pha                             ; Save the subpixel.
    inc     recognizer_bottom
    jsr     draw_recognizer
    lda     recognizer_bottom
    cmp     #recognizer_top_bound + #recognizer_height
    bne     .retrieve_subpixel_y
    lda     #0                      ; Clipping is over.
    sta     recognizer_from_top
    jmp     .retrieve_subpixel_y
.no_top_entry_hacks
    pha                             ; Save the subpixel.
    lda     recognizer_ypos         ; Check for bottom exit hacks.
    sec
    sbc     #$ad
    bcc     .normal_move
    sta     $ff                     ; Save the offscreen count.
    lda     recognizer_ypos
    clc
    adc     #recognizer_height + 1
    sec
    sbc     $ff
    jmp     .vertical_loop
.normal_move
    lda     recognizer_ypos
    clc
    adc     #recognizer_height + 1  ; Rows in the recognizer,
    tay                             ; plus the one being moved to.
.vertical_loop
    lda     $14ff,y
    sta     $1500,y
    lda     $15ff,y
    sta     $1600,y
    cpy     recognizer_ypos
    beq     .vertical_move_done
    dey
    jmp     .vertical_loop
.vertical_move_done
    inc     recognizer_ypos
    jsr     draw_recognizer_eye
.retrieve_subpixel_y
    pla
    sec
    sbc     #subpixel_limit
.update_sub_y
    sta     recognizer_sub_y
    rts
.move_up
    dec     recognizer_ypos
    clc
    adc     #subpixel_limit
    sta     recognizer_sub_y
    ldy     recognizer_ypos
    ldx     #recognizer_height + 1
.move_up_loop
    lda     $1501,y
    sta     $1500,y
    lda     $1601,y
    sta     $1600,y
    dex
    beq     .move_up_done
    iny
    jmp     .move_up_loop
.move_up_done
    jmp     draw_recognizer_eye

; Erase the recognizer probe.  Includes a jump-in point
; for erasing the door panel as well.
erase_recognizer_probe subroutine
    ldy     recognizer_probe_y      ; Back up one byte to cover
    dey                             ; distance the probe moved.
    ldx     #15                     ; Erase two extra bytes, one
recognizer_common_erase             ; on top and one on the bottom.
    lda     #0
.loop
    sta     $1700,y
    dex
    beq     .done
    iny
    jmp     .loop
.done
    rts

; Erase the recognizer door panel.
erase_recognizer_door subroutine
    ldy     recognizer_repair_y     ; Back up one byte, see the
    dey                             ; probe code for why.
    ldx     #18                     ; Erase two extra bytes, see
    jmp     recognizer_common_erase ; probe code for why.

; Convert a player's disc to the floppy state
; that occurs after hitting a hard point on
; the recognizer.
make_floppy_disc subroutine
    lda     #$14
    jsr     DOTUNE
    lda     #$15
    jsr     DOTUNE
    lda     #disc_floppy
    sta     player_data_1 + player_disc_state
    lda     player_data_1 + player_disc_x_speed
    jsr     invert
    sta     player_data_1 + player_disc_x_speed
    lda     player_data_1 + player_disc_y_speed
    jsr     invert
    sta     player_data_1 + player_disc_y_speed
    lda     #60
    sta     player_data_1 + player_disc_timer
    rts

; Set the recognizer entry state if it
; isn't already active.
summon_recognizer subroutine
    lda     recognizer_state
    bne     .already_active
    inc     recognizer_state
.already_active
    rts

; Converts the POT values into a direction code.
; Code is placed in the accumulator.  Expects the
; horizontal POT value in x and the vertical in y.
check_joystick_pots subroutine
    cpx     #80
    bcc     .left
    cpx     #140
    bcc     .h_neutral
    lda     #1                      ; Save the direction.
    jmp     .store_h
.left
    lda     #-1
    jmp     .store_h
.h_neutral
    lda     #0
.store_h
    sta     $ff
    cpy     #80
    bcc     .up
    cpy     #140
    bcc     .v_neutral
    lda     $ff                     ; Down, check horizontal.
    bmi     .down_left
    beq     .straight_down
    lda     #7                      ; Down and right.
    rts
.down_left
    lda     #11
    rts
.straight_down
    cpx     #95
    bcc     .down_down_left
    cpx     #125
    bcs     .down_down_right
    lda     #9
    rts
.down_down_left
    lda     #10
    rts
.down_down_right
    lda     #8
    rts
.up
    lda     $ff                     ; Up, check horizontal.
    bmi     .up_left
    beq     .straight_up
    lda     #3                      ; Up and right.
    rts
.up_left
    lda     #15
    rts
.straight_up
    cpx     #95
    bcc     .up_up_left
    cpx     #125
    bcs     .up_up_right
    lda     #1
    rts
.up_up_left
    lda     #16
    rts
.up_up_right
    lda     #2
    rts
.v_neutral
    lda     $ff                     ; Neutral, check horizontal.
    bmi     .straight_left
    beq     .neutral
    cpy     #95
    bcc     .right_right_up
    cpy     #125
    bcs     .right_right_down
    lda     #5                      ; Straight right.
    rts
.right_right_up
    lda     #4
    rts
.right_right_down
    lda     #6
    rts
.straight_left
    cpy     #95
    bcc     .left_left_up
    cpy     #125
    bcs     .left_left_down
    lda     #13
.neutral
    rts
.left_left_up
    lda     #14
    rts
.left_left_down
    lda     #12
    rts

; Disc input handlers.  Each one updates the
; disc action in the player's data block with
; the appropriate action.
keypad_disc_handler subroutine
kdh_point_1
    ldx     keypad_last_key
    bmi     .no_input
    beq     .no_input
    cpx     #10
    bpl     .no_input
    cpx     #5
    beq     .block
kdh_point_2
    lda     player_data_1 + player_disc_state
    cmp     #disc_in_flight
    beq     .check_return
    lda     key_to_disc_direction,x
kdh_point_3
    stx     player_throw_key
kdh_point_4
    sta     player_data_1 + player_disc_action
    lda     #$ff
kdh_point_5
    sta     keypad_last_key
    rts
.no_input
    lda     #action_none
kdh_point_6
    sta     player_data_1 + player_disc_action
    rts
.block
    lda     #action_block
kdh_point_7
    sta     player_data_1 + player_disc_action
    lda     #$ff
kdh_point_8
    sta     keypad_last_key
    rts
.check_return
kdh_point_9
    cpx     player_throw_key
    beq     .no_input
    lda     #action_return
kdh_point_10
    sta     player_data_1 + player_disc_action
    lda     #$ff
kdh_point_11
    sta     keypad_last_key
    rts
keypad_disc_handler_end

trigger_disc_handler subroutine
tdh_point_1
    lda     TRIG0
tdh_point_2
    cmp     trigger_states
    beq     .no_primary_fire        ; No change in trigger.
tdh_point_3
    sta     trigger_states          ; Update last trigger.
    cmp     #0
    bne     .no_primary_fire        ; Trigger release.
tdh_point_4
    lda     player_data_1 + player_disc_state
    cmp     #disc_in_flight
    beq     .disc_return
tdh_point_5
    ldx     player_data_1 + player_last_angle
    dex                             ; Convert to disc angle.
tdh_point_6
    stx     player_data_1 + player_disc_action
    rts
.disc_return
    lda     #action_return
tdh_point_7
    sta     player_data_1 + player_disc_action
    rts
.no_primary_fire
tdh_point_8
    lda     keypad_shift
tdh_point_9
    cmp     shift_states
    beq     .no_secondary_fire      ; No change in trigger.
tdh_point_10
    sta     shift_states            ; Update last trigger.
    cmp     #0
    bne     .no_secondary_fire      ; Trigger release.
    lda     #action_block
tdh_point_11
    sta     player_data_1 + player_disc_action
    rts
.no_secondary_fire
    lda     #action_none
tdh_point_12
    sta     player_data_1 + player_disc_action
    rts
trigger_disc_handler_end

twin_stick_disc_handler subroutine
    ldx     sPOT2
    ldy     sPOT3
    jsr     check_joystick_pots
    beq     .no_stick_input
    ldx     player_data_1 + player_disc_state
    cpx     #disc_in_flight
    beq     .check_return
    sta     player_throw_key
    tax
    dex                             ; Convert to disc direction.
    stx     player_data_1 + player_disc_action
    rts
.check_return
    cmp     #8
    beq     .check_8
    bpl     .check_9_plus
    clc
    adc     #9
    sec
    sbc     player_throw_key
    cmp     #3
    bcs     .no_stick_input
.return
    lda     #action_return
    sta     player_data_1 + player_disc_action
    rts
.check_8
    lda     player_throw_key
    cmp     #1
    beq     .return
    cmp     #15
    bpl     .return
    bmi     .no_stick_input
.check_9_plus
    cmp     #9
    beq     .check_9
    sec
    sbc     #7
    sbc     player_throw_key
    cmp     #3
    bcs     .no_stick_input
    bcc     .return
.check_9
    lda     player_throw_key
    cmp     #16
    beq     .return
    cmp     #3
    bmi     .return
    bpl     .no_stick_input
.no_stick_input
    lda     keypad_last_key + 1     ; Check for any key on stick 2.
    bmi     .no_input
    lda     #$ff
    sta     keypad_last_key + 1
    lda     #action_block
    sta     player_data_1 + player_disc_action
    rts
.no_input
    lda     #action_none
    sta     player_data_1 + player_disc_action
    rts
twin_stick_disc_handler_end

; Convenience macros for loading y with
; modification offsets into the disc control
; functions in memory and updating with the
; value in the accumulator.
    mac kdh_point
    ldy     #kdh_point_{1} - #keypad_disc_handler + #1
    sta     ($fe),y
    endm

    mac tdh_point
    ldy     #tdh_point_{1} - #trigger_disc_handler + #1
    sta     ($fe),y
    endm

; Alter an in-memory disc handler for the player
; number in x.  Expects the address of the copied
; routine in $fe-$ff.
adjust_input_for_player subroutine
    cpx     #1                      ; Check the player number.
    beq     .update_for_1
    cpx     #2
    beq     .update_for_2
    cpx     #3
    beq     .update_for_3
    rts                             ; Nothing to do for player 0.
.update_for_1
    ldy     #player_data_2
    jmp     .player_ready
.update_for_2
    ldy     #player_data_3
    jmp     .player_ready
.update_for_3
    ldy     #player_data_4
.player_ready
    sty     $fd                     ; Save the data pointer.
    lda     control_modes,x
    beq     .update_keypad_handler
    txa
    clc
    adc     #<TRIG0
    tdh_point 1
    txa
    clc
    adc     #<trigger_states
    tdh_point 2
    tdh_point 3
    lda     $fd
    clc
    adc     #player_disc_state
    tdh_point 4
    lda     $fd
    clc
    adc     #player_last_angle
    tdh_point 5
    lda     $fd
    clc
    adc     #player_disc_action
    tdh_point 6
    tdh_point 7
    tdh_point 11
    tdh_point 12
    txa
    clc
    adc     #<keypad_shift
    tdh_point 8
    txa
    clc
    adc     #<shift_states
    tdh_point 9
    tdh_point 10
    rts
.update_keypad_handler
    txa
    clc
    adc     #<keypad_last_key
    kdh_point 1
    kdh_point 5
    kdh_point 8
    kdh_point 11
    lda     $fd
    clc
    adc     #player_disc_state
    kdh_point 2
    txa
    clc
    adc     #player_throw_key
    kdh_point 3
    kdh_point 9
    lda     $fd
    clc
    adc     #player_disc_action
    kdh_point 4
    kdh_point 6
    kdh_point 7
    kdh_point 10
    rts

; Game vertical blank routine.
game_vblank subroutine
    lda     $fe
    pha
    lda     $ff
    pha
    jsr     TUNER
    pla
    sta     $ff
    pla
    sta     $fe
    jsr     keypad_cooldown_process
    lda     #0                      ; Reset the attract mode timer.
    sta     ATRACT
    lda     #$f8                    ; Restore the character set.
    sta     CHBASE
    jmp     ENDINT

