; This file is part of Intellidiscs.

; Intellidiscs is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; Intellidiscs is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with Intellidiscs.  If not, see <http://www.gnu.org/licenses/>.

; Copyright 2021 Phaser Cat Games LLC

; Common keypad handling code for all modes

; Keypad data.
keypad_active_player    equ $600    ; Currently active controller.
keypad_last_key         equ $601    ; Last key pressed by each pad.
keypad_cooldown         equ $605    ; Cooldown timers for each pad.
keypad_shift            equ $609    ; Current state of the shift key
                                    ; (secondary trigger) for each pad.
keypad_change           equ $60d    ; Trigger for changing the active
                                    ; keypad.  Change every other frame.

; Cooldown length.
keypad_cooldown_length  equ 15

; Initialize all keypad data.
keypad_init subroutine
    lda     #0
    sta     keypad_active_player
    lda     #4
    sta     CONSOL
    ; Fall through!

; Clear all of the recorded keys and reset cooldown.
keypad_reset subroutine
    ldy     #3
.loop
    lda     #$ff
    sta     keypad_last_key,y
    lda     #0
    sta     keypad_cooldown,y
    lda     #8
    sta     keypad_shift,y
    dey
    bpl     .loop
    rts

; Process cooldowns and advance to the next keypad.
keypad_process subroutine
    sei
    lda     SKSTAT                  ; Process secondary fire.
    and     #8
    ldy     keypad_active_player
    sta     keypad_shift,y
    cpy     #3
    bne     .no_loop
    ldy     #-1
.no_loop
    lda     #0
    sta     IRQEN                   ; Clear any lingering keypress.
    iny
    sty     keypad_active_player
    tya
    ora     #4
    sta     CONSOL
    lda     #$40
    sta     IRQEN                   ; Reenable key interrupts.
    cli
    rts

; Process the keypad cooldown timers.
keypad_cooldown_process subroutine
    ldy     #3
.loop
    lda     keypad_cooldown,y
    beq     .next
    sec
    sbc     #1
    sta     keypad_cooldown,y
.next
    dey
    bpl     .loop
    rts

; Keypad interrupt routine.
keypad_interrupt subroutine
    tax                             ; Save the key.
    ldy     keypad_active_player
    lda     keypad_cooldown,y       ; Check cooldown.
    bne     .ignore_key
    txa
    sta     keypad_last_key,y
    lda     #keypad_cooldown_length
    sta     keypad_cooldown,y
.ignore_key
    jmp     ENDINT

