; This file is part of Intellidiscs.

; Intellidiscs is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; Intellidiscs is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with Intellidiscs.  If not, see <http://www.gnu.org/licenses/>.

; Copyright 2021 Phaser Cat Games LLC

; This is a somewhat modified version of code given to
; me to Bobby Clark, who also created all the music and
; sound effects.

; Originally intended for the Atari 8-bit computers, I've
; stripped out some stuff that isn't relevant to the 5200
; (PAL timing for example, there is no PAL 5200) and removed
; some parts of the subroutines that weren't relevant to my
; particular project.  Many routines saved registers on the
; stack, which was largely unneeded in my case, for example.

; A few things have been moved around in memory as well, but
; I expect this code will run on an A8 with minimum modification.

;	TUNE VARIABLES - $34 BYTES
TUNEAREA		EQU	$700
TUNESOFF		EQU	TUNEAREA+$0000				;1 BYTE  - FLAG FOR ALL TUNES OFF
TUNNUM			EQU	TUNEAREA+$0001				;1 BYTE  - CURRENT TUNE NUMBER BEING PROCESSED
TUNCHANNEL		EQU	TUNEAREA+$0002				;1 BYTE  - CONVERTED FOR INDEX INTO POKEY CHANNEL REGISTERS
PALCNT			EQU	TUNEAREA+$0003				;1 BYTE  - PAL index
TUNON			EQU	TUNEAREA+$0004				;4 BYTES - FLAG FOR TUNE PLAYING BY CHANNEL
NOTELO			EQU	TUNEAREA+$0008				;4 BYTES - INDEX INTO NOTE TABLE LOW
NOTEHI			EQU	TUNEAREA+$000C				;4 BYTES - INDEX INTO NOTE TABLE HI
CTLVOL			EQU	TUNEAREA+$0010				;4 BYTES - CONTROL / VOLUME VALUE BY CHANNEL
TUNINDEX		EQU	TUNEAREA+$0014				;4 BYTES - TUNE NUMBER BY CHANNEL
TUNFRM			EQU	TUNEAREA+$0018				;4 BYTES - NUMBER OF FRAMES (DURATION COUNT)
TUNPRIOR		EQU	TUNEAREA+$001C				;4 BYTES - TUNE PRIORITY BY CHANNEL
DURNLO			EQU	TUNEAREA+$0020				;4 BYTES - INDEX INTO DURATION TABLE LOW
DURNHI			EQU	TUNEAREA+$0024				;4 BYTES - INDEX INTO DURATION TABLE HI
DCYSTOR			EQU	TUNEAREA+$0028				;4 BYTES - FOR NOTE DECAY BY CHANNEL
FREQCNT			EQU	TUNEAREA+$002C				;4 BYTES - FOR NOTE DECAY BY CHANNEL
CTLSAV			EQU	TUNEAREA+$0030				;4 BYTES - TO SAVE THE CONTROL VALUE
MUTEMASK		EQU	TUNEAREA+$0036				;4 BYTES - FLAG FOR MUTEMASK

SOUNDZP equ $fe

;	SOUND -------------------------------------------------------------------------------------------------------------------->

;	TUNES - THESE ROUTINES HANDLE ALL OF THE SOUNDS

;	TURN OFF ALL SOUNDS
STOPTUN
	LDA #$00
	STA AUDC1
	STA AUDC2
	STA AUDC3
	STA AUDC4
	LDA #$01
	STA TUNESOFF
	RTS

;	TURN ON ALL SOUNDS
STARTTUN
	LDA #$00
	STA TUNESOFF
	RTS

;	THIS ROUTINE ERASES ALL TUNES
CLEARTUN
	LDX #$03
CTLOOP
	JSR ENDTUNE									;ERASE CURRENT TUNE
	DEX
	BPL CTLOOP
	RTS

;	ROUTINE TO KILL A PARTICULAR TUNE - IF IT IS RUNNING
;	INPUT: TUNE NUMBER IN A
KILLTUNE
	STA TUNNUM									;SAVE IT
	LDX #$03									;CHECK ALL CHANNELS
KTLOOP
	LDA TUNON,X									;SEE IF CHANNEL ON
	BEQ KTNEXT
	LDA TUNINDEX,X								;SEE IF HAS TUNE TO BE KILLED
	CMP TUNNUM
	BNE KTNEXT
	JSR ENDTUNE									;ERASE IT
KTNEXT
	DEX
	BPL KTLOOP
	RTS

;	THIS ROUTINE CLEARS OUT A TUNE CHANNEL
;	INPUT: X IS CHANNEL
ENDTUNE
	LDA #$00
	STA TUNON,X									;INDICATE CHANNEL CLEAR
	STA TUNINDEX,X								;CLEAR TUNE INDEX
	STA DCYSTOR,X
	STA FREQCNT,X
	RTS

;	THIS ROUTINE ENTERS A TUNE INTO ONE OF THE SOUND CHANNELS IF IT CAN
;	INPUT:  TUNE NUMBER IN A
DOTUNE
	STA TUNNUM									;SAVE IT
	LDY TUNNUM									;SEE IF WE CAN PUT IT IN
	LDX CHANNLTBL,Y								;GET WHAT CHANNEL TO TRY TO PUT IT IN
	LDA TUNON,X									;SEE IF CHANNEL OPEN
	BEQ DTDOIT
	LDA PRIRTYTBL,Y								;SEE IF WE CAN BUMP CHANNEL
	CMP TUNPRIOR,X
	BMI DTOUT
DTDOIT
	LDA TUNNUM
	TAY											;PUT TUNE IN Y
	STA TUNINDEX,X								;SET THE TUNE INDEX
	LDA #$00									;TURN TUNE OFF WHILE CHANGING IT
	STA TUNON,X
	LDA CNTVOLTBL,Y								;GET TUNE CONTROL / VOLUME
	STA CTLVOL,X
	STA CTLSAV,X								;USED TO RESTORE AFTER DECAY
	LDA NOTETBLLO,Y								;GET TUNE FREQUENCY LOW ADDRESS
	STA NOTELO,X
	LDA NOTETBLHI,Y								;GET TUNE FREQUENCY HIGH ADDRESS
	STA NOTEHI,X
	LDA DURNTBLLO,Y								;GET TUNE DURATION LOW ADDRESS
	STA DURNLO,X
	LDA DURNTBLHI,Y								;GET TUNE DURATION HIGH ADDRESS
	STA DURNHI,X
	LDA PRIRTYTBL,Y								;SET PRIORITY
	STA TUNPRIOR,X
	LDA #$01									;SET FREQ, CTL, AND VOL TO BE SET
	STA TUNFRM,X
	STA TUNON,X									;AND TURN THE TUNE ON!
DTOUT
	RTS

;  THIS ROUTINE IS CALLED EVERY VBLANK TO TAKE CARE OF TUNES
;  REGISTERS ARE NOT SAVED
TUNER
	LDX #$03									;FOUR TUNES CHANNELS, START WITH LAST
	LDA TUNESOFF
	BEQ TUNLOOP
	RTS
TUNLOOP
	TXA
	ASL
	STA TUNCHANNEL								;CHANNELS ARE OFFSET 0, 2, 4, 6 IN THE POKEY
	TAY
	LDA TUNON,X
	BNE TUNBODY
	STA AUDC1,Y									;CHANNEL OFF - MAKE SURE VOLUME OFF
	JMP TUNNEXT
TUNBODY
	DEC TUNFRM,X								;SEE IF WE'RE DONE WITH THIS SOUND
	BEQ TUNFRMFRQ								;YES - GET NEXT NOTE / DURATION
	DEC FREQCNT,X                               ;REDUCE THE NUMBER OF FRAMES UNTIL NEXT DECAY
	BEQ DEC_VOLUME
	JMP TUNNEXT                                 ;IF WE AREN'T AT ZERO YET, DON'T DECAY

DEC_VOLUME
	LDA DCYSTOR,X                               ;RESET THE DECAY FOR THE NEXT COUNT
	STA FREQCNT,X

	LDA CTLVOL,X                                ;IF VOLUME ALREADY 0 DO NOT DECREMENT
	AND #$0F
	BEQ TUNNEXT
	
	DEC CTLVOL,X                                ;DECREMENT THE VOLUME
	LDA CTLVOL,X
	LDY TUNCHANNEL
	AND MUTEMASK,X
	STA AUDC1,Y
	JMP TUNNEXT                                 ;GO TO NEXT CHANNEL
TUNFRMFRQ
	LDA DURNLO,X							;GET THE CURRENT DURATION
	STA SOUNDZP
	LDA DURNHI,X
	STA SOUNDZP+1
	LDY #$00
	LDA (SOUNDZP),Y
	BEQ TUNEND								;$00 IN DURATION MEANS TUNE IS OVER
	STA TUNFRM,X
	TAY
	LDA DECAYTBL,Y							;GET THE CURRENT DECAY VALUE INDEXED BY NOTE
	STA DCYSTOR,X							;STORE IT HERE TO REFRESH THE COUNTER FOR THE NEXT DECAY
	STA FREQCNT,X							;ALSO STORE IT HERE FOR TUNER
	LDY TUNCHANNEL
	LDA CTLSAV,X
	STA CTLVOL,X							;RESTORE THE ORIGINAL CONTROL AND VOLUME FOR NEXT NOTE
 	AND #$0f 								;REVENG - STATIC REMOVAL - ensure 0 write is skipped
 	BEQ SKIPWRITEZERO
	LDA CTLSAV,X

 

	STA AUDC1,Y
SKIPWRITEZERO
	LDA NOTELO,X							;GET THE CURRENT FREQUENCY
	STA SOUNDZP
	LDA NOTEHI,X
	STA SOUNDZP+1

 
	LDA #$ff
	STA MUTEMASK,X


	LDY #$00
	LDA (SOUNDZP),Y
	LDY TUNCHANNEL

 
        ; if you want to use 254 as a REST indicator, change the "CMP #0" to "CMP #254"
	CMP #0									;+ MUTEMASK KLUDGE
	BNE SKIPFLAGREST
	LDA #$f0
	STA MUTEMASK,X
	AND CTLSAV,X 
	STA AUDC1,Y 
	
	LDA #$00
SKIPFLAGREST
 

	STA AUDF1,Y
	INC NOTELO,X
	BNE TUNNEXTNOTE
	INC NOTEHI,X
TUNNEXTNOTE
	INC DURNLO,X
	BNE TUNNEXT
	INC DURNHI,X
TUNNEXT
	DEX
	CPX #$00
	BMI TUNEXIT
	JMP TUNLOOP
TUNEXIT

	RTS
	

TUNEND
	LDA NOTELO,X								;SEE IF WE SHOULD REPEAT
	STA SOUNDZP
	LDA NOTEHI,X
	STA SOUNDZP+1
	LDY #$00
	LDA (SOUNDZP),Y
	BMI TUNRESTART
	JSR ENDTUNE
	JMP TUNNEXT
TUNRESTART
	LDA TUNINDEX,X								;GET TUNE NUMBER
	TAY
	LDA CNTVOLTBL,Y								;GET TUNE CONTROL / VOLUME
	STA CTLVOL,X
	STA CTLSAV,X								;USED TO RESTORE AFTER DECAY
	LDA NOTETBLLO,Y								;GET TUNE FREQUENCY LOW ADDRESS
	STA NOTELO,X
	LDA NOTETBLHI,Y								;GET TUNE FREQUENCY HIGH ADDRESS
	STA NOTEHI,X
	LDA DURNTBLLO,Y								;GET TUNE DURATION LOW ADDRESS
	STA DURNLO,X
	LDA DURNTBLHI,Y								;GET TUNE DURATION HIGH ADDRESS
	STA DURNHI,X
	LDY TUNCHANNEL
	LDA CTLVOL,X
	STA AUDC1,Y									;STORE THE CONTROL / VOLUME IN THE CHANNEL
	LDA #$01									;SET FREQ, CTL, AND VOL TO BE SET
	STA TUNFRM,X
	JMP TUNNEXT
	
;	DATA FOR TUNES

;	CNTVOLTBL - HIGH NYBBLE FOR VOICE CONTROL (7654XXXX), AND LOW NYBBLE FOR VOLUME (XXXX3210)
CNTVOLTBL:
	.byte $a7,$a7,$a0,$a7		 	;$A0 on channel 2 to silence channel
    .byte $46,$00,$46,$00
	.byte $46,$00,$46,$00
	.byte $86,$00,$86,$00
	.byte $06,$00,$06,$00
	.byte $26,$00,$A6,$A4
	.byte $46,$00
	.byte $a7,$a7,$a0,$a7


;	PRIORITY TABLE
PRIRTYTBL:
	.byte $05,$05,$05,$05
    .byte $02,$02,$02,$02
	.byte $02,$02,$02,$02	
	.byte $02,$02,$02,$02
	.byte $03,$03,$02,$02
	.byte $02,$02,$02,$02
	.byte $02,$02
	.byte $05,$05,$05,$05

	
;	CHANNEL TABLE ($00 - $03 ARE VALID)
CHANNLTBL:
	.byte $00,$01,$02,$03			; must set $00 and $01 to use 16-bit combined channel
    .byte $00,$01,$00,$01
	.byte $00,$01,$00,$01
	.byte $00,$01,$00,$01
	.byte $00,$01,$00,$01
	.byte $00,$01,$02,$03
	.byte $00,$01
	.byte $00,$01,$02,$03


;	DECAY TABLE - DEFINES DECAY COUNT INDEXED BY NOTE DURATION
DECAYTBL:
	.byte $01,$01,$01,$01,$02,$02,$02,$02
	.byte $02,$03,$03,$03,$03,$03,$03,$03
	.byte $04,$04,$04,$04,$04,$04,$04,$04
	.byte $05,$05,$05,$05,$05,$05,$05,$05
	.byte $05,$05,$06,$06,$06,$06,$06,$06
	.byte $06,$06,$06,$06,$07,$07,$07,$07
	.byte $07,$07,$07,$07,$07,$07,$07,$07
	.byte $07,$08,$08,$08,$08,$08,$08,$08

;	FREQUENCY TABLE
NOTETBLHI:
	.byte >(TUNE00),>(TUNE01),>(TUNE02),>(TUNE03)
	.byte >(TUNE04),>(TUNE05),>(TUNE06),>(TUNE07)
	.byte >(TUNE08),>(TUNE09),>(TUNE0A),>(TUNE0B)
    .byte >(TUNE0C),>(TUNE0D),>(TUNE0E),>(TUNE0F)
	.byte >(TUNE10),>(TUNE11),>(TUNE12),>(TUNE13)
	.byte >(TUNE14),>(TUNE15),>(TUNE16),>(TUNE17)
	.byte >(TUNE18),>(TUNE19)
	.byte >(TUNE1A),>(TUNE1B),>(TUNE1C),>(TUNE1D)

NOTETBLLO:
	.byte <(TUNE00),<(TUNE01),<(TUNE02),<(TUNE03)
	.byte <(TUNE04),<(TUNE05),<(TUNE06),<(TUNE07)
	.byte <(TUNE08),<(TUNE09),<(TUNE0A),<(TUNE0B)
    .byte <(TUNE0C),<(TUNE0D),<(TUNE0E),<(TUNE0F)
	.byte <(TUNE10),<(TUNE11),<(TUNE12),<(TUNE13)
	.byte <(TUNE14),<(TUNE15),<(TUNE16),<(TUNE17)
	.byte <(TUNE18),<(TUNE19)
	.byte <(TUNE1A),<(TUNE1B),<(TUNE1C),<(TUNE1D)

;	DURATION TABLE
DURNTBLHI:
	.byte >(DURN00),>(DURN01),>(DURN02),>(DURN03)
	.byte >(DURN04),>(DURN05),>(DURN06),>(DURN07)
	.byte >(DURN08),>(DURN09),>(DURN0A),>(DURN0B)
    .byte >(DURN0C),>(DURN0D),>(DURN0E),>(DURN0F)
	.byte >(DURN10),>(DURN11),>(DURN12),>(DURN13)
	.byte >(DURN14),>(DURN15),>(DURN16),>(DURN17)
	.byte >(DURN18),>(DURN19)
	.byte >(DURN1A),>(DURN1B),>(DURN1C),>(DURN1D)

DURNTBLLO:
	.byte <(DURN00),<(DURN01),<(DURN02),<(DURN03)
	.byte <(DURN04),<(DURN05),<(DURN06),<(DURN07)
	.byte <(DURN08),<(DURN09),<(DURN0A),<(DURN0B)
    .byte <(DURN0C),<(DURN0D),<(DURN0E),<(DURN0F)
	.byte <(DURN10),<(DURN11),<(DURN12),<(DURN13)
	.byte <(DURN14),<(DURN15),<(DURN16),<(DURN17)
	.byte <(DURN18),<(DURN19)
	.byte <(DURN1A),<(DURN1B),<(DURN1C),<(DURN1D)

;TEMPOS
T1	=	$03
T2      =       (T1*2)
T3      =       (T1*3)
T4      =       (T1*4)
T5      =       (T1*5)
T6      =       (T1*6)
T7      =       (T1*7)
T8      =       (T1*8)
T9      =       (T1*9)
TA      =       (T1*10)
TB      =       (T1*11)
TC      =       (T1*12)
TD      =       (T1*13)
TE      =       (T1*14)
TF      =       (T1*15)
T0      =       (T1*16)

;A	16264	3F88			
;A#	15351	3BF7			
;B	14489	3899			
;C	13675	356B				OCTAVE 2
;C#	12907	326B			
;D	12182	2F96			
;D#	11498	2CEA			
;E	10852	2A64	
;
;F	10243	2803			
;F#	9668	25C4			
;G	9125	23A5			
;G#	8612	21A4			
;A	8128	1FC0			
;A#	7672	1DF8			
;B	7241	1C49			
;C	6834	1AB2	243	F3		OCTAVE 3
;C#	6450	1932	230	E6	
;D	6088	17C8	217	D9	
;D#	5746	1672	204	CC	
;E	5423	152F	193	C1	
;F	5118	13FE	182	B6	
;F#	4830	12DE	172	AC	
;G	4559	11CF	162	A2	
;G#	4303	10CF	153	99	
;A	4061	0FDD	144	90	

TUNE00:

	.byte $2e,$36,$43,$45,$2e,$36,$43,$2e,$36,$43,$45,$2e,$36,$43,$45
	
	.byte 0
	
	.byte $2a,$31,$39,$3b,$2a,$31,$39,$2a,$31,$39,$3b,$2a,$31,$39,$3b
	
	
	.byte 0

	
	.byte $26,$2e,$49,$26,$2e,$49,$26,$2e,$49,$26,$2e,$49 
	
	
	.byte $49,$4d,$56,$34,$73,$6c,$73,$6c,$4c,0
	.byte $31,$73,$34,$48,$54,$26,$54,$26,$57,0
TUNE1A:	
	.byte $26,$57,$26,$4c,$6c,$48,$54
	.byte $34,$56,$6c,$73,$50,0,$56,$26,$54,$4d,$2e
	.byte $62
	
	
	
	.byte 0
	
DURN00:
	
	
	
	.byte 10,10,10,30,10,10,10,10,10,10,30,10,10,10,60
	
	.byte 30
	
	.byte 10,10,10,30,10,10,10,10,10,10,30,10,10,10,60
	
	
	.byte 30
	
	.byte 10,10,10,10,10,10,10,10,10,10,10,10
	
	.byte 30,20,10,30,30,10,10,10,30,60
	.byte 30,20,10,30,30,10,10,10,30,60
DURN1A:
	.byte 30,20,10,20,20,10,10
	.byte 10,10,10,20,10,10,10,10,10,10,10
	
	.byte 60

	
	.byte 0
	

;
;	TUNE 1 - COMBINED 16-BIT CHANNEL (High order bytes stored in AUDF1)

TUNE01:

	
	
	.byte 10,14,22,23,10,14,22,10,14,22,23,10,14,22,23
	
	.byte 162
	
	.byte 8,11,16,17,8,11,16,8,11,16,17,8,11,16,17
	
	.byte 162
	
	

	
	
	.byte 17,15,17,15,17,15,17,15
	
	.byte 79,79,75,75,70,70,79,150
	.byte 75,75,70,70,66,66,75,142
TUNE1B:
	.byte 64, 47,50, 53, 60, 64
	
	.byte 68
	
	.byte 0
	

	
	
	
	
;A	16264	3F88			
;A#	15351	3BF7			
;B	14489	3899			
;C	13675	356B				OCTAVE 2
;C#	12907	326B			
;D	12182	2F96			
;D#	11498	2CEA			
;E	10852	2A64	
;F	10243	2803			
;F#	9668	25C4			
;G	9125	23A5			
;G#	8612	21A4			
;A	8128	1FC0			
;A#	7672	1DF8			
;B	7241	1C49			
;C	6834	1AB2	243	F3		OCTAVE 3
;C#	6450	1932	230	E6	
;D	6088	17C8	217	D9	
;D#	5746	1672	204	CC	
;E	5423	152F	193	C1	
;F	5118	13FE	182	B6	
;F#	4830	12DE	172	AC	
;G	4559	11CF	162	A2	
;G#	4303	10CF	153	99	
;A	4061	0FDD	144	90	
DURN01: ;this table MUST be identical to DURN00:
	
	
	.byte 10,10,10,30,10,10,10,10,10,10,30,10,10,10,30
	
	.byte 60
	
	.byte 10,10,10,30,10,10,10,10,10,10,30,10,10,10,30
	
	
	.byte 60
	
	.byte 15,15,15,15,15,15,15,15
	
	.byte 30,30,30,30,30,30,30,30
	.byte 30,30,30,30,30,30,30,30
DURN1B:
	.byte 60, 50,10, 60, 30, 30
	
	.byte 60
	
	.byte 0
	
	
		
	
	
	
;
;	TUNE 2 - NORMAL 8-BIT CHANNEL

TUNE02:
	.byte $2f,$37,$44,$46,$2f,$37,$44,$2f,$37,$44,$46,$2f,$37,$44,$46
	
	.byte 0
	
	.byte $2b,$32,$3a,$3c,$2b,$32,$3a,$2b,$32,$3a,$3c,$2b,$32,$3a,$3c
	
	
	.byte 0
	
	.byte $27,$2f,$4b,$27,$2f,$4b,$27,$2f,$4b,$27,$2f,$4b
	
	.byte $4b,$50,$5b,$36,$7a,$73,$7a,$73,$51,0
	.byte $32,$7a,$36,$4c,$58,$27,$58,$27,$5e,0
TUNE1C:
	.byte $27,$5e,$27,$51,$73,$4c,$58
	.byte $36,$5b,$73,$7a,$53,0,$5b,$27,$58,$50,$2f
	.byte $66
	
	.byte 0
	

DURN02:
	
	.byte 10,10,10,30,10,10,10,10,10,10,30,10,10,10,60
	
	.byte 30
	
	.byte 10,10,10,30,10,10,10,10,10,10,30,10,10,10,60
	
	
	.byte 30
	
	.byte 10,10,10,10,10,10,10,10,10,10,10,10
	
	.byte 30,20,10,30,30,10,10,10,30,60
	.byte 30,20,10,30,30,10,10,10,30,60
DURN1C:
	.byte 30,20,10,20,20,10,10
	.byte 10,10,10,20,10,10,10,10,10,10,10
	
	.byte 60
	

	.byte 0
	
	
	

;
;	TUNE 3 - NORMAL 8-BIT CHANNEL
TUNE03:

	.byte 0,0
	
	.byte 136
	
	.byte 0,0
	
	.byte 136
	
	.byte 84,84,84,84
	
	.byte 52,52,52,52,52,52,49,49,49,49,49,49,47,47,47,47,47,47
	.byte 22,29,39,59
	
	.byte 49,49,49,49,49,49,47,47,47,47,47,47,44,44,44,44,44,44
	.byte 21,27,37,56
TUNE1D:
	.byte 45, 33,35, 37, 42, 45
	
	.byte 47
	
	.byte 0
	
	
	
	
		;A	4061	0FDD	144	90	
;A#	3832	0EF8	136	88	
;B	3617	0E21	128	80	
;C	3414	0D56	121	79		OCTAVE 4
;C#	3222	0C96	114	72	
;D	3040	0BE0	108	6C	
;D#	2869	0B35	102	66	
;E	2708	0A94	96	60	
;F	2555	09FB	91	5B	
;F#	2412	096C	85	55	
;G	2276	08E4	81	51	
;G#	2148	0864	76	4C	
;A	2027	07EB	72	48	
;A#	1913	0779	68	44	
;B	1805	070D	64	40	
;C	1703	06A7	60	3C		OCTAVE 5
;C	1703	06A7	60	3C		OCTAVE 5
;C#	1607	0647	57	39	
;D	1517	05ED	53	35	
	
	
DURN03:
	
	.byte 105,105
	
	.byte 60
	
	.byte 105,105
	
	.byte 60
	
	.byte 30,30,30,30
	
	.byte 10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10
	.byte 10,10,10,30
	
	.byte 10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10
	.byte 10,10,10,30
DURN1D:
	.byte 60, 50,10, 60, 30, 30
	
	.byte 60
	
	.byte 0

;guard hit

TUNE04:

	.byte 89
	
	
	
	.byte 0
	
DURN04:
	
	
	
	.byte 47

	
	.byte 0
	


TUNE05:

	
	
	.byte 62
	
	
	
	.byte 0
	

	
	

DURN05: ;this table MUST be identical to DURN00:
	
	
	.byte 47

	
	.byte 0
	
	
	
	
		
	
	
	

; door open
TUNE06:
	.byte 27
	
	
	.byte 0
	

DURN06:
	
	.byte 47
	
	
	

	.byte 0
	
	
	

;
;	TUNE 3 - NORMAL 8-BIT CHANNEL
TUNE07:

	.byte 33
	
	
	
	.byte 0
	
	
	
	

	
DURN07:
	
	.byte 47
	

	
	.byte 0
	
;Disk shatter

TUNE08:   
	.byte 89,0	
	
	
DURN08:
	.byte 47,0
	
TUNE09:
	.byte 202,0
	
	
	
	
	
	
	
	


DURN09:
	.byte 47,0
	
	
	
	

;Player death

	
TUNE0A:
	.byte 58,0
	
	
	
	
DURN0A:
    .byte 47,0
	
	
	


TUNE0B:
   .byte 67,0
	

DURN0B: ;this table MUST be identical to DURN00:
	
	

    
    .byte 47,0

;Door Close 
TUNE0C:
	.byte 142,0
	
	
	
	
	
DURN0C:
	.byte 47,0
	

TUNE0D:
	
	.byte 99,0
	
	
	
DURN0D: 
	
	
	.byte 47,0
	
	
; Recognizer spark
TUNE0E:
	.byte 69,0
	


DURN0E:
	
	.byte 47,0
	
	

TUNE0F:   
	.byte 255,0
			

	
DURN0F:
	.byte 47
	
	.byte 0

;Recognizer Destruction

TUNE10:
	.byte 27,58,89,120,151,182,213,244,213,182,151,120,89,58
	
	
	.byte -1
	
DURN10:
	.byte 2,2,2,2,2,2,2,2,2,2,2,2,2,2
	.byte 0

;
;	TUNE 1 - COMBINED 16-BIT CHANNEL (High order bytes stored in AUDF1)
TUNE11:
	
	.byte 240
	
	
	
	.byte -1

DURN11: ;this table MUST be identical to DURN00:
	
	.byte 28
	.byte 0

; Disc hit
TUNE12:
	.byte 213
	
	
	
	.byte 0


DURN12:
	
	.byte 47
	
	.byte 0

;
;	TUNE 3 - NORMAL 8-BIT CHANNEL
TUNE13:   
	.byte 192,0

	
DURN13:
	.byte 47
	
	.byte 0

;Disc bounce


TUNE14:
	.byte 231,0
	
DURN14:
	.byte 47,0


TUNE15:
	.byte 191,0
DURN15:
	.byte 47,0

;Recognizer Hum

TUNE16:
	.byte 137,142,147,152,157,162,167,162,157,152,147,142
	.byte -1
DURN16:
	.byte 2,2,2,2,2,2,2,2,2,2,2,2
	.byte 0

TUNE17:
	.byte 4,-1
	
	
	
	
DURN17:
	.byte 24
	
	.byte 0

;disc throw

TUNE18:
	
	.byte 56,56,71,71,86,86,101,101,116,116
	.byte 131,131,146,146,161,161,176,176,191,191
	.byte 206,206,221,221,236,236,251,251
	.byte 0
	
DURN18: ;this table MUST be identical to DURN00:
	
	.byte 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
	.byte 1,1,1,1,1,1,1,1
	
	.byte 0



TUNE19:

	.byte 10,35,12,41,15,16,53,18,62,21
	.byte 22,71,25,27,28,30,31,33,36,37
	.byte 40,43,45,48,51,55,57,61
	
	.byte 0
DURN19:
	
	.byte 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
	.byte 1,1,1,1,1,1,1,1
	
	.byte 0

