; This file is part of Intellidiscs.

; Intellidiscs is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; Intellidiscs is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with Intellidiscs.  If not, see <http://www.gnu.org/licenses/>.

; Copyright 2021 Phaser Cat Games LLC

; Title menu code

; Constants.
title_table_duration        equ 3
title_table_length          equ 13
title_key_cooldown_length   equ 15

title_color_table       equ $80     ; Title color shading table.
title_table_fill        equ $a0     ; Title fill direction.
title_table_time        equ $a1     ; Time until table change.
title_animation_index   equ $a2
title_animation_timer   equ $a3
title_control_mode      equ $a4
title_recognizer_style  equ $a8

; Page 2 data shared between states.
high_score_zone         equ $230
low_score_zone          equ $258
last_score_zone         equ $280
player_count            equ $2fa
control_modes           equ $2fb
recognizer_body_style   equ $2ff

; Title screen entry point.
start_title_screen subroutine
    jsr     start_transition
    mwa     title_display_list,sDLISTL
    mwa     title_vblank,VVBLKD
    mwa     title_interrupt,VDSLST
    jsr     safe_clear_page_0
    jsr     clear_sprites
    mva     #$f8,CHBASE             ; Normal character set.
    lda     #$ff
    sta     keypad_last_key         ; Clear player 1 input.
    lda     #0
    ldy     #0
.memory_clear_loop
    sta     $3000,y
    sta     $3100,y
    sta     $3200,y
    dey
    bne     .memory_clear_loop
    sta     sCOLBK
    sta     sCOLOR2
    sta     SIZEP1
    sta     SIZEP2
    sta     title_color_table + title_table_length
    lda     #$0e
    sta     sCOLOR0
    sta     sCOLOR1
    lda     #player_1_color
    sta     sCOLPM0
    lda     #drone_color
    sta     sCOLPM1
    lda     #leader_color
    sta     sCOLPM2
    lda     #-2                     ; Will reach 0 on first
    sta     title_animation_index   ; time through the loop.
    ldx     #title_table_length - 1
.title_table_loop
    lda     title_table,x
    sta     title_color_table,x
    dex
    bpl     .title_table_loop
    lda     #title_table_duration
    sta     title_table_time
    lda     #1
    sta     title_animation_timer
    sta     title_table_fill        ; Fill direction up.
    ldy     #2
.disc_loop
    lda     #$03
    sta     $1368 + held_disc_y_delta,y
    lda     #$3c
    sta     $1380 + held_disc_y_delta,y
    sta     $1398 + held_disc_y_delta,y
    lda     #$30
    sta     $13b0 + held_disc_y_delta,y
    dey
    bpl     .disc_loop
    ldy     #7
.stick_loop
    lda     #$04
    ora     $13b3,y
    sta     $13b3,y
    dey
    bpl     .stick_loop
    mwa     copyright_text,$fc
    mwa     $32aa,$fe
    jsr     copy_text_full
    mva     #$65,AUDCTL
    mva     #2,SKCTL
    lda     #0
    jsr     DOTUNE
    lda     #1
    jsr     DOTUNE
    lda     #2
    jsr     DOTUNE
    lda     #3
    jsr     DOTUNE
    jsr     STARTTUN
    jsr     complete_transition
    ; Fall through to the main loop.

; Main title processing loop.
show_main_title subroutine
    jsr     clear_text_area
    mwa     hero_text,$fc
    mwa     $3089,$fe
    jsr     copy_text_full
    mwa     drone_text,$fc
    mwa     $30f8,$fe
    jsr     copy_text_full
    mwa     armored_text,$fc
    mwa     $3170,$fe
    jsr     copy_text_full
    mwa     guard_text,$fc
    mwa     $31e8,$fe
    jsr     copy_text_full
    mwa     leader_text,$fc
    mwa     $310c,$fe
    jsr     copy_text_full
    mwa     hunter_text,$fc
    mwa     $3184,$fe
    jsr     copy_text_full
    mwa     seeker_text,$fc
    mwa     $31fc,$fe
    jsr     copy_text_full
    mwa     press_start_text,$fc
    mwa     $3262,$fe
    jsr     copy_text_full
    mwa     press_for_license_text,$fc
    mwa     $3288,$fe
    jsr     copy_text_full
    jsr     position_title_sprites
    lda     #1
    sta     player_count
main_title_loop
    jsr     wait_for_vb
    jsr     keypad_process
    lda     keypad_last_key
    cmp     #KEYOCTOTHORPE
    bne     .not_octothorpe
    jmp     show_license
.not_octothorpe
    cmp     #KEYSTART
    bne     .not_start
    jmp     show_game_selection
.not_start
    dec     title_animation_timer
    bne     main_title_loop
    lda     #player_frame_time
    sta     title_animation_timer
    ldy     title_animation_index
    cpy     #player_right_last_frame
    bne     .increment_player_index
    ldy     #0
    jmp     .store_player_index
.increment_player_index
    iny
    iny
.store_player_index
    sty     title_animation_index
    lda     player_right_cycle,y
    sta     $fc
    lda     player_right_cycle + 1,y
    sta     $fd
    ldy     #player_sprite_height - 1
.sprite_update_loop
    lda     ($fc),y
    sta     $1468,y
    sta     $1580,y
    sta     $1680,y
    sta     $1598,y
    sta     $1698,y
    sta     $15b0,y
    sta     $16b0,y
    dey
    bpl     .sprite_update_loop
    jmp     main_title_loop

; License screen setup and loop.
show_license subroutine
    lda     #$ff
    sta     keypad_last_key
    jsr     hide_title_sprites
    jsr     wait_for_vb
    mwa     license_display_list,sDLISTL
.license_loop
    jsr     wait_for_vb
    jsr     keypad_process
    lda     keypad_last_key
    cmp     #KEYOCTOTHORPE
    bne     .license_loop
    mwa     title_display_list,sDLISTL
    jsr     position_title_sprites
    lda     #$ff
    sta     keypad_last_key
    jmp     main_title_loop

; Game selection mode setup and loop.
show_game_selection subroutine
    lda     #$ff
    sta     keypad_last_key
    jsr     hide_title_sprites
    jsr     clear_text_area
    mwa     select_game_mode_text,$fc
    mwa     $3083,$fe
    jsr     copy_text_full
    mwa     one_player_text,$fc
    mwa     $30d2,$fe
    jsr     copy_text_full
    mwa     battle_player_text,$fc
    mwa     $30fa,$fe
    jsr     copy_text_full
    mwa     $3122,$fe
    jsr     copy_text_full
    mva     #$4a,$fe
    jsr     copy_text_full
    inc     $3122
    inc     $3126
    lda     #$14
    sta     $314a
    sta     $314e
.game_selection_loop
    jsr     wait_for_vb
    jsr     keypad_process
    lda     keypad_last_key
    cmp     #KEYRESET
    bne     .not_reset
    jmp     show_main_title
.not_reset
    cmp     #1
    bcc     .game_selection_loop
    cmp     #5
    bcs     .game_selection_loop
    sta     player_count
    tax
    lda     #$ff
    sta     keypad_last_key
    jsr     clear_text_area
    mwa     control_mode_text,$fc
    mwa     $3085,$fe
    jsr     copy_text_full
    mwa     press_start_text,$fc
    mwa     $3262,$fe
    jsr     copy_text_full
    mwa     control_keypad_text,$fc ; This is somewhat redundant, but
    mwa     $30d5,$fe               ; it puts some of the needed menu
    jsr     copy_text_full          ; elements in place.
    cpx     #1
    beq     show_arcade_config
    jmp     show_battle_config

; Arcade mode configuration.
show_arcade_config subroutine
    mwa     identifier_style_text,$fc
    mwa     $3121,$fe
    jsr     copy_text_full
    mwa     identifier_new_text,$fc ; Redundant, but needed for
    mwa     $3175,$fe               ; certain meny elements.
    jsr     copy_text_full
    lda     control_modes
    sta     title_control_mode
    lda     recognizer_body_style
    sta     title_recognizer_style
.update_selections
    lda     title_control_mode
    beq     .control_keypad
    cmp     #1
    beq     .control_trigger
    mwa     control_twin_text,$fc
    mwa     $30d9,$fe
    jmp     .update_control_option
.control_keypad
    mwa     control_keypad_text,$fc
    mwa     $30d5,$fe
    jmp     .update_control_option
.control_trigger
    mwa     control_trigger_text,$fc
    mwa     $30d9,$fe
.update_control_option
    jsr     copy_text_full
    lda     title_recognizer_style
    bne     .recognizer_classic
    mwa     identifier_new_text,$fc
    mwa     $3175,$fe
    jmp     .update_recognizer_option
.recognizer_classic
    mwa     identifier_classic_text,$fc
    mwa     $3179,$fe
.update_recognizer_option
    jsr     copy_text_full
    lda     #$ff
    sta     keypad_last_key
.arcade_config_loop
    jsr     wait_for_vb
    jsr     keypad_process
    lda     keypad_last_key
    cmp     #KEYRESET
    bne     .not_reset
    jmp     show_main_title
.not_reset
    cmp     #KEYSTAR
    beq     .change_control
    cmp     #KEYOCTOTHORPE
    beq     .change_recognizer
    cmp     #KEYSTART
    bne     .arcade_config_loop
    lda     title_control_mode
    sta     control_modes
    lda     title_recognizer_style
    sta     recognizer_body_style
    jmp     start_game
.change_control
    lda     title_control_mode
    cmp     #2
    beq     .control_loop
    inc     title_control_mode
    jmp     .update_selections
.control_loop
    lda     #0
    sta     title_control_mode
    jmp     .update_selections
.change_recognizer
    lda     title_recognizer_style
    eor     #1
    sta     title_recognizer_style
    jmp     .update_selections

; Battle mode configuration.  Player count is in x.
show_battle_config subroutine
    mwa     control_keypad_text,$fc ; Set everyone to keypad at
    ldy     #$11                    ; first to get all the menu
    sty     $30d2                   ; text on the screen.
    iny
    sty     $30fa
    lda     #$30
    sta     $30d3
    sta     $30fb
    mva     #$fd,$fe
    jsr     copy_text_full
    cpx     #3
    bmi     .menu_done
    lda     #$13
    sta     $3122
    lda     #$30
    sta     $3123
    mwa     $3125,$fe
    jsr     copy_text_full
    cpx     #4
    bne     .menu_done
    lda     #$14
    sta     $314a
    lda     #$30
    sta     $314b
    mva     #$4d,$fe
    jsr     copy_text_full
.menu_done
    lda     control_modes           ; No twin-stick for player 1.
    cmp     #2
    bne     .not_twin_stick
    lda     #0                      ; Fall back to keypad.
.not_twin_stick
    sta     title_control_mode
    ldy     #3
.old_control_loop
    lda     control_modes,y
    sta     title_control_mode,y
    dey
    bne     .old_control_loop
    jsr     keypad_reset
.update_selections
    mwa     $30d9,$fe
    ldy     #0
.update_selection_loop
    lda     title_control_mode,y
    beq     .mode_keypad
    mwa     control_trigger_text,$fc
    jmp     .apply_text
.mode_keypad
    mwa     [control_keypad_text + 4],$fc
.apply_text
    sty     $f0
    jsr     copy_text_full
    ldy     $f0
    iny
    cpy     player_count
    bpl     .battle_config_loop
    lda     $fe
    clc
    adc     #40
    sta     $fe
    bcc     .update_selection_loop
    inc     $ff
    jmp     .update_selection_loop
.battle_config_loop
    jsr     wait_for_vb
    jsr     keypad_process
    lda     keypad_last_key
    cmp     #KEYRESET
    bne     .not_reset
    jmp     show_main_title
.not_reset
    cmp     #KEYSTART
    beq     .start_game
    ldy     keypad_active_player
    cpy     player_count
    bpl     .battle_config_loop
    lda     keypad_last_key,y
    cmp     #KEYSTAR
    bne     .battle_config_loop
    lda     #$ff
    sta     keypad_last_key,y
    lda     title_control_mode,y
    eor     #1
    sta     title_control_mode,y
    jmp     .update_selections
.start_game
    ldy     #3
.control_loop
    lda     title_control_mode,y
    sta     control_modes,y
    dey
    bpl     .control_loop
    jmp     start_game

; Clear the main text area in preparation
; for a new menu to be displayed.
clear_text_area subroutine
    lda     #0
    tay
.loop
    sta     $3000,y
    sta     $3100,y
    cpy     #168
    bcs     .skip_32
    sta     $3200,y
.skip_32
    dey
    bne     .loop
    rts

; Position the player and missile sprites
; for the enemy display.
position_title_sprites subroutine
    lda     #106
    sta     HPOSP0
    lda     #70
    sta     HPOSP1
    lda     #150
    sta     HPOSP2
    lda     #112
    sta     HPOSM0
    lda     #76
    sta     HPOSM1
    lda     #156
    sta     HPOSM2
    rts

; Move all title screen sprites offscreen.
hide_title_sprites subroutine
    lda     #0
    sta     HPOSP0
    sta     HPOSP1
    sta     HPOSP2
    sta     HPOSM0
    sta     HPOSM1
    sta     HPOSM2
    rts

; Title screen vertical blank routine.
title_vblank subroutine
    lda     $fe
    pha
    lda     $ff
    pha
    jsr     TUNER
    pla
    sta     $ff
    pla
    sta     $fe
    jsr     keypad_cooldown_process
    lda     #0
    sta     ATRACT
    dec     title_table_time
    bne     .no_logo_change
    lda     #title_table_duration
    sta     title_table_time
    ldx     #title_table_length - 1
.title_shift_loop
    lda     title_color_table - 1,x
    sta     title_color_table,x
    dex
    bne     .title_shift_loop
    lda     sCOLOR0
    sta     title_color_table
    ldx     title_table_fill
    beq     .fill_down
    tay                             ; Save the value.
    and     #$0f                    ; Check brightness
    cmp     #$0e
    beq     .shift_to_down
    iny
    iny
    sty     sCOLOR0
    jmp     .no_logo_change
.shift_to_down
    dec     title_table_fill
    jmp     .drop_color
.fill_down
    tay                             ; Save the value.
    and     #$0f                    ; Check brightness.
    cmp     #$02
    beq     .shift_to_up
.drop_color
    dey
    dey
    sty     sCOLOR0
    jmp     .no_logo_change
.shift_to_up
    tya                             ; Carry is set, so add $0f
    adc     #$0f                    ; to go to the next color.
    sta     sCOLOR0
    inc     title_table_fill
.no_logo_change
    jmp     ENDINT

